//==============================================
//      File rClasses.pas
//
//      Components library for Delphi and BCB++.
//      .
//      Classes.
//
//      Copyright 1998-2001 Polaris Software
//      http://polesoft.da.ru
//      mailto: PolarisLib@mail.ru
//==============================================
unit rClasses;

{$I POLARIS.INC}

interface

uses Classes, rUtils, Windows, SysUtils,
     {$IFDEF POLARIS_D6}RxStrUtils, Variants{$ELSE}StrUtils{$ENDIF};

type

  { TVariantList }

  TVarListAction = (vlnAdd, vlnInsert, vlnExtract, vlnDelete, vlnChange, vlnClear);

  TVariantList = class(TObject)
  private
    FList: TList;
    function Get(const Index: Integer): Variant;
    procedure Put(const Index: Integer; const Value: Variant);
    function GetCount: Integer;
  protected
    procedure Notify(Value: Variant; Action: TVarListAction); virtual;
  public
    constructor Create;
    destructor Destroy; override;
    function IndexOf(Value: Variant): Integer;
    function Add(Item: Variant): Integer; virtual;
    function Remove(Item: Variant): Integer; virtual;
    function Extract(Item: Variant): Variant; virtual;
    function First: Variant;
    function Last: Variant;
    procedure Clear; virtual;
    procedure Toggle(Item: Variant);
    procedure Delete(Index: Integer); virtual;
    procedure Insert(Index : Integer; Item: Variant); virtual;
    procedure LoadFromString(S: String; Delims: TCharSet {$IFDEF POLARIS_D4}= [',']{$ENDIF};
      const AVarType: Integer {$IFDEF POLARIS_D4}= varInteger{$ENDIF});
    function SaveToString(Delim: String {$IFDEF POLARIS_D4}= ','{$ENDIF}): String;
    procedure Assign(Source: TVariantList);
    property Count: Integer read GetCount;
    property Items[const Index: Integer]: Variant read Get write Put; default;
  end;

implementation

{ TVariantList }

constructor TVariantList.Create;
begin
  inherited Create;
  FList := TList.Create;
end;

destructor TVariantList.Destroy;
begin
  FList.Free;
  FList := nil;
  inherited Destroy;
end;

function TVariantList.Add(Item: Variant): Integer;
var
  P : PVariant;
begin
  //if IndexOf(Item) < 0 then begin
    New(P);
    VarCopy(P^, Item);
    Result := FList.Add(P);
  {end
  else Result := -1;}
  Notify(Item, vlnAdd);
end;

procedure TVariantList.Clear;
var
  I: Integer;
begin
  for i := 0 to FList.Count - 1 do Dispose(PVariant(FList.Items[I]));
  FList.Clear;
  FList.Pack;
  Notify(Null, vlnClear);
end;

procedure TVariantList.Delete(Index: Integer);
var
  P: PVariant;
begin
  P := PVariant(FList.Items[Index]);
  FList.Delete(Index);
  FList.Pack;
  Notify(P^, vlnDelete);
  Dispose(P);
end;

procedure TVariantList.Put(const Index: Integer; const Value: Variant);
begin
  if (Index >= 0) and (Index < FList.Count) then
    VarCopy(PVariant(FList.Items[Index])^, Value);
  Notify(Value, vlnChange);
end;

function TVariantList.Get(const Index: Integer): Variant;
begin
  Result := Null;
  if (Index >= 0) and (Index < FList.Count) then
    Result := PVariant(FList.Items[Index])^;
end;

function TVariantList.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TVariantList.IndexOf(Value: Variant): Integer;
begin
  Result := 0;
  while (Result < FList.Count) and (CompareVar(Items[Result], Value) <> 0) do
    Inc(Result);
  if Result = FList.Count then Result := -1;
end;

procedure TVariantList.Notify(Value: Variant; Action: TVarListAction);
begin
end;

function TVariantList.Remove(Item: Variant): Integer;
begin
  Result := IndexOf(Item);
  if Result >= 0 then
    Delete(Result);
end;

procedure TVariantList.Toggle(Item: Variant);
var
  Index: Integer;
begin
  Index := IndexOf(Item);
  if Index < 0 then Add(Item)
  else Delete(Index);
end;

function TVariantList.Extract(Item: Variant): Variant;
var
  I: Integer;
begin
  Result := Null;
  I := IndexOf(Item);
  if I >= 0 then
  begin
    Result := Item;
    Delete(I);
    Notify(Result, vlnExtract);
  end;
end;

function TVariantList.First: Variant;
begin
  Result := Get(0);
end;

function TVariantList.Last: Variant;
begin
  Result := Get(Count - 1);
end;

procedure TVariantList.Insert(Index : Integer; Item: Variant);
var
  P: PVariant;
begin
  New(P);
  VarCopy(P^, Item);
  FList.Insert(Index, P);
  Notify(Item, vlnInsert);
end;

procedure TVariantList.LoadFromString(S: String; Delims: TCharSet;
  const AVarType: Integer);
var
  I: Integer;
  Value: Variant;
begin
  S := DelChars(S, '''');
  for i := 0 to WordCount(S, Delims) - 1 do begin
    case AVarType of
      varInteger: Value := StrToInt(ExtractWord(i + 1, S, Delims));
      varDate: Value := StrToDate(ExtractWord(i + 1, S, Delims));
      else Value := ExtractWord(i + 1, S, Delims);
    end;
    Add(Value);
  end;
end;

function TVariantList.SaveToString(Delim: String): String;
var
  J: Integer;
begin
  Result := EmptyStr;
  for J := 0 to Count - 1 do
    if VarType(Items[J]) = varDate then
      Result := Result + '''' + DateToStr(Items[J]) + '''' + Delim
    else
    if VarType(Items[J]) = varString then
      Result := Result + '''' + String(Items[J]) + '''' + Delim
    else
      Result := Result + String(Items[J]) + Delim;
  System.Delete(Result, Length(Result), 1);
end;

procedure TVariantList.Assign(Source: TVariantList);
var
  i: Integer;
begin
  if not Assigned(Source) then Exit;
  Clear;
  for i := 0 to Source.Count - 1 do
    Add(Source[i]);
end;

end.
