//==============================================
//      File rcomboedit.pas
//
//      Components library for Delphi and BCB++.
//      TrComboEdit.
//
//      Copyright 1998-2001 Polaris Software
//      http://members.xoom.com/PolarisSoft
//      mailto: PolarisLib@mail.ru
//==============================================
unit rComboEdit;

{$I POLARIS.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  rUtils, rVCLUtils, ToolEdit;

type
  TrCustomComboEdit = class(TCustomComboEdit)
  private
    FGlyphResKind: TBitmapResKind;
    FGlyphResource: string;
    FOnCloseUp: TNotifyEvent;
    function IsGlyph: Boolean;
    procedure SetGlyphResKind(Value: TBitmapResKind);
    procedure SetGlyphResource(Value: string);
  protected
    property GlyphResKind:  TBitmapResKind read FGlyphResKind write SetGlyphResKind
                                           default DefBRK;
    property GlyphResource: string read FGlyphResource write SetGlyphResource;
    property OnCloseUp: TNotifyEvent read FOnCloseUp write FOnCloseUp;
    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override;
    property Button;
  published
    property Align;
    property Alignment;
    property AutoSelect;
    property BorderStyle;
    property ButtonHint;
    property CharCase;
    property ClickKey;
    property Color;
    property Ctl3D;
    property DirectInput;
    property DragCursor;
    property DragMode;
    property EditMask;
    property Enabled;
    property Font;
    property GlyphKind;
    { Ensure GlyphKind is published before Glyph and ButtonWidth }
    property Glyph stored IsGlyph;
    property ButtonWidth;
    property HideSelection;
{$IFDEF POLARIS_D4}
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DragKind;
    property ParentBiDiMode;
{$ENDIF}
    property ImeMode;
    property ImeName;
    property MaxLength;
    property NumGlyphs;
    property OEMConvert;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Text;
    property Visible;
    property OnButtonClick;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
{$IFDEF POLARIS_D5}
    property OnContextPopup;
{$ENDIF}
{$IFDEF POLARIS_D4}
    property OnEndDock;
    property OnStartDock;
{$ENDIF}
  end;

  TrComboEdit = class(TrCustomComboEdit)
  published
    property GlyphResKind;
    property GlyphResource;
  end;

implementation

{ TrCustomComboEdit }

constructor TrCustomComboEdit.Create(AOwner: TComponent);
begin
  inherited;
  FGlyphResKind := DefBRK;
end;

function TrCustomComboEdit.IsGlyph;
begin
  Result := ((FGlyphResKind = rkNone) or (Trim(FGlyphResource) = '')) and (GlyphKind = gkCustom);
end;

procedure TrCustomComboEdit.SetGlyphResource;
begin
  if FGlyphResource <> Value then begin
    FGlyphResource := Value;
    UpdateGlyph(Self, Button.Glyph, FGlyphResKind, FGlyphResource);
  end;
end;

procedure TrCustomComboEdit.SetGlyphResKind;
begin
  if FGlyphResKind <> Value then begin
    FGlyphResKind := Value;
    UpdateGlyph(Self, Button.Glyph, FGlyphResKind, FGlyphResource);
  end;
end;

procedure TrCustomComboEdit.Loaded;
begin
  inherited;
  UpdateGlyph(Self, Button.Glyph, FGlyphResKind, FGlyphResource);
end;

end.
