//==============================================
//      File rdbgrid.pas
//
//      Components library for Delphi and BCB++.
//       TRxDBGrid.
//      Descendant of TRxDBGrid.
//
//      Copyright 1998-2002 Polaris Software
//      http://polesoft.da.ru
//      mailto: PolarisLib@mail.ru
//==============================================

unit rDBGrid;

{$I POLARIS.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Menus,
  DB, rSetupFields, RxDBCtrl, DBGrids, Grids, rDBConst, rUtils, rVCLUtils, rClasses,
  {$IFDEF POLARIS_D6}RxStrUtils, Variants{$ELSE}StrUtils{$ENDIF};

const
  scCheckKey = scCtrl + vk_Space;

type

  TBookmarks = class(TBookmarkList);
  TrDBGrid = class;
  TCharSet = Set of Char;
  TCellAttribute = procedure(Sender: TObject; Field: TField; Canvas: TCanvas; State: TGridDrawState) of object;
  TCheckListChangingEvent = procedure(Sender: TObject; Item: Variant;
    Action: TVarListAction; var AllowChange: Boolean) of object;
  TCheckListChange = procedure(Sender: TObject; Item: Variant;
    Action: TVarListAction) of object;
  TOnAfterSetupColumns = procedure(Sender: TObject; ModalResult: TModalResult) of object;
  TOnSetCanModify = procedure(Sender: TObject; Column: TColumn; var CanModify: Boolean) of object;


  //  TCheckListCheckInactiveEvent = procedure (Sender: TObject; Item: Variant; var Inactive: Boolean) of object;

  { TCheckList }

  TCheckMethod = (clmCheck, clmUncheck, clmToggle);

  TCheckList = class(TVariantList)
  private
    FGrid: TrDBGrid;
    FGroupChange: Boolean;
    FCheckListChanging: TCheckListChangingEvent;
    FCheckListChange: TCheckListChange;
    procedure SetGroupChange(const Value: Boolean);
  protected
    procedure Notify(Value: Variant; Action: TVarListAction); override;
  public
    constructor Create(AGrid: TrDBGrid);
    function Add(Item: Variant): Integer; override;
    function Remove(Item: Variant): Integer; override;
    procedure Insert(Index: Integer; Item: Variant); override;
    procedure Delete(Index: Integer); override;
    procedure Clear; override;
    function AllowChange(Item: Variant; Action: TVarListAction): Boolean;
    property GroupChange: Boolean read FGroupChange write SetGroupChange;
  end;

  { TrDBGrid }
  TrCheckedStyle = (csNone, csCheck, csRadio);

  TrDBGrid = class(TRxDBGrid)
  private
    FSelectedRect: TRect;
    //
    FRealChecked: Boolean;
    FCheckList: TCheckList;
    FCheckType: TrCheckedStyle;
    FCheckFieldName: string;
    FCheckShift: TShiftState;
    FCheckFields: TList;
    //    FOnCheckListCheckInactive: TCheckListCheckInactiveEvent;
    FOnAfterSetupColumns: TOnAfterSetupColumns;
    FOnSetCanModify: TOnSetCanModify;
    FFlatCheck: Boolean;
    FSaveOptions: TDBGridOptions;
    //
    FSetupField: TrSetupFields;
    FRegKey: string;
    //
    FNewDefaultRowHeight: Integer;
    FOnApplyCellAttribute: TCellAttribute;
    FGridAutoSize: Boolean;
    FGridAutoWidth: Boolean;
    FDisplayMemo: Boolean;
    FMultiLineTitles: Boolean;
    FDisplayImages: Boolean;
    FDisplayBool: Boolean;
    FRowSizingAllowed: Boolean;
    FCol: Integer;
    FRow: Integer;
    FResizeFieldName: string;
    FCheckKey: TShortCut;
    FSelecting: Boolean;
    FSelectionAnchor: TBookMarkStr;
    //
    function GetDataSource: TDataSource;
    procedure SetDataSource(Value: TDataSource);
    procedure UpdateTitlesHeight;
    procedure SetGridAutoSize(Value: Boolean);
    procedure SetGridAutoWidth(Value: Boolean);
    function GetDefaultRowHeight: Integer;
    procedure SetDisplayMemo(Value: Boolean);
    procedure SetMultiLineTitles(Value: Boolean);
    procedure SetRowSizingAllowed(Value: Boolean);
    procedure SetDisplayImages(Value: Boolean);
    procedure SetDisplayBool(Value: Boolean);
    procedure SetCheckBoxSize;
    procedure SetFixedCols(Value: Integer);
    function GetFixedCols: Integer;
    procedure SetRealChecked(Value: Boolean);
    procedure RestoreDefaults;
    //
    procedure SetCheckType(Value: TrCheckedStyle);
    function GetID: Variant;
    procedure SetOnCheckListChanging(const Value: TCheckListChangingEvent);
    function GetOnCheckListChanging: TCheckListChangingEvent;
    function GetOnCheckListChange: TCheckListChange;
    procedure SetOnCheckListChange(const Value: TCheckListChange);
    procedure SetResizeFieldName(const Value: string);
    procedure ResizeColumns;
    procedure WMChar(var Msg: TWMChar); message WM_CHAR;
    procedure SetFlatCheck(const Value: Boolean);
    function GetCheckFieldValue: Variant;
  protected
    FOnColResize: TNotifyEvent;
    procedure LinkActive(Value: Boolean); override;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure DrawColumnCell(const Rect: TRect; DataCol: Integer; Column: TColumn; State: TGridDrawState); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure DblClick; override;
    procedure DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState); override;
    procedure ColWidthsChanged; override;
    procedure LayoutChanged; override;
    procedure SetDefaultRowHeight(Value: Integer);
    procedure RowHeightsChanged; override;
    function  CanEditModify: Boolean; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure ApplyCellAttribute(Field: TField; Canvas: TCanvas; State: TGridDrawState);
    procedure CellAttribute(Field: TField; Canvas: TCanvas; State: TGridDrawState); dynamic;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function IndexColumnOf(AFieldName: string): TColumn;
    property DataLink;
    property ColCount;
    procedure RunSetup;
    procedure SetRegKey(Value: string);
    property SetupFields: TrSetupFields read FSetupField;
    //
    procedure SetCheckFieldName(Value: string);
    property CheckList: TCheckList read FChecklist;
    property RealChecked: Boolean read FRealChecked write SetRealChecked default False;
    procedure CheckSelected(CheckMethod: TCheckMethod); //        
    procedure CheckAll(CheckMethod: TCheckMethod); //  
    procedure DoCheck(CheckMethod: TCheckMethod{$IFDEF POLARIS_D4} = clmToggle{$ENDIF});
    procedure CreateNewCheckColumn;
    procedure CheckFromString(S: String; CheckMethod: TCheckMethod; Delims: TCharSet{$IFDEF POLARIS_D4} = [',']{$ENDIF});
    //
    property SelectedRect: TRect read FSelectedRect;
    property CheckFieldValue: Variant read GetCheckFieldValue;
  published
    { Published declarations }
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property RegKey: string read FRegKey write SetRegKey;
    //
    property GridAutoSize: Boolean read FGridAutoSize write SetGridAutoSize default False;
    property GridAutoWidth: Boolean read FGridAutoWidth write SetGridAutoWidth default False;
    property DisplayMemo: Boolean read FDisplayMemo write SetDisplayMemo default False;
    property MultiLineTitles: Boolean read FMultiLineTitles write SetMultiLineTitles default False;
    property DefaultRowHeight: Integer read GetDefaultRowHeight write SetDefaultRowHeight default 17;
    property RowSizingAllowed: Boolean read FRowSizingAllowed write SetRowSizingAllowed default False;
    property OnApplyCellAttribute: TCellAttribute read FOnApplyCellAttribute write FOnApplyCellAttribute;
    property DisplayImages: Boolean read FDisplayImages write SetDisplayImages default False;
    property DisplayBoolean: Boolean read FDisplayBool write SetDisplayBool default False;
    property FixedCols: Integer read GetFixedCols write SetFixedCols default 0;
    property ResizeFieldName: string read FResizeFieldName write SetResizeFieldName {stored IsEmpty };
    //
    property CheckFieldName: string read FCheckFieldName write SetCheckFieldName;
    property CheckType: TrCheckedStyle read FCheckType write SetCheckType default csNone;
    property CheckKey: TShortCut read FCheckKey write FCheckKey default scCheckKey;
    property FlatCheck: Boolean read FFlatCheck write SetFlatCheck default False;
    property OnCheckListChanging: TCheckListChangingEvent read GetOnCheckListChanging write SetOnCheckListChanging;
    property OnCheckListChange: TCheckListChange read GetOnCheckListChange write SetOnCheckListChange;
    //    property OnCheckListCheckInactive: TCheckListCheckInactiveEvent read FOnCheckListCheckInactive write FOnCheckListCheckInactive;
    property OnAfterSetupColumns: TOnAfterSetupColumns read FOnAfterSetupColumns write FOnAfterSetupColumns;
    property OnSetCanModify: TOnSetCanModify read FOnSetCanModify write FOnSetCanModify;
    property CheckShift: TShiftState read FCheckShift write FCheckShift default [ssShift];
  end;

  TrDBGridMod = class(TCustomGrid)
  public
    property Options;
  end;

const
  MinRowHeight = 17;

implementation

uses rDBGridCtls;

var
  GCheckWidth, GCheckHeight: Integer;

procedure DrawGridColumnButton(Canvas: TCanvas; Rect: TRect; Style: TrCheckedStyle;
  Checked: Boolean{$IFDEF POLARIS_D4} = False{$ENDIF}; Inactive: Boolean{$IFDEF POLARIS_D4} = False{$ENDIF};
  FFlatCheck: Boolean{$IFDEF POLARIS_D4} = False{$ENDIF});
begin
  Canvas.FillRect(Rect);
  if FFlatCheck then InFlateRect(Rect, -2, -2)
  else InFlateRect(Rect, -1, -1);
  DrawFrameControl(Canvas.Handle, Rect, DFC_BUTTON,
    IIF(Style = csCheck,
    DFCS_BUTTONCHECK,
    DFCS_BUTTONRADIO)
    or (DFCS_CHECKED * Integer(Checked))
    or (DFCS_INACTIVE * Integer(InActive) or DFCS_FLAT * Integer(FFlatCheck)));
end;

{ TCheckList }

constructor TCheckList.Create(AGrid: TrDBGrid);
begin
  inherited Create;
  FGroupChange := False;
  FGrid := AGrid;
end;

procedure TCheckList.Notify(Value: Variant; Action: TVarListAction);
var
  R: TRect;
begin
  if Assigned(FGrid) and not FGrid.ScrollDisabled and not FGroupChange then
  begin
    R := FGrid.ClientRect;
    R.Left := 0;
    R.Right := R.Left + FGrid.ClientWidth;
    RefreshWindow(FGrid, @R);
  end;
end;

procedure TCheckList.SetGroupChange(const Value: Boolean);
begin
  if Value <> FGroupChange then FGroupChange := Value;
end;

function TCheckList.AllowChange(Item: Variant; Action: TVarListAction): Boolean;
begin
  Result := True;
  if Assigned(FCheckListChanging) then FCheckListChanging(FGrid, Item, Action, Result);
end;

function TCheckList.Add(Item: Variant): Integer;
begin
  if AllowChange(Item, vlnAdd) then begin
    if (FGrid <> nil) and (FGrid.CheckType = csRadio) then begin
      if Count = 0 then inherited Add(Item)
      else items[0] := Item;
      Result := 0;
    end
    else
      Result := inherited Add(Item);
    if Assigned(FCheckListChange) then FCheckListChange(FGrid, Item, vlnAdd);
  end
  else Result := -1;
end;

function TCheckList.Remove(Item: Variant): Integer;
begin
  if AllowChange(Item, vlnDelete) then begin
    Result := inherited Remove(Item);
    if Assigned(FCheckListChange) then FCheckListChange(FGrid, Item, vlnDelete);
  end
  else Result := -1;
end;

procedure TCheckList.Insert(Index: Integer; Item: Variant);
begin
  if AllowChange(Item, vlnInsert) then begin
    inherited Insert(Index, Item);
    if Assigned(FCheckListChange) then FCheckListChange(FGrid, Item, vlnInsert);
  end;
end;

procedure TCheckList.Delete(Index: Integer);
begin
  if AllowChange(Items[Index], vlnDelete) then begin
    inherited Delete(Index);
    if Assigned(FCheckListChange) then FCheckListChange(FGrid, Items[Index], vlnDelete);
  end;
end;

procedure TCheckList.Clear;
begin
  inherited Clear;
  if Assigned(FCheckListChange) then FCheckListChange(FGrid, Null, vlnClear);
end;

{ TrDBGrid }

constructor TrDBGrid.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetCheckBoxSize;
  GridAutoSize := False;
  GridAutoWidth := False;
  MultiLineTitles := False;
  FRowSizingAllowed := False;
  FDisplayMemo := False;
  FDisplayBool := False;
  FDisplayImages := False;
  FCheckType := csNone;
  FRealChecked := False;
  FFlatCheck := False;
  FCheckKey := scCheckKey;
  FCheckShift := [ssShift];
  if not (csDesigning in ComponentState) then
    FSetupField := TrSetupFields.Create(AOwner);
  FCheckList := TCheckList.Create(Self);
  DefaultRowHeight := 17;
  FSaveOptions := Options;
  FCheckFields := TList.Create;
end;

destructor TrDBGrid.Destroy;
begin
  if Assigned(FSetupField) then FSetupField.Free;
  if Assigned(FCheckList) then FCheckList.Free;
  if Assigned(FCheckFields) then FCheckFields.Free;
  inherited Destroy;
end;

procedure TrDBGrid.Loaded;
begin
  inherited Loaded;
  RowSizingAllowed := FRowSizingAllowed;
  SetCheckType(FCheckType);
end;

procedure TrDBGrid.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited;
  if (Operation = opRemove) then
    if (AComponent = FSetupField) and (FSetupField <> nil) then
      FSetupField := nil;
end;

function TrDBGrid.GetFixedcols: Integer;
begin
  Result := inherited FixedCols;
  if FRealChecked then Result := Result - 1;
end;

procedure TrDBGrid.SetFixedCols(Value: Integer);
begin
  if FRealChecked then Value := Value + 1;
  inherited FixedCols := Value;
end;

procedure TrDBGrid.SetRealChecked(Value: Boolean);
begin
  if FRealChecked <> Value then begin
    FRealChecked := Value;
    if Value then FixedCols := FixedCols + 1
    else FixedCols := FixedCols - 1;
  end;
end;

procedure TrDBGrid.RunSetup;
begin
  if FSetupField <> nil then SetupFields.Execute;
end;

procedure TrDBGrid.SetRegKey;
begin
  if (Value <> FRegKey) then begin
    FRegKey := Value;
    if not (csDesigning in ComponentState) then
      FSetupField.RegKey := Value;
  end;
end;

function TrDBGrid.GetDataSource;
begin
  Result := inherited DataSource;
end;

procedure TrDBGrid.SetDataSource(Value: TDataSource);
begin
  inherited DataSource := Value;
  if not (csDesigning in ComponentState) then begin
    if Value <> nil then begin
      FSetupField.DBGrid := Self;
      FSetupField.RegKey := FRegKey;
    end else FSetupField.DBGrid := nil;
  end;
end;

procedure TrDBGrid.SetGridAutoSize(Value: Boolean);
var
  RowNo, ColNo, WidthNew, WidthMax: Integer;
begin
  FGridAutoSize := Value;
  if FGridAutoSize = True then FGridAutoWidth := False;
  if not (csDesigning in ComponentState) then begin
    if (DataLink <> nil) and (DataLink.Active) then begin
      DataLink.DataSet.DisableControls;
      try
        if FGridAutoSize = True then begin
          for ColNo := Byte(FRealChecked) to (ColCount - 2) do begin
            WidthMax := 0;
            WidthNew := 0;
            DataLink.DataSet.First;
            for RowNo := 0 to (RowCount - 1) do
              if Fields[ColNo].Value <> Null then
                WidthNew := Canvas.TextWidth(Fields[ColNo].DisplayText);
            if WidthNew > WidthMax then WidthMax := WidthNew;
            DataLink.DataSet.Next;
            try
              Columns[ColNo].Width := (WidthMax + (GridLineWidth + 5))
            except
            end;
            DataLink.DataSet.First;
          end;
          Invalidate;
        end
        else RestoreDefaults;
      finally
        DataLink.DataSet.EnableControls;
      end;
    end;
  end;
end;

procedure TrDBGrid.SetGridAutoWidth(Value: Boolean);
var
  Available, Remaining, i, w, n: Integer;
begin
  FGridAutoWidth := Value;
  if FGridAutoWidth = True then FGridAutoSize := False;
  if not (csDesigning in ComponentState) then begin
    if (DataLink <> nil) and (DataLink.Active) then begin
      DataLink.DataSet.DisableControls;
      try
        if FGridAutoWidth = True then begin
          with TrDBGrid(Self) do begin
            Available := ClientWidth;
            n := FixedCols + Byte(FRealChecked);
            for i := 0 to (n) do
              Dec(Available, ColWidths[i] + GridlineWidth);
            Remaining := ColCount - (n + 1);
            w := Available div Remaining;
            for i := (n + 1) to ColCount - 2 do begin
              ColWidths[i] := w - GridLineWidth;
              Dec(Available, w);
            end;
            ColWidths[ColCount - 1] := Available;
          end;
          Invalidate;
        end
        else SetGridAutoSize(True);
      finally
        DataLink.DataSet.EnableControls;
      end;
    end;
  end;
end;

procedure TrDBGrid.SetDisplayMemo(Value: Boolean);
begin
  if (FDisplayMemo <> Value) then begin
    FDisplayMemo := Value;
    Invalidate;
  end;
end;

procedure TrDBGrid.SetMultiLineTitles(Value: Boolean);
begin
  if FMultiLineTitles <> Value then
    FMultiLineTitles := Value;
  UpdateTitlesHeight;
end;

procedure TrDBGrid.UpdateTitlesHeight;
var
  i, j, MaxHeight: Integer;
  RRect: TRect;
begin
  if FMultiLineTitles then begin
    if (dgTitles in Options) then begin
      MaxHeight := 0;
      for i := 0 to Columns.Count - 1 do begin
        RRect := CellRect(0, 0);
        RRect.Left := 0;
        RRect.Right := Columns[i].Width;

        Canvas.Font := Columns[i].Title.Font;
          j := DrawText(Canvas.Handle,
          PChar(Columns[i].Title.Caption),
          Length(Columns[i].Title.Caption),
          RRect,
          DT_EXPANDTABS or DT_CALCRECT or DT_WORDBREAK);
        if j > MaxHeight then MaxHeight := j;
      end;
      if (MaxHeight <> 0) then begin
        if (dgRowLines in Options) then Inc(MaxHeight, 3)
        else Inc(MaxHeight, 2);
        RowHeights[0] := MaxHeight;
        if RowHeights[0] < MinRowHeight then RowHeights[0] := MinRowHeight;
      end;
    end;
  end
  else if RowHeights[0] <> MinRowHeight then RowHeights[0] := MinRowHeight;
end;

procedure TrDBGrid.DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState);
const
  AlignFlags: array[TAlignment] of Integer =
  (DT_LEFT or DT_WORDBREAK or DT_EXPANDTABS or DT_NOPREFIX,
    DT_RIGHT or DT_WORDBREAK or DT_EXPANDTABS or DT_NOPREFIX,
    DT_CENTER or DT_WORDBREAK or DT_EXPANDTABS or DT_NOPREFIX);
var
  TitleText: string;
  TmpCol: LongInt;
  TmpRect: TRect;
  AField: TField;
  BackColor: TColor;
  SortMarker: TSortMarker;
begin
  inherited DrawCell(ACol, ARow, ARect, AState);

  if AState = [gdSelected, gdFocused] then FSelectedRect := ARect;

  if (dgIndicator in Options) then TmpCol := ACol - 1
  else TmpCol := ACol;

  with ARect do begin
    if (FMultiLineTitles) and (ARow = 0) and (TmpCol >= 0)
      and (dgTitles in Options) and (Assigned(Columns[TmpCol])) then begin
      TitleText := Columns[TmpCol].Title.Caption;
      if TitleText <> EmptyStr then begin
        Canvas.Font := Columns[TmpCol].Title.Font;
        Canvas.Brush.Color := Columns[TmpCol].Title.Color;
        TmpRect := ARect;
        Inc(TmpRect.Top, 2);
        Inc(TmpRect.Left, 2);
        Dec(TmpRect.Right, 1);
        Dec(TmpRect.Bottom, 1);
        Canvas.FillRect(TmpRect);
        DrawText(Canvas.Handle, PChar(TitleText), Length(TitleText), TmpRect, AlignFlags[Columns[TmpCol].Title.Alignment]);
        SortMarker := smNone;
        AField := Columns[TmpCol].Field;
        if (AField <> nil) and (Assigned(OnGetBtnParams)) then begin
          BackColor := Canvas.Brush.Color;
          OnGetBtnParams(Self, AField, Canvas.Font, BackColor, SortMarker, False);
          Canvas.Brush.Color := BackColor;
        end;

        if (AField <> nil) then with Canvas, ARect do
            case SortMarker of
              smDown: begin
                  Pen.Color := clBtnShadow;
                  MoveTo(Right - 13, Bottom - 12);
                  LineTo(Right - 6, Bottom - 12);
                  Pen.Color := clBtnHighlight;
                  LineTo(Right - 9, Bottom - 5);
                  Pen.Color := clBtnShadow;
                  LineTo(Right - 13, Bottom - 12);
                  Pixels[Right - 5, Bottom - 12] := clBtnShadow;
                end;
              smUp: begin
                  Pen.Color := clBtnHighlight;
                  MoveTo(Right - 13, Bottom - 5);
                  LineTo(Right - 5, Bottom - 6);
                  LineTo(Right - 9, Bottom - 12);
                  Pen.Color := clBtnShadow;
                  LineTo(Right - 13, Bottom - 5);
                  Pixels[Right - 13, Bottom - 5] := clBtnShadow;
                end;
            end;
      end;
    end;
  end;
end;

procedure TrDBGrid.ColWidthsChanged;
begin
  inherited;
  try
    if Assigned(FOnColResize) then FOnColResize(TrDBGrid(Self));
  except
  end;
end;

procedure TrDBGrid.LayoutChanged;
begin
  inherited LayoutChanged;
  SetDefaultRowHeight(FNewDefaultRowHeight);
  if FMultiLineTitles then UpdateTitlesHeight;
  ResizeColumns;
end;

function TrDBGrid.GetDefaultRowHeight: Integer;
begin
  Result := inherited DefaultRowHeight;
end;

procedure TrDBGrid.SetDefaultRowHeight(Value: Integer);
var
  RRow: Integer;
begin
  if not (csDestroying in ComponentState) then begin
    if dgTitles in Options then RRow := RowHeights[0]
    else RRow := 17;
    if Value <= 0 then Value := inherited DefaultRowHeight;
    inherited DefaultRowHeight := Value;
    FNewDefaultRowHeight := Value;
    if dgTitles in Options then RowHeights[0] := RRow;
  end;
end;

procedure TrDBGrid.RowHeightsChanged;
var
  i, ThisHasChanged, Def: Integer;
begin
  ThisHasChanged := -1;
  Def := DefaultRowHeight;
  for i := Ord(dgTitles in Options) to RowCount do
    if RowHeights[i] <> Def then begin
      ThisHasChanged := i;
      Break;
    end;
  if ThisHasChanged <> -1 then begin
    SetDefaultRowHeight(RowHeights[ThisHasChanged]);
    RecreateWnd;
  end;
  inherited;
end;

procedure TrDBGrid.CMFontChanged(var Message: TMessage);
var
  h: integer;
begin
  inherited;
  Canvas.Font.Assign(Font);
  h := Canvas.TextHeight('Wg');
  if MinRowHeight < h then SetDefaultRowHeight(h);
  if (csDesigning in ComponentState) then Invalidate;
end;

procedure TrDBGrid.SetRowSizingAllowed(Value: Boolean);
begin
  FRowSizingAllowed := Value;
  with TrDBGridMod(Self) do begin
    Options := Options + [goColSizing, goThumbTracking, goColMoving, goRowMoving, goDrawFocusSelected];
    if Value then
      Options := Options + [goRowSizing]
    else
      Options := Options - [goRowSizing];
  end;
end;

procedure TrDBGrid.DrawColumnCell(const Rect: TRect; DataCol: Integer; Column: TColumn; State: TGridDrawState);
var
  Bmp: TBitmap;
  Style: UINT;
  TmpRect: TRect;
  FactorX, FactorY: Double;
  TmpWidth, TmpHeight: Integer;
  CWidth, CHeight, BWidth, BHeight: Integer;
  Inactive: Boolean;
  ID: Variant;
begin
  inherited DrawColumnCell(Rect, DataCol, Column, State);
  if DefaultDrawing then
    with Canvas do begin
      if (Column <> nil) and (Column.Field <> nil) then begin
        if not (gdSelected in State) and not (gdFixed in State) then
          ApplyCellAttribute(Column.Field, Canvas, State);
        if (Column.Field.DataType = ftGraphic) or
          (Column.Field.DataType = ftTypedBinary) then begin
          if (FDisplayImages) then begin
            Bmp := TBitmap.Create;
            try
              Bmp.Assign(Column.Field);
              begin
                CWidth := Rect.Right - Rect.Left + 1 - 1;
                CHeight := Rect.Bottom - Rect.Top + 1 - 1;
                BWidth := Bmp.Width;
                BHeight := Bmp.Height;
                try
                  if BWidth = 0 then FactorX := 0
                  else FactorX := CWidth / BWidth;
                  if BHeight = 0 then FactorY := 0
                  else FactorY := CHeight / BHeight;
                except
                  FactorX := 1;
                  FactorY := 1;
                end;
                if FactorY < FactorX then begin
                  TmpWidth := Round(BWidth * FactorY);
                  TmpHeight := CHeight;
                end
                else begin
                  TmpHeight := Round(BHeight * FactorX);
                  TmpWidth := CWidth;
                end;
                TmpRect := Rect;
                FillRect(Rect);
                SetRect(TmpRect, 0, 0, TmpWidth, TmpHeight);
                OffSetRect(TmpRect, Rect.Left + ((CWidth - TmpWidth) div 2), Rect.Top + ((CHeight - TmpHeight) div 2));
                StretchDraw(TmpRect, Bmp);
              end
            finally
              Bmp.Free;
            end;
          end
        end

        else if (Column.Field is TMemoField) then begin
          if (FDisplayMemo) then begin
            TmpRect := Rect;
            FillRect(Rect);
            Inc(TmpRect.Top, 2);
            Inc(TmpRect.Left, 2);
            DrawText(Handle, PChar(Column.Field.AsString), Length(Column.Field.AsString), TmpRect, DT_WORDBREAK or DT_NOPREFIX);
          end
        end

        else if (Column.Field is TBooleanField) then begin
          if (FDisplayBool) then begin
            TmpRect := Rect;
            FillRect(Rect);
            TmpRect.Left := Rect.Left + (Rect.Right - Rect.Left - GCheckWidth) div 2;
            TmpRect.Right := TmpRect.Left + GCheckWidth;
            TmpRect.Top := Rect.Top + (Rect.Bottom - Rect.Top - GCheckWidth) div 2;
            TmpRect.Bottom := TmpRect.Top + GCheckHeight;
            Style := DFCS_CHECKED;
            if Column.Field.AsBoolean = True then Style := DFCS_CHECKED
            else if Column.Field.AsBoolean = False then Style := DFCS_BUTTONCHECK;
            DrawFrameControl(Handle, TmpRect, DFC_BUTTON, Style);
          end
        end
      end;
    end;

  if FRealChecked and (Column.Index = 0) then begin
    if FFlatCheck then Canvas.Brush.Color := Color;
    ID := GetID;
    if VarIsNull(ID) then
      DrawGridColumnButton(Canvas, Rect, FCheckType, False, True, FFlatCheck)
    else begin
      Inactive := False;
      //      if Assigned(FOnCheckListCheckInactive) then FOnCheckListCheckInactive(Self, Integer(ID), Inactive);
      DrawGridColumnButton(Canvas, Rect, FCheckType, FCheckList.IndexOf(ID) > -1, Inactive, FFlatCheck);
    end;
  end;
end;

procedure TrDBGrid.ApplyCellAttribute(Field: TField; Canvas: TCanvas; State: TGridDrawState);
begin
  if Assigned(FOnApplyCellAttribute) then
    FOnApplyCellAttribute(Self, Field, Canvas, State);
end;

procedure TrDBGrid.CellAttribute(Field: TField; Canvas: TCanvas; State: TGridDrawState);
begin
  ApplyCellAttribute(Field, Canvas, State);
end;

procedure TrDBGrid.SetDisplayImages(Value: Boolean);
begin
  if (FDisplayImages <> Value) then begin
    FDisplayImages := Value;
    Invalidate;
  end;
end;

procedure TrDBGrid.WMSize(var Message: TWMSize);
begin
  try
    inherited;
    if (DataLink <> nil) and (Datalink.Active) then begin
      if FGridAutoSize then SetGridAutoSize(FGridAutoSize);
      if FGridAutoWidth then SetGridAutoWidth(FGridAutoWidth);
      ResizeColumns;
    end;
  except
  end;
end;

procedure TrDBGrid.MouseUp(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var
  ID: Variant;
  Inactive: Boolean;
begin
  inherited MouseUp(Button, Shift, X, Y);
  TrDBGrid(Self).MouseToCell(X, Y, FCol, FRow);
  if (dgIndicator in Options) then FCol := FCol - Byte(FCol <> 0);

  if FRealChecked then begin
    ID := GetID;
    if not (ssDouble in Shift) and (Button = mbLeft) and not VarIsNull(ID) then begin
      if ((FRow > 0) or (not (dgTitles in Options))) and (FCol = 0) then begin
        Inactive := False;
        //        if Assigned(FOnCheckListCheckInactive) then FOnCheckListCheckInactive(Self, Integer(ID), Inactive);
        if not Inactive then begin
          if FCheckType = csCheck then FCheckList.Toggle(ID)
          else FCheckList.Add(ID);
        end;
      end;
    end;
  end;
end;

procedure TrDBGrid.DblClick;
var
  AField: TField;
begin
  if Assigned(OnDblClick) then inherited DblClick
  else begin
    if (FCol >= 0) and Assigned(Columns[FCol].Field) then AField := Columns[FCol].Field
    else AField := nil;
    if not (csDesigning in ComponentState) then begin
      BeginUpdate;
      if ((Datalink <> nil) and (Datalink.Active) and (AField <> nil)) then begin
        if (FDisplayBool) and (AField.DataType = ftBoolean)
          and CanEditModify then begin
          if not (ReadOnly) and not (DataLink.Readonly) then begin
            DataSource.Dataset.Edit;
            try
              AField.AsBoolean := not AField.AsBoolean;
            except
              raise;
            end;
          end;
        end;
        if AField.DataType in [ftGraphic, ftTypedBinary, ftMemo, ftFmtMemo] then begin
          with TrDBGridCtlsForm.Create(Application) do try
            if AField.DataType in [ftGraphic, ftTypedBinary] then
              BlobCaption := srGridImage
            else
              BlobCaption := srGridMemo;
            BlobField := AField;
            DataSource := Self.DataSource;
            if not (ReadOnly) and not (DataLink.ReadOnly) then begin
              DataSource.DataSet.Edit;
              try
                ShowModal;
              except
                raise;
              end;
            end else ShowModal;
          finally
            Free;
          end;
        end
      end;
      EndUpdate;
    end;
  end;
end;

procedure TrDBGrid.SetDisplayBool(value: Boolean);
begin
  if (FDisplayBool <> Value) then begin
    FDisplayBool := Value;
    Invalidate;
  end;
end;

procedure TrDBGrid.SetCheckBoxSize;
begin
  with TBitmap.Create do try
    Handle := LoadBitmap(0, PChar(32759));
    GCheckWidth := Width div 4;
    GCheckHeight := Height div 3;
  finally
    Free;
  end;
end;

procedure TrDBGrid.SetCheckFieldName(Value: string);

  procedure SetCheckColumn(F: TField);
  var
    FColumn: TColumn;
    i: Integer;
  begin
    if F <> nil then F.Tag := fSetupField.ExcludeTag;
    if (Columns.State = csDefault) then
    begin
      if (F <> nil) then begin
        F.Index := 0;
        F.Visible := True;
        F.DisplayLabel := ' ';
      end
      else begin
        if not FRealChecked then
          with Columns.Add do begin
            Title.Caption := EmptyStr;
            Index := 0;
            Width := DefaultRowHeight;
          end;
        for i := 0 to DataSource.DataSet.Fields.Count - 1 do
          if DataSource.DataSet.Fields[I].Visible then
            with Columns.Add do
              Field := DataSource.DataSet.Fields[I];
      end;
    end
    else begin
      if not FRealChecked and ((F = nil) or (Columns[0].Field <> F)) then
      begin
        FColumn := Columns.Add;
        FColumn.Title.Caption := EmptyStr;
        FColumn.Index := 0;
      end;
      if F <> nil then Columns[0].Field := F;
    end;
    Columns[0].Width := DefaultRowHeight;
    RealChecked := True;
    Invalidate;
  end;

  procedure SetUnCheckColumn(F: TField);
  begin
    //    ....
    if FRealChecked then begin
      try
        RealChecked := False;
        FCheckList.Clear;
      finally
        if Columns.State = csDefault then
          DataSource.Dataset.Fields[0].Visible := False
        else Columns[0].Destroy;
        Invalidate;
      end;
    end;
  end;

var
  F: TField;
begin
  FCheckFieldName := Value;
  if not (csDesigning in ComponentState) then begin
    if (DataSource <> nil) and (DataSource.DataSet <> nil)
      and (DataSource.DataSet.Active) then
    begin
      FCheckFields.Clear;
      DataSource.DataSet.GetFieldList(FCheckFields, FCheckFieldName);
      if FCheckFields.Count = 0 then Exit
      else if FCheckFields.Count = 1 then F := FCheckFields[0]
      else F := nil;
      //F := DataSource.DataSet.FindField(Value);
      if (FCheckType <> csNone) {and (F <> nil)} then SetCheckColumn(F)
      else SetUnCheckColumn(F);
    end
    //else FRealChecked := False;
  end;
end;

procedure TrDBGrid.LinkActive(Value: Boolean);
begin
  inherited LinkActive(Value);
  if not (csLoading in ComponentState) then
    SetCheckType(FCheckType);
end;

procedure TrDBGrid.SetCheckType(Value: TrCheckedStyle);
begin
  if FCheckType <> Value then FCheckList.Clear;
  FCheckType := Value;
  SetCheckFieldName(FCheckFieldName);
end;

procedure TrDBGrid.CheckSelected(CheckMethod: TCheckMethod);
var
  BM: TBookmark;
  i: Integer;
begin
  if not (FRealChecked and Assigned(DataSource) and
    Assigned(DataSource.DataSet) and DataSource.DataSet.Active) then
    Exit;
  with DataSource.DataSet do
    if not (Bof and Eof) then
      if SelectedRows.Count > 0 then begin
        BM := GetBookmark;
        DisableControls;
        FChecklist.GroupChange := True;
        try
          for i := 0 to SelectedRows.Count - 1 do begin
            GotoBookmark(Pointer(SelectedRows.Items[I]));
              DoCheck(CheckMethod);
          end;
          GotoBookmark(BM);
        finally
          FChecklist.GroupChange := False;
          FreeBookmark(BM);
          SelectedRows.Clear;
          EnableControls;
        end;
      end
      else DoCheck(CheckMethod);
end;

procedure TrDBGrid.CheckAll(CheckMethod: TCheckMethod);
var
  BM: TBookmark;
begin
  if not (FRealChecked and Assigned(DataSource) and
    Assigned(DataSource.DataSet) and DataSource.DataSet.Active) then
    Exit;
  with DataSource.DataSet do begin
    DisableControls;
    BM := GetBookmark;
    FCheckList.GroupChange := True;
    try
      First;
      while not Eof do begin
        DoCheck(CheckMethod);
        Next;
      end;
      if (BM <> nil) and BookmarkValid(BM) then GotoBookmark(BM);
    finally
      FChecklist.GroupChange := False;
      FreeBookmark(BM);
      SelectedRows.Clear;
      EnableControls;
    end;
  end;
end;

function TrDBGrid.GetID: Variant;
begin
  if FRealChecked then
    Result := CheckFieldValue
  else
    Result := Null;
end;

procedure TrDBGrid.DoCheck(CheckMethod: TCheckMethod);
var
  Item: Variant;
begin
  if FRealChecked then begin
    Item := CheckFieldValue;
    if not VarIsNull(Item) then
      case CheckMethod of
        clmCheck: if FCheckList.IndexOf(Item) < 0 then
          FCheckList.Add(Item);
        clmUncheck: FCheckList.Remove(Item);
        clmToggle: FCheckList.Toggle(Item);
      end;
  end;
end;

procedure TrDBGrid.RestoreDefaults;
var
  I: Integer;
begin
  with Columns do begin
    BeginUpdate;
    try
      for I := Byte(FRealChecked) to Count - 1 do
        Items[I].RestoreDefaults;
    finally
      EndUpdate;
    end;
  end;
end;

procedure TrDBGrid.CreateNewCheckColumn;
begin
  with Columns.Add do begin
    Index := 0;
    Title.Caption := EmptyStr;
    if FCheckFields.Count = 1 then Field := FCheckFields[0]
    else Field := nil;
//    FieldName := FCheckFieldName;
    Width := DefaultRowHeight;
  end;
end;

procedure TrDBGrid.SetOnCheckListChanging(
  const Value: TCheckListChangingEvent);
begin
  if Assigned(FCheckList) then
    FCheckList.FCheckListChanging := Value;
end;

function TrDBGrid.GetOnCheckListChanging: TCheckListChangingEvent;
begin
  Result := nil;
  if Assigned(FCheckList) then
    Result := FCheckList.FCheckListChanging;
end;

function TrDBGrid.GetOnCheckListChange: TCheckListChange;
begin
  Result := nil;
  if Assigned(FCheckList) then
    Result := FCheckList.FCheckListChange;
end;

procedure TrDBGrid.SetOnCheckListChange(const Value: TCheckListChange);
begin
  if Assigned(FCheckList) then
    FCheckList.FCheckListChange := Value;
end;

procedure TrDBGrid.SetResizeFieldName(const Value: string);
begin
  if FResizeFieldName <> Value then begin;
    FResizeFieldName := Value;
    ResizeColumns;
  end;
end;

procedure TrDBGrid.ResizeColumns;
var
  w, i, j: Integer;
begin
  if not (csDesigning in ComponentState) then
  begin
    if (FResizeFieldName <> EmptyStr) and (Columns.State = csDefault) then
    begin
      w := ClientWidth - Integer(dgIndicator in Options) * 12 - Columns.Count - 1;
      j := -1;
      for i := 0 to Columns.Count - 1 do
        if (Columns[i].FieldName <> EmptyStr) and (Columns[i].FieldName <> FResizeFieldName) then
          w := w - Columns[i].Width
        else j := i;
      if (w > 0) and (j <> -1) then Columns[j].Width := w;
    end;
  end;
end;

procedure TrDBGrid.WMChar(var Msg: TWMChar);
var
  sc: TShortCut;
  ID: Variant;
  Inactive: Boolean;
begin
  sc := ShortCut(Msg.CharCode, KeyDataToShiftState(Msg.KeyData));
  if FRealChecked and (sc = FCheckKey) then begin
    ID := GetID;
    if not VarIsNull(ID) then begin
      Inactive := False;
      //        if Assigned(FOnCheckListCheckInactive) then FOnCheckListCheckInactive(Self, Integer(ID), Inactive);
      if not Inactive then begin
        if FCheckType = csCheck then FCheckList.Toggle(ID)
        else FCheckList.Add(ID);
      end;
    end;
    Msg.CharCode := 0;
    Msg.Result := 1;
  end
  else inherited;
end;

procedure TrDBGrid.SetFlatCheck(const Value: Boolean);
begin
  if FFlatCheck <> Value then begin;
    FFlatCheck := Value;
    if FFlatCheck then begin
      FSaveOptions := Options;
      Options := Options - [dgRowLines, dgColLines, dgTitles, dgIndicator];
    end
    else Options := FSaveOptions;
  end;
end;

function TrDBGrid.IndexColumnOf(AFieldName: string): TColumn;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to Columns.Count - 1 do
    if AnsiCompareText(Columns[i].FieldName, AFieldName) = 0 then begin
      Result := Columns[i];
      Break;
    end;
end;

procedure TrDBGrid.CheckFromString(S: String; CheckMethod: TCheckMethod;
  Delims: TCharSet);
var
  I: Integer;
  Value: String;
begin
  for i := 0 to WordCount(S, Delims) - 1 do begin
    Value := ExtractWord(i + 1, S, Delims);
    case CheckMethod of
      clmCheck: FCheckList.Add(Value);
      clmUncheck: FCheckList.Remove(Value);
      clmToggle: FCheckList.Toggle(Value);
    end;
  end;  
end;

function TrDBGrid.CanEditModify: Boolean;
var
  OldResult: Boolean;
begin
  Result := inherited CanEditModify;
  OldResult := Result;
  if Assigned(FOnSetCanModify) then FOnSetCanModify(Self, Columns[SelectedIndex], Result);
  if not OldResult and Result then begin
    Datalink.Edit;
    Result := Datalink.Editing;
    if Result then Datalink.Modified;
  end;
end;

procedure TrDBGrid.KeyDown(var Key: Word; Shift: TShiftState);
var
  AddAfter, AChecked: Boolean;
begin
  AddAfter := False;
  AChecked := FRealChecked and (FCheckShift = Shift) and (Key in [VK_DOWN, VK_UP]);
  if AChecked then
  begin
    if not FSelecting then begin
      FSelectionAnchor := TBookmarks(SelectedRows).CurrentRow;
      DoCheck(clmToggle);
      FSelecting := True;
      AddAfter := True;
    end
    else
      with TBookmarks(SelectedRows) do begin
        AddAfter := Compare(CurrentRow, FSelectionAnchor) <> IIF(Key = VK_DOWN, -1, 1);
        if not AddAfter then DoCheck(clmToggle);
      end
  end
  else FSelecting := False;
  inherited;
  if AChecked and AddAfter then DoCheck(clmToggle);
end;

function TrDBGrid.GetCheckFieldValue: Variant;
var
  i: Integer;
begin
  if FCheckFields.Count = 0 then Result := Null
  else if FCheckFields.Count = 1 then Result := TField(FCheckFields[0]).Value
  else begin
    if TField(FCheckFields[0]).IsNull then Result := Null
    else begin
      Result := VarArrayCreate([0, FCheckFields.Count - 1], varVariant);
      for i := 0 to FCheckFields.Count - 1 do Result[i] := TField(FCheckFields[i]).Value;
    end;
  end;
end;

end.

