//==============================================
//      File rdbtreeview.pas
//
//      Components library for Delphi and BCB++.
//      DBAware-tree for TDataSet.
//
//      Copyright 1998-2002 Polaris Software
//      http://polesoft.da.ru
//      mailto: PolarisLib@mail.ru
//==============================================
unit rDBTreeView;

{$I POLARIS.INC}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Menus,
  ComCtrls, DB, DBCtrls, ToolEdit, Stdctrls, Buttons, rUtils, rDBConst, Consts,
  {$IFDEF POLARIS_D4}ImgList,{$ENDIF} rClasses, CommCtrl;

const
  RCT_LOOKUPCHANGE = WM_USER + $2000;
  scCheckKey = scCtrl + vk_Space;

type

  ErDBTreeView = class(Exception);
  TrCustomDBTreeView = class;

  { TrTreeViewDataLink }
  TrTreeViewDataLink = class(TDataLink)
  private
    FControl: TrCustomDBTreeView;
  protected
    procedure ActiveChanged;  override;
    procedure DataSetChanged; override;
    procedure EditingChanged; override;
    procedure RecordChanged(Field: TField); override;
//**
    procedure DataSetScrolled(Distance: Integer); override;
    procedure LayoutChanged; override;
    procedure UpdateData; override;
//**
  end;

  { TTreeDataList }

  PTreeDataEntry = ^TTreeDataEntry;
  TTreeImageIndex = packed record
    ParentIndex,
    ParentSelectedIndex,
    ChildIndex,
    ChildSelectedIndex: Integer;
  end;
  TTreeDataEntry = record
    KeyValue,
    ChildValue,
    ParentValue: Variant;
    Images: TTreeImageIndex;
    Node: TTreeNode;
//    Text: String;
  end;

  TrTreeBuildMode = (trbmDefault, trbmAsIs{, trsmQuickSort});

  TTreeDataList = class(TObject)
  private

    FList:    TList;
    FStrings: TStringList;
    FIndex:   Integer;
    FEmpty,
    FCurrent: TTreeDataEntry;
    FBuildMode: TrTreeBuildMode;

    procedure SetIndex(Item: Integer);
    function  GetTreeEntry(Index: Integer): TTreeDataEntry;
    procedure SetTreeEntry(Index: Integer; AValue: TTreeDataEntry);
    function  GetTreeNode(Index: Integer): TTreeNode;
    procedure SetTreeNode(Index: Integer; Node: TTReeNode);
    function  GetItems(Index: Integer): PTreeDataEntry;
    procedure SetItems(Index: Integer; Item: PTreeDataEntry);
    function  GetCount: Integer;
    function  GetEntryLevel(Index: Integer): Integer;
    function  GetStringLevel(Index: Integer): Integer;
    function  GetDisplayString(Index: Integer): String;
    function  GetDisplayText(Index: Integer): String;
  protected

//    FOnMySort: TStringListSortCompare;

    function Compare(Index1, Index2: Integer): Boolean;
    function CompareEntryData(Index1, Index2: Integer): Boolean;
    function CompareEntry(Value1, Value2: TTreeDataEntry): Boolean;
    function FullCompareEntry(Value1, Value2: TTreeDataEntry): Boolean;
    procedure ChangePosition(OldPosition, NewPosition: Integer);
    function GetEntryIndex(Index: Integer): TTreeImageIndex;
  public
    constructor Create;
    destructor Destroy; override;

    function Add(const KeyValue, ChildValue, ParentValue: Variant;
                  DisplayString: String): PTreeDataEntry;
    function AddEntry(const Entry: TTreeDataEntry;
                  DisplayString: String): PTreeDataEntry;
    procedure Clear;

    procedure EndOfCreate;
    function CalcEntryLevel(Item: TTreeDataEntry): Integer;

    function GetItem(Item: TTreeDataEntry; Delta: Integer; NoChild: Boolean): PTreeDataEntry;

    procedure BuildTreeNode(Tree: TTreeView; SelValue: Variant);

    function IndexOf      (Item: Pointer):        Integer;

    function IndexOfNode  (Node: TTreeNode):      Integer;

    function IndexOfEntry (Item: TTreeDataEntry): Integer;
    function IndexOfKey   (KeyValue: Variant):   Integer;
    function IndexOfParent(ParentValue: Variant): Integer;
    function ChildrenPos  (ChildValue: Variant; var Pos: Integer): Integer;
    function HasChildren  (Item: TTreeDataEntry): Boolean;
    function HasParent    (Item: TTreeDataEntry): Boolean;

    function IsEmptyEntry(Index: Integer): Boolean;

    property Strings: TStringList read FStrings;

    property EntryLevel[Index: Integer]: Integer read GetEntryLevel;
    property StringLevel[Index: Integer]: Integer read GetStringLevel;
    property DisplayString[Index: Integer]: String read GetDisplayString;
    property DisplayText[Index: Integer]: String read GetDisplayText;

    property BuildMode: TrTreeBuildMode read FBuildMode write FBuildMode;
    property List: TList read FList;
    property Index: Integer read FIndex write SetIndex;
    property Count: Integer read GetCount;
    property CurrentEntry: TTreeDataEntry read FCurrent;
    property Items[Index: Integer]: PTreeDataEntry read GetItems write SetItems;
    property EmptyEntry: TTreeDataEntry read FEmpty;
    property TreeEntry[Index: Integer]: TTreeDataEntry read GetTreeEntry write SetTreeEntry;
    property TreeNode[Index: Integer]: TTreeNode read GetTreeNode write SetTreeNode;
  end;

  { TrCustomDBTreeView }

  TrTreeDisableStyle = (tdsDefault, tdsExplorer, tdsGrayText);

  TrTreeGetTextEvent   = procedure(Sender: TObject; DataSet: TDataSet; var Text: String) of object;
  TrTreeKeyValueChanged = procedure(Sender: TObject; Node: TTreeNode) of object;
  TrTreeGetImageIndexesEvent = procedure(Sender: TObject; DataSet: TDataSet; DataEntry: TTreeDataList;
    Entry: PTreeDataEntry) of object;

//  TrTreeGetImageIndexesEvent = procedure(Sender: TObject; DataSet: TDataSet; Node: TTreeNode) of object;

  { TrTreeCheckIndexes }
  TrTreeCheckIndexes = class(TPersistent)
  private
    FOwner: TPersistent;
    FImageIndexes: Array[0..4] of Integer;
    function GetImageIndexs(Index: Integer): Integer;
    procedure SetImageIndexes(Index: Integer; Value: Integer);
  protected
    function GetOwner: TPersistent; override;
  public
    constructor Create(AOwner: TPersistent);
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    function Compare(CheckIndex1, CheckIndex2: TrTreeCheckIndexes): Boolean;
  published
{
UnCheckIndex            0 -  ;
CheckIndex              1 - 
CheckChildIndex         2 -   (  )
InactiveUnCheckIndex    3 -  (       )
InactiveCheckChildIndex 4 -    (    )
}
    property UnCheckIndex: Integer index 0 read GetImageIndexs write SetImageIndexes default -1;
    property CheckIndex: Integer index 2 read GetImageIndexs write SetImageIndexes default -1;
    property CheckChildIndex: Integer index 3 read GetImageIndexs write SetImageIndexes default -1;
    property InactiveUnCheckIndex: Integer index 1 read GetImageIndexs write SetImageIndexes default -1;
    property InactiveCheckChildIndex: Integer index 4 read GetImageIndexs write SetImageIndexes default -1;
  end;

//  TrMultiSelect = (tmsNone, tmsAll, tmsAllNotHaveChild);
  //tmsNone -   
  //tmsAll -  
  //tmsAllNotHaveChild -    
  TrParentOption = (tpoSelectNode, tpoSelectChild);
  //  Check
  //tpoSelectNode -   
  //tpoSelectChild -     
  TrParentOptions = Set of TrParentOption;

  TrTreeMultiSelect = class(TPersistent)
  private
//    FOption: TrMultiSelect;
    FEnabled: Boolean;
    FCheckIndexes: TrTreeCheckIndexes;
    FParentOptions: TrParentOptions;
    FTreeView: TrCustomDBTreeView;
    procedure SetEnabled(Value: Boolean);
    procedure SetParentOptions(Value: TrParentOptions);
    procedure SetCheckIndexes(Value: TrTreeCheckIndexes);
  protected
    function StoreCheckIndexes: Boolean;
  public
    constructor Create(AOwner: TrCustomDBTreeView);
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;    
  published
//    property Option: TrMultiSelect read FOption write SetOption default tmsNone;
    property Enabled: Boolean read FEnabled write SetEnabled default False;
    property ParentOptions: TrParentOptions read FParentOptions write SetParentOptions default [tpoSelectNode];
    property CheckIndexes: TrTreeCheckIndexes read FCheckIndexes write SetCheckIndexes;//stored StoreCheckIndexes;
  end;

  TrDBTreeNodeCheckState = (trncsUnChecked, trncsChecked, trncsCheckChild);
  TrDBTreeNodeCheckStates = set of TrDBTreeNodeCheckState;

  TrDBTreeChangeMode = (trcmChange, trcmToggle);

  TrDBTreeNode = class(TTreeNode)
  private
    FCheckState: TrDBTreeNodeCheckStates;
    procedure SetChecked(Value: Boolean);
    function GetDBTree: TrCustomDBTreeView;
    function GetChecked: Boolean;
    function GetTreeDataEntry: TTreeDataEntry;
    procedure SetTreeDataEntry(Value: TTreeDataEntry);
    function GetMultiSelect: Boolean;
    function GetParentOptions: TrParentOptions;
//    function GetOptions: TrMultiSelect;
  protected
//    procedure DoCheckChange(CheckValue, ChangeAtChild: Boolean); virtual;
    procedure DoCheckChange(CheckValue: Boolean; CheckMode: TrDBTreeChangeMode); virtual;
    procedure SetCheckState(CheckValue: Boolean);
    procedure ChangeParentState; virtual;
    function GetChildState: TrDBTreeNodeCheckStates; virtual;
    function GetCheckImageIndex: Integer; virtual;
    procedure TestMultiSelect;
//    property MultiSelectOptions: TrMultiSelect read GetMultiSelectOptions;
    property MultiSelect: Boolean read GetMultiSelect;
    property ParentOptions: TrParentOptions read GetParentOptions;
//    property Options: TrMultiSelect read GetOptions;
    property DBTree: TrCustomDBTreeView read GetDBTree;
  public
    constructor Create(AOwner: TTreeNodes);
    destructor Destroy; override;
//    procedure ChangeCheckNode(Value: Boolean; ChangeAtChild: Boolean = False);
    procedure ChangeCheckNode(Value: Boolean);
    property CheckState: TrDBTreeNodeCheckStates read FCheckState;
    property Checked: Boolean read GetChecked write SetChecked default False;
    property DataEntry: TTreeDataEntry read GetTreeDataEntry write SetTreeDataEntry;
  published
  end;

  { TrTreeCheckList }
  TTreeCheckListChangingEvent = procedure(Sender: TObject; Item: Variant;
    Action: TVarListAction; var AllowChange{, AllowChangeChild}: Boolean) of object;
  TTreeCheckListChange = procedure(Sender: TObject; Item: Variant;
    Action: TVarListAction) of object;

  TrTreeCheckList = class(TVariantList)
  private
    FTreeView: TrCustomDBTreeView;
    FCheckListChanging: TTreeCheckListChangingEvent;
    FCheckListChange: TTreeCheckListChange;
  protected
  public
    constructor Create(ATree: TrCustomDBTreeView);
    function Add(Item: Variant): Integer; override;
    function Remove(Item: Variant): Integer; override;
    procedure Insert(Index: Integer; Item: Variant); override;
    procedure Delete(Index: Integer); override;
    procedure Clear; override;
    function AllowChange(Item: Variant; Action: TVarListAction): Boolean;
  end;

  TrCustomDBTreeView = class(TCustomTreeView)
  private
    { Private declarations }
    FInternalCheck: Boolean;
    FCheckList: TrTreeCheckList;
    FCheckListChange: TTreeCheckListChange;
    FCheckListChanging: TTreeCheckListChangingEvent;
    FCheckKey: TShortCut;

    FDisableStyle: TrTreeDisableStyle;
    FOnKeyChanged: TrTreeKeyValueChanged;
    FOnGetText: TrTreeGetTextEvent;
    FOnGetImageIndexes: TrTreeGetImageIndexesEvent;
    FValue,
    FChildrenFieldName,
    FParentFieldName: String;

    FIsResync,
    FSelectedEnabled,
    FAutoMove,
    FAutoSelect:      Boolean;
    FLinkData:        TrTreeViewDataLink;
    FKeyValue:        Variant;
    FListFields:      TList;
    FTreeData:        TTreeDataList;
                           {Selected, Parent}
    FImageItedexes:   Array[Boolean,  Boolean] of Integer;
    FPrevState:       TDataSetState;

    FCurrentRecord,
    FRecordIndex,
    FRecordCount:     Integer;
    FSearchString:    String;

    FDestroingWindow,
    FNotSelectParent,
    FIsFirst: Boolean; //new check
    //Multi
    FSelecting: Boolean;
    FSelectionAnchor: TTreeNode;
    Direction: Word;
    FMultiSelect: TrTreeMultiSelect;
    FDefaultCheckImage: TCustomImageList;
    FCheckShift: TShiftState;

    function GetChildIndex: Integer;
    function GetChildSelected: Integer;
    function GetParentIndex: Integer;
    function GetParentSelected: Integer;
    function GetDisplayValue: String;
    procedure SetChildIndex(AValue: Integer);
    procedure SetChildSelected(AValue: Integer);
    procedure SetParentIndex(AValue: Integer);
    procedure SetParentSelected(AValue: Integer);

    procedure SetChildrenFieldName(AValue: String);
    procedure SetParentFieldName  (AValue: String);
    procedure SetValue            (AValue: String); virtual;
    procedure SetKeyValue         (AValue: Variant); virtual;
    procedure SetNotSelectParent    (Value: Boolean);

    function GetBuildMode: TrTreeBuildMode;
    procedure SetBuildMode(Value: TrTreeBuildMode);

    procedure ChangeImages;

    procedure GetImageIndexes(Node: TTreeNode);

{$IFDEF POLARIS_D4}
    procedure SetDisableStyle(Value: TrTreeDisableStyle);
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CNNotify(var Message: TWMNotify); message CN_Notify;
{$ENDIF}

//    procedure TVMGETDISPINFO(var Message: TMessage); message TVM_GETITEMRECT;

    procedure SetMultiSelect(const Value: TrTreeMultiSelect);
    procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
    procedure WMChar(var Message: TWMChar); message WM_CHAR;
    procedure SetCheckListChange(const Value: TTreeCheckListChange);
    procedure SetCheckListChanging(const Value: TTreeCheckListChangingEvent);
  protected
    { Protected declarations }
    FLastSelected: TVariantList;
    FNoCloseUp,
    FOnlyChild: Boolean;

{$IFDEF POLARIS_D4}
    FSaveProc: TTVCustomDrawItemEvent;
{$ENDIF}
    FUpdateEnbled,
    FDataChangedCalled,
    FLookupControl,
    FListActive,
    FHasMoved,
    FRefreshEnabled,
    FHasInsert,
    FHasCreated:   Boolean;
    FKeyField,
    FChildrenField,
    FParentField:  TField;
    FDataSet:      TDataSet;
    FEditNode:     TTreeNode;

    FKeyFieldName, FDisplayFields: String;

    function GetLinkDataSource: TDataSource;
    procedure SetLinkDataSource   (AValue: TDataSource);
    procedure SetKeyFieldName     (AValue: String);
    procedure SetDisplayFields    (AValue: String);

    function GetNodeFromItem(const Item: TTVItem): TTreeNode;
    function TestChildEnabled: Boolean; virtual;
    procedure PreNotify(var Message: TWMNotify); virtual;

    function CanCollapse(Node: TTreeNode): Boolean; override;

    function CreateNode: TTreeNode; override;
    function GetChildNode(Node: TTreeNode; Delta: Integer): TTreeNode;

    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;

    procedure WndProc(var Message: TMessage); override;
    //Multi
    function ChangeCheckNode(Node: TTreeNode): Boolean; virtual;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure AfterCheckChange(Item: Variant; Action: TVarListAction); virtual;
{$IFDEF POLARIS_D4}
    procedure CheckDisableColor(Repaint: Boolean);
    procedure InternalDrawItem(Sender: TCustomTreeView;
      Node: TTreeNode; State: TCustomDrawState; var DefaultDraw: Boolean);
    procedure CustomDrawParam(Node: TTreeNode; State: TCustomDrawState;
      Stage: TCustomDrawStage); virtual;
    function CustomDrawItem(Node: TTreeNode; State: TCustomDrawState;
      Stage: TCustomDrawStage{$IFDEF POLARIS_D5}; var PaintImages: Boolean{$ENDIF}): Boolean; override;
{$ENDIF}

    procedure InheritedReadOnly(AValue: Boolean);
    procedure KeyValueChanged; virtual;
    function  CanEdit(Node: TTreeNode): Boolean; override;
    procedure Change(Node: TTreeNode); override;

    function GetListField(Index: Integer): TField;
    function GetListFieldCount:            Integer;

    procedure CreatedNodes; virtual;
    procedure CreateEntryData; virtual;
    procedure ClearNodes;

    function  CheckActive:  Boolean;
    function  CheckDataSet: Boolean;
    function  CheckSource:  Boolean;
    function  CheckBrowse:  Boolean;
    procedure CheckNotCircular; virtual;
    procedure SyncData(Node: TTreeNode);
    procedure SyncDataSet; virtual;
//new
//    procedure ChangeDisplayNames;
    procedure NewLayout; virtual;
    procedure InitFields(ShowError: Boolean);

    function  DisplayString: String;
    procedure ActiveChanged;                virtual;
    procedure DataChanged;                  virtual;
    procedure LayoutChanged;                virtual;
    procedure EditingChanged;               virtual;
    procedure RecordChanged(Field: TField); virtual;
    procedure DataSetScrolled(Distance: Integer); virtual;

    procedure UpdatingData;                 virtual;
    procedure ReSyncRecords;                virtual;
    procedure ChangeChildOf (Node: TTreeNode; ChildValue: Variant);  virtual; //dynamic;
    procedure ChangeParentOf(Node: TTreeNode; ParentValue: Variant); virtual; //dynamic;
    procedure ChangeDataOf  (Node: TTreeNode; KeyValue: Variant);   virtual; //dynamic;

    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;

    procedure SetValueKey(const AValue: Variant; const AString: String); virtual; //dynamic;
    procedure SelectKeyValue(const AValue: Variant); virtual; //dynamic;
    procedure ChangeNodeText(Node: TTreeNode; const NewText: String);
    procedure SetSearchString(AValue: String);

    procedure DoEnter; override;
    procedure DoExit; override;

    //Multi

    function UpdateRecordCount: Integer; virtual;
    //new
    procedure ScrollTo(Pos: Smallint);
    function ScrollPos: SmallInt;

    function GetItemHeight: Integer;

    { Properties }
    { Internal }
    {R}
    property DisplayValue: String read GetDisplayValue;
    property SearchString: String read FSearchString write SetSearchString;
    {RW}
    property AutoSelectItem:Boolean read FAutoSelect write FAutoSelect default False;
    property ChildIndex:    Integer read GetChildIndex write SetChildIndex default -1;
    property ChildSelected: Integer read GetChildSelected write SetChildSelected default -1;
    property ParentIndex:   Integer read GetParentIndex write SetParentIndex default -1;
    property ParentSelected:Integer read GetParentSelected write SetParentSelected default -1;
    property LinkDataSource:TDataSource read  GetLinkDataSource write SetLinkDataSource;
    property ChildrenField: String read  FChildrenFieldName write SetChildrenFieldName;
    property KeyField:      String read  FKeyFieldName write SetKeyFieldName;
    property ParentField:   String read  FParentFieldName write SetParentFieldName;
    property DisplayFields: String read  FDisplayFields write SetDisplayFields;
    property Items stored False;
    property NotSelectParent: Boolean read FNotSelectParent write SetNotSelectParent default False;

    {Events}
    property OnGetText: TrTreeGetTextEvent read FOnGetText write FOnGetText;
    property OnGetImageIndexes: TrTreeGetImageIndexesEvent read FOnGetImageIndexes write FOnGetImageIndexes;
    property OnKeyChanged: TrTreeKeyValueChanged read FOnKeyChanged write FOnKeyChanged;
{$IFDEF POLARIS_D4}
    property DisableStyle: TrTreeDisableStyle read FDisableStyle write SetDisableStyle default tdsDefault;
{$ENDIF}

    property MultiSelect: TrTreeMultiSelect read FMultiSelect write SetMultiSelect;
    property OnCheckListChange: TTreeCheckListChange read FCheckListChange write SetCheckListChange;
    property OnCheckListChanging: TTreeCheckListChangingEvent read FCheckListChanging write SetCheckListChanging;
    property CheckKey: TShortCut read FCheckKey write FCheckKey default scCheckKey;
    property CheckShift: TShiftState read FCheckShift write FCheckShift default [ssShift];

  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy;                    override;
    procedure RefreshNodes;
    function Locate(const S: String; var Index: Integer; var Node: TTreeNode): Boolean;

{$IFDEF POLARIS_D4}
    function ExecuteAction(Action: TBasicAction): Boolean; override;
    function UpdateAction(Action: TBasicAction): Boolean; override;
{$ENDIF}

    {R}
    property ListFieldCount:Integer  read  GetListFieldCount;
    property ChildFld:      TField   read  FChildrenField;
    property ParentFld:     TField   read  FParentField;
    property ListField[Index: Integer]: TField read  GetListField;
    {RW}
    property BuildMode: TrTreeBuildMode read GetBuildMode write SetBuildMode default trbmDefault;
    property Value: String read FValue write SetValue;
    property KeyValue: Variant read FKeyValue write SetKeyValue;

    property TreeData: TTreeDataList read FTreeData;
    property CheckList: TrTreeCheckList read FCheckList;
    procedure DoCheck;
    procedure CheckAll;
    function GetNodeByID(Item: Variant): TTreeNode;
    function IsMultiSelect: Boolean;
  end;

  TrDBTreeView = class(TrCustomDBTreeView)
  public
    property Items;

  published
    property Align;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property Images;
    property Indent;
    property ParentColor default False;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property RightClickSelect;
    property ShowButtons;
    property ShowHint;
    property ShowLines;
    property ShowRoot;
    property SortType;
    property StateImages;
    property TabOrder;
    property TabStop default True;
    property Visible;
{$IFDEF POLARIS_D5}
    property OnAdvancedCustomDraw;
    property OnAdvancedCustomDrawItem;
{$ENDIF}
    property OnChange;
    property OnChanging;
    property OnClick;
    property OnCollapsing;
    property OnCollapsed;
    property OnCompare;
    property OnDblClick;
    property OnDeletion;
    property OnDragDrop;
    property OnDragOver;
    property OnEdited;
    property OnEditing;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnExpanding;
    property OnExpanded;
    property OnGetImageIndex;
    property OnGetSelectedIndex;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    {$IFDEF POLARIS_D4}
    property Anchors;
    property AutoExpand;
    property BiDiMode;
    property BorderWidth;
    property ChangeDelay;
    property Constraints;
    property DragKind;
    property HotTrack;
    property ParentBiDiMode;
    property RowSelect;
    property ToolTips;
    property OnCustomDraw;
    property OnCustomDrawItem;
    property OnEndDock;
    property OnStartDock;
    property DisableStyle;
    {$ENDIF}
    { New Property }
    property BuildMode;
    property NotSelectParent;
    property AutoSelectItem;
    property ChildIndex;
    property ChildSelected;
    property ParentIndex;
    property ParentSelected;
    property ChildrenField;
    property KeyField;
    property ParentField;
    property DisplayFields;
    property DataSource:TDataSource read  GetLinkDataSource
                                    write SetLinkDataSource;
    property OnGetText;
    property OnGetImageIndexes;
    property OnKeyChanged;

    property MultiSelect;
    property OnCheckListChange;
    property OnCheckListChanging;
    property CheckKey;
    property CheckShift;
  end;

  { TrDBLookupTree }
  TrDBLookupTree = class;

  TrDataSourceLink = class(TDataLink)
  private
    FControl: TrDBLookupTree;
  protected
    procedure ActiveChanged; override;
    procedure LayoutChanged; override;
    procedure FocusControl(Field: TFieldRef); override;
    procedure DataSetChanged; override;
    procedure RecordChanged(Field: TField); override;
{???}
    procedure EditingChanged; override;
  end;

  TrDBLookupTree = class(TrCustomDBTreeView)
  private
    FReadOnly: Boolean;
    FDataLink: TrDataSourceLink;
    FDataFieldName: String;
    FDataField: TField;
    FMasterField: TField;

    procedure SetReadOnly(AValue: Boolean);
    procedure SetDataFieldName(AValue: String);
    function GetDataSource: TDataSource;
    procedure SetDataSource(AValue: TDataSource);

    procedure DataLinkActiveChanged; virtual;
    procedure DataLinkDataSetChanged; virtual;
    procedure CheckDataLinkActiveChanged; virtual;
    procedure DataLinkRecordChanged(Field: TField); virtual;
    {???}
    procedure DataLinkEditChanged; virtual;

  protected
    FHasEdit: Boolean;
    FPrevValue: Variant;

    procedure Notification(AComponent: TComponent;
                           Operation: TOperation); override;
    procedure CheckNotCircular; override;
    procedure SetKeyValue         (AValue: Variant); override;
    procedure SelectKeyValue(const AValue: Variant); override;
    function CanModify: Boolean;

    procedure Change(Node: TTreeNode); override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Items;
  published
    property Align;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property Images;
    property Indent;
    property ParentColor default False;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly: Boolean read FReadOnly write SetReadOnly;
    property RightClickSelect;
    property ShowButtons;
    property ShowHint;
    property ShowLines;
    property ShowRoot;
    property SortType;
    property StateImages;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property OnChange;
    property OnChanging;
    property OnClick;
    property OnCollapsing;
    property OnCollapsed;
    property OnCompare;
    property OnDblClick;
    property OnDeletion;
    property OnDragDrop;
    property OnDragOver;
    property OnEdited;
    property OnEditing;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnExpanding;
    property OnExpanded;
    property OnGetImageIndex;
    property OnGetSelectedIndex;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
    {$IFDEF POLARIS_D4}
    property Anchors;
    property AutoExpand;
    property BiDiMode;
    property BorderWidth;
    property ChangeDelay;
    property Constraints;
    property DragKind;
    property HotTrack;
    property ParentBiDiMode;
    property RowSelect;
    property ToolTips;
    property OnCustomDraw;
    property OnCustomDrawItem;
    property OnEndDock;
    property OnStartDock;
    property DisableStyle;
    {$ENDIF}
    { New property }
    property BuildMode;
    property DataField: String read FDataFieldName write SetDataFieldName;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property LookupSource: TDataSource read GetLinkDataSource write SetLinkDataSource;
    property LookupField: String read FKeyFieldName write SetKeyFieldName;
    property LookupDisplay: String read FDisplayFields write SetDisplayFields;
    property ChildrenField;
    property ParentField;
    property ChildIndex;
    property ChildSelected;
    property ParentIndex;
    property ParentSelected;
    property OnGetText;
    property OnGetImageIndexes;
    property OnKeyChanged;

    property MultiSelect;
    property OnCheckListChange;
    property OnCheckListChanging;
    property CheckKey;
    property CheckShift;
  end;

  TrDBLookupComboTree = class;

  TrPopupTree = class(TPopupWindow)
  private
//    FTreeList: TrDBLookupTree;
    FIsCreateTree: Boolean;
    FTreeList: TrCustomDBTreeView;
    FNoneLabel: TStaticText;
    FDropDownRows: Integer;
    SNoneLabel: String;
    FPrevValue,
    FIValue: Variant;

    procedure SetNoneLabel(AValue: String);

    procedure WMMouseActivate(var Message: TMessage); message WM_MOUSEACTIVATE;
    procedure WMKeyDown(var Message: TWMKeyDown); message WM_KEYDOWN;
    procedure RCTLookupChange(var Message: TMessage); message RCT_LOOKUPCHANGE;

    procedure SetDropDownRows(AValue: Integer);
    procedure TreeMouseUp(Sender: TObject; Button: TMouseButton;
                          Shift: TShiftState; X, Y: Integer);
    procedure LabelClick(Sender: TObject);
    procedure TreeChange(Sender: TObject; Node: TTreeNode);
  protected
//    procedure Click; override;

    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure DestroyWnd; override;

    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    function GetValue: Variant; override;
    procedure SetValue(const Value: Variant); override;

    function GetDisplayValue: String;
    function GetSearchString: String;
    procedure SetSearchString(AValue: String);

    procedure EndCreated;

    property DisplayValue: String read GetDisplayValue;
    property SearchString: String read GetSearchString write SetSearchString;

    property NoneLabel: String read SNoneLabel write SetNoneLabel;

    procedure ProcessSearchKey(Key: Char);
  public
    constructor Create(AOwner: TComponent); override;
    function GetPopupText: String; override;

//    procedure Hide; override;
    procedure Show(Origin: TPoint); override;

    property DropDownRows: Integer read FDropDownRows write SetDropDownRows;
  end;

  { TrDBLookupComboTree }
  TrDBLookupComboTree = class(TCustomComboEdit)
  private
// Design time only
    FDTBuildMode:        TrTreeBuildMode;
    FDTNotSelectParent:    Boolean;
    FDTDataSource:     TDataSource;
    FDTLookupField:    String;
    FDTLookupDisplay:  String;
    FDTChildrenField:  String;
    FDTParentField:    String;
    FDTChildIndex:     Integer;
    FDTChildSelected:  Integer;
    FDTParentIndex:    Integer;
    FDTParentSelected: Integer;
    FDTImages:         TImageList;
    FDTDropDownCount:  Integer;
// ****************
    FValue:         String;
    FKeyValue:      Variant;
    FDisplayString: String;
    FDataLink:      TFieldDataLink;
//    FCharPress,
    FEscapeClear,
    FOnDataChange,
    FChanging,
    FListVisible:   Boolean;
//    FDropDownCount: Integer;
    FDropDownWidth: Integer;
    FOnDropDown:    TNotifyEvent;
    FOnCloseUp:     TNotifyEvent;
//    FInListDown:    Boolean;
    FOnGetImageIndex: TTVExpandedEvent;
    FOnGetSelectedIndex: TTVExpandedEvent;

    FCharKeyDown: Boolean;

    SNoneLabel: String;

    FMultiSelect: TrTreeMultiSelect;
    FCheckKey: TShortCut;
    FCheckShift: TShiftState;
    FCheckListChange: TTreeCheckListChange;
    FCheckListChanging: TTreeCheckListChangingEvent;

    procedure SetNoneLabel(AValue: String);
    procedure SetOnGetImageIndex(AValue: TTVExpandedEvent);
    procedure SetOnGetSelectedIndex(AValue: TTVExpandedEvent);

    procedure ActiveChange(Sender: TObject);
    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
    procedure UpdateData(Sender: TObject);
    procedure KeyChanged(Sender: TObject; Node: TTreeNode);
//    procedure PopupCloseUp(Sender: TObject; Accept: Boolean);

    procedure SetValue(AValue: String);
    procedure SetKeyValue(AValue: Variant);

    procedure SetDropDownWidth(AValue: Integer);
    function GetDropDownCount: Integer;
    procedure SetDropDownCount(AValue: Integer);
    function GetDataField: String;
    procedure SetDataField(AValue: String);
    function GetDataSource: TDataSource;
    procedure SetDataSource(AValue: TDataSource);
    function GetLinkDataSource: TDataSource;
    procedure SetLinkDataSource(AValue: TDataSource);
    function GetKeyField: String;
    procedure SetKeyField(AValue: String);
    function GetDisplayFields: String;
    procedure SetDisplayFields(AValue: String);
    function GetChildrenField: String;
    procedure SetChildrenField(AValue: String);
    function GetParentField: String;
    procedure SetParentField(AValue: String);
    procedure SetReadOnly(AValue: Boolean);

    function GetChildIndex: Integer;
    procedure SetChildIndex(AValue: Integer);
    function GetChildSelected: Integer;
    procedure SetChildSelected(AValue: Integer);
    function GetParentIndex: Integer;
    procedure SetParentIndex(AValue: Integer);
    function GetParentSelected: Integer;
    procedure SetParentSelected(AValue: Integer);

    function GetImageList: TImageList;
    procedure SetImageList(AValue: TImageList);
    function GetBuildMode: TrTreeBuildMode;
    procedure SetBuildMode(Value: TrTreeBuildMode);
    function GetNotSelectParent: Boolean;
    procedure SetNotSelectParent(Value: Boolean);

    {    procedure InvalidateText;}
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMKeyDown(var Message: TWMKeyDown); message WM_KEYDOWN;
    procedure WMChar(var Message: TWMChar); message WM_CHAR;
    procedure CMGetDataLink(var Message: TMessage); message CM_GETDATALINK;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;

    procedure RCTLookupChange(var Message: TMessage); message RCT_LOOKUPCHANGE;

    procedure SetMultiSelect(const Value: TrTreeMultiSelect);
    function GetCheckList: TrTreeCheckList;
    procedure SetCheckListChange(const Value: TTreeCheckListChange);
    procedure SetCheckListChanging(const Value: TTreeCheckListChangingEvent);
    function GetDisplayString: String;
    procedure SetCheckKey(const Value: TShortCut);
    procedure SetCheckShift(const Value: TShiftState);
    function GetIDText: String;
    procedure SetIDText(const Value: String);

  protected

    procedure CheckNotCircular; virtual;

    procedure DoEnter; override;
    procedure DoExit; override;

    procedure ResetField;

    function GetReadOnly: Boolean; override;
//    function EditCanModify: Boolean; override;
    function CanModify: Boolean;

    procedure Notification(AComponent: TComponent;
              Operation: TOperation); override;

    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
              X, Y: Integer); override;

    procedure ButtonClick; override;
    procedure Change; override;
    procedure DoChange; override;
    procedure PopupChange; override;
    procedure PopupDropDown(DisableEdit: Boolean); override;
    procedure PopupCloseUp(Sender: TObject; Accept: Boolean); override;
    procedure AcceptValue(const Value: Variant); override;
    function AcceptPopup(var Value: Variant): Boolean; override;
    procedure SetPopupValue(const Value: Variant); override;
    function GetPopupValue: Variant; override;

    procedure CreateWnd; override;
    property DisplayString: String read GetDisplayString;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure RefreshNodes;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
{$IFDEF POLARIS_D4}
    function ExecuteAction(Action: TBasicAction): Boolean; override;
    function UpdateAction(Action: TBasicAction): Boolean; override;
{$ENDIF}
    procedure DropDown;

    property KeyValue: Variant read FKeyValue write SetKeyValue;
    property Value: String read FValue write SetValue;
    property Text;
    property GlyphKind;
    { Ensure GlyphKind is published before Glyph and ButtonWidth }
    property Glyph;
    property ButtonWidth;
    property CheckList: TrTreeCheckList read GetCheckList;
    property IDText: String read GetIDText write SetIDText;
  published
    property NotSelectParent: Boolean read GetNotSelectParent write SetNotSelectParent default False;
    property BuildMode: TrTreeBuildMode read GetBuildMode write SetBuildMode default trbmDefault;

    property DropDownWidth: Integer read FDropDownWidth write SetDropDownWidth default 0;
    property DropDownCount: Integer read GetDropDownCount write SetDropDownCount default 8;
    property OnDropDown: TNotifyEvent read FOnDropDown write FOnDropDown;
    property OnCloseUp:  TNotifyEvent read FOnCloseUp write FOnCloseUp;

    property DataField: String read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property LookupSource: TDataSource read GetLinkDataSource write SetLinkDataSource;
    property LookupField: String read GetKeyField write SetKeyField;
    property LookupDisplay: String read GetDisplayFields write SetDisplayFields;
    property ChildrenField: String read GetChildrenField write SetChildrenField;
    property ParentField: String read GetParentField write SetParentField;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;

    property ChildIndex: Integer read GetChildIndex write SetChildIndex default -1;
    property ChildSelected: Integer read GetChildSelected write SetChildSelected default -1;
    property ParentIndex: Integer read GetParentIndex write SetParentIndex default -1;
    property ParentSelected: Integer read GetParentSelected write SetParentSelected default -1;
    property Images: TImageList read GetImageList write SetImageList;

    property DisplayEmpty: String read SNoneLabel write SetNoneLabel;
    property EscapeClear: Boolean read FEscapeClear write FEscapeClear default True;

    property Alignment;
    property BorderStyle;
    property ButtonHint;
    property CharCase;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
{$IFDEF POLARIS_D4}
    property Anchors;
    property BiDiMode;
    property Constraints;
    property DragKind;
    property ParentBiDiMode;
{$ENDIF}
{$IFNDEF VER90}
    property ImeMode;
    property ImeName;
{$ENDIF}
//    property NumGlyphs;
    property OEMConvert;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnButtonClick;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;

    property OnGetImageIndex: TTVExpandedEvent read FOnGetImageIndex
                                               write SetOnGetImageIndex;
    property OnGetSelectedIndex: TTVExpandedEvent read FOnGetSelectedIndex
                                                  write SetOnGetSelectedIndex;

    property OnStartDrag;
{$IFDEF POLARIS_D4}
    property OnEndDock;
    property OnStartDock;
{$ENDIF}
    property MultiSelect: TrTreeMultiSelect read FMultiSelect write SetMultiSelect;
    property OnCheckListChange: TTreeCheckListChange read FCheckListChange write SetCheckListChange;
    property OnCheckListChanging: TTreeCheckListChangingEvent read FCheckListChanging write SetCheckListChanging;
    property CheckKey: TShortCut read FCheckKey write SetCheckKey default scCheckKey;
    property CheckShift: TShiftState read FCheckShift write SetCheckShift default [ssShift];
  end;

  procedure SetNodeState(node :TTreeNode; Flags: Integer);

implementation

uses DBConsts, DBUtils, MaxMin, extctrls, rConst,
     {$IFDEF POLARIS_D6}RxStrUtils, Variants{$ELSE}StrUtils{$ENDIF};

{$IFNDEF POLARIS_D4}
type
  TCrackCustomImageList = class(TCustomImageList);
{$ENDIF}

procedure DrawDefaultCheck(Image: {$IFNDEF POLARIS_D4}TCrackCustomImageList{$ELSE}TCustomImageList{$ENDIF}; Height: Integer);
const
  pfCount = 4;
  PaintFlags : Array [0..pfCount] of Integer = (DFCS_BUTTONCHECK or DFCS_FLAT,
                                     DFCS_BUTTONCHECK or DFCS_FLAT or DFCS_CHECKED,
                                     DFCS_BUTTONCHECK or DFCS_FLAT or DFCS_CHECKED or DFCS_INACTIVE,
                                     DFCS_BUTTONCHECK or DFCS_FLAT or DFCS_INACTIVE,
//                                     DFCS_BUTTONCHECK or DFCS_FLAT or DFCS_CHECKED or DFCS_PUSHED or DFCS_INACTIVE
                                     DFCS_BUTTONCHECK or DFCS_FLAT or DFCS_CHECKED or DFCS_PUSHED{ or DFCS_INACTIVE}
                                     );
{
UnCheckIndex            0 -  ;
CheckIndex              1 - 
CheckChildIndex         2 -   (  )
InactiveUnCheckIndex    3 -  (       )
InactiveCheckChildIndex 4 -    (    )
}
var
  B: TBitmap;
  I: Integer;
  iR, dR, R: TRect;
  mColor: TColor;
begin
  if Assigned(Image) then begin
    B := TBitmap.Create;
    try
      if Height > 0
      then B.Height := Height
      else B.Height := 16;
      B.Width := B.Height;
      Image.Clear;
      Image.Height := B.Height;
      Image.Width  := B.Width;
      R := Rect(0,0,B.Width,B.Height);
      iR.Left   := B.Width div 2-6;
      iR.Top    := B.Width div 2-6;
      iR.Right  := iR.Left+13;
      iR.Bottom := iR.Top+13;
      for I := -1 to pfCount do begin
        SelectClipRgn(B.Canvas.Handle, 0);
        IntersectClipRect(B.Canvas.Handle, R.Left, R.Top, R.Right, R.Bottom);
        if I = 2
        then mColor := clBtnFace
        else mColor := clFuchsia;

        B.Canvas.Brush.Color := mColor;
        B.Canvas.Brush.Style := bsSolid;
        B.Canvas.FillRect(R);
        if (I >= 0) and ( I<= pfCount) then begin
          dR := iR;
          InflateRect(dR, -2, -2);
          SelectClipRgn(B.Canvas.Handle, 0);
          IntersectClipRect(B.Canvas.Handle, dR.Left, dR.Top, dR.Right, dR.Bottom);
          DrawFrameControl(B.Canvas.Handle, iR, DFC_BUTTON, PaintFlags[I]);
          InflateRect(dR, 1, 1);
          SelectClipRgn(B.Canvas.Handle, 0);
//          IntersectClipRect(B.Canvas.Handle, dR.Left, dR.Top, dR.Right, dR.Bottom);
          B.Canvas.Pen.Color := clBtnShadow;
          B.Canvas.Brush.Style := bsClear;
{$IFDEF POLARIS_D5}
          B.Canvas.Rectangle(dR);
{$ELSE}
          B.Canvas.Rectangle(dR.Left, dR.Top, dR.Right, dR.Bottom);
{$ENDIF}
        end;
        Image.AddMasked(B, mColor);
      end;
    finally
      B.Free;
    end;
  end;
end;

function TestKeyShiftState(Shift, TestShift: TShiftState): Boolean;
var
  B: Boolean;
begin
  B := False;
  if ssShift in TestShift then B := B or (ssShift in Shift);
  if ssAlt in TestShift then B := B or (ssAlt in Shift);
  if ssCtrl in TestShift then B := B or (ssCtrl in Shift);
  if ssLeft in TestShift then B := B or (ssLeft in Shift);
  if ssRight in TestShift then B := B or (ssRight in Shift);
  if ssMiddle in TestShift then B := B or (ssMiddle in Shift);
  if ssDouble in TestShift then B := B or (ssDOuble in Shift);
  Result := B;
end;

procedure SetNodeState(node :TTreeNode; Flags: Integer);
var
  tvi: TTVItem;
begin
//RSV new
  FillChar(tvi, Sizeof(tvi), 0);
  tvi.hItem := node.ItemID;
  tvi.mask := TVIF_STATE;
  tvi.stateMask := TVIS_BOLD or TVIS_CUT;
  tvi.state := Flags;
  TreeView_SetItem(node.Handle, tvi);
end;

procedure FontSetDefault(Font: TFont);
begin
  Font.Name := 'MS Sans Serif';
  Font.Size := 8;
  Font.Style := [];
  Font.Color := clWindowText;
end;

{ TrTreeCheckIndexes }

constructor TrTreeCheckIndexes.Create(AOwner: TPersistent);
begin
  inherited Create;
  FOwner := AOwner;
  FillChar(FImageIndexes, SizeOf(FImageIndexes), -1);
end;

destructor TrTreeCheckIndexes.Destroy;
begin
  FOwner := nil;
  inherited Destroy;
end;

procedure TrTreeCheckIndexes.Assign(Source: TPersistent);
begin
  if Assigned(Source) and (Source is TrTreeCheckIndexes)
  then begin
    Move(TrTreeCheckIndexes(Source).FImageIndexes, FImageIndexes, SizeOf(FImageIndexes));
    FOwner := TrTreeCheckIndexes(Source).FOwner;
  end else inherited Assign(Source);
end;

function TrTreeCheckIndexes.GetOwner: TPersistent;
begin
  Result := FOwner;
end;

function TrTreeCheckIndexes.GetImageIndexs(Index: Integer): Integer;
begin
  Result := -1;
  if Index in [0..4]
  then Result := FImageIndexes[Index]
  else Exception.CreateFmt(SOutOfRange, [0, 4]);
end;

procedure TrTreeCheckIndexes.SetImageIndexes(Index: Integer; Value: Integer);
begin
  if Index in [0..4]
  then begin
    if FImageIndexes[Index] <> Value then begin
      FImageIndexes[Index] := Value;
      //!!!!     -      
      // 
    end
  end else Exception.CreateFmt(SOutOfRange, [0, 4]);
end;

function TrTreeCheckIndexes.Compare(CheckIndex1, CheckIndex2: TrTreeCheckIndexes): Boolean;
var
  I: Integer;
begin
  Result := Assigned(CheckIndex1) and Assigned(CheckIndex2);
  for I := 0 to 4 do
    if Result
    then Result := Result and (CheckIndex1.FImageIndexes[I] = CheckIndex2.FImageIndexes[I])
    else Break;
end;

{TrTreeMultiSelect}

constructor TrTreeMultiSelect.Create(AOwner: TrCustomDBTreeView);
begin
  inherited Create;
  FTreeView := AOwner;
  FEnabled := False;
  FParentOptions := [tpoSelectNode];
  FCheckIndexes := TrTreeCheckIndexes.Create(Self);
end;

destructor TrTreeMultiSelect.Destroy;
begin
  FCheckIndexes.Free;
  FCheckIndexes := nil;
  FTreeView := nil;
  inherited Destroy;
end;

procedure TrTreeMultiSelect.SetEnabled(Value: Boolean);
begin
  if FEnabled <> Value then begin
    FEnabled := Value;
    if Assigned(FTreeView) and FTreeView.HandleAllocated then
      with FTreeView do begin
        if FEnabled then
          TreeView_SetImageList(Handle, FDefaultCheckImage.Handle, TVSIL_STATE)
        else TreeView_SetImageList(Handle, 0, TVSIL_STATE);
        CreatedNodes;
      end;
  end;
end;

procedure TrTreeMultiSelect.SetParentOptions(Value: TrParentOptions);
begin
  if FParentOptions <> Value then begin
    FParentOptions := Value;
    //!!!!     
    if Assigned(FTreeView)
    then FTreeView.Invalidate;
  end;
end;

procedure TrTreeMultiSelect.SetCheckIndexes(Value: TrTreeCheckIndexes);
begin
  FCheckIndexes.Assign(Value);
end;

function TrTreeMultiSelect.StoreCheckIndexes: Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to 4 do
    Result := Result or (FCheckIndexes.FImageIndexes[I] <> -1);
end;

procedure TrTreeMultiSelect.Assign(Source: TPersistent);
begin
  if Assigned(Source) and (Source is TrTreeMultiSelect)
  then begin
    Enabled := TrTreeMultiSelect(Source).Enabled;
    ParentOptions := TrTreeMultiSelect(Source).ParentOptions;
  end else inherited Assign(Source);
end;

{ TrDBTreeNode }

constructor TrDBTreeNode.Create(AOwner: TTreeNodes);
begin
  inherited Create(AOwner);
  FCheckState    := [];
  TestMultiSelect;
end;

destructor TrDBTreeNode.Destroy;
begin
  inherited Destroy;
end;

procedure TrDBTreeNode.SetChecked(Value: Boolean);
begin
  {if GetChecked <> Value
  then} DoCheckChange(Value, trcmToggle);
end;

function TrDBTreeNode.GetDBTree: TrCustomDBTreeView;
begin
  Result := TrCustomDBTreeView(TreeView);
end;

//procedure TrDBTreeNode.DoCheckChange(CheckValue, ChangeAtChild: Boolean);
procedure TrDBTreeNode.DoCheckChange(CheckValue: Boolean; CheckMode: TrDBTreeChangeMode);
var
  AllowChangeChecked: Boolean;
  Node: TrDBTreeNode;
begin
//trncsUnChecked, trncsChecked
  AllowChangeChecked := True;
// BeforeCheckChange
//                             node  dataset  var         var              var
//  DBTree.DoBeforeCheckChange(Self, DataSet, CheckValue, CheckAtChildren, AllowChangeChecked);
//   AllowChangeChecked  FChecked   StateIndex   
//        Parent    StateIndex
  if AllowChangeChecked then begin
    //   tmsAllNotHaveChild        
    //  
    if HasChildren and (ParentOptions * [tpoSelectChild] = [tpoSelectChild])
    //tmsAllNotHaveChild
    then begin
      Node := TrDBTreeNode(getFirstChild);
      while Node <> nil do begin
        // check 
        Node.DoCheckChange(CheckValue, trcmChange);
        Node := TrDBTreeNode(Node.getNextSibling);
      end;
    end;
    SetCheckState(CheckValue);
  end;
// AfterChackChane
//  DBTree.DoafterCheckChange(Self, DataSet);
end;

procedure TrDBTreeNode.SetCheckState(CheckValue: Boolean);
begin
//  TrDBTreeNodeCheckState = (trncsUnChecked, trncsChecked, trncsCheckChild);
//tmsAllNotHaveChild
  DBTree.FInternalCheck := True;
  try
    CheckValue := CheckValue and
    (   (HasChildren and (   (ParentOptions * [tpoSelectNode, tpoSelectChild] = [tpoSelectNode]))
                          or (ParentOptions * [tpoSelectNode] = [tpoSelectNode]))
     or (not HasChildren and (ParentOptions * [tpoSelectNode, tpoSelectChild] <> [])));
    if CheckValue then
      if not (trncsChecked in FCheckState) then
      begin
        if DBTree.CheckList.Add(DataEntry.KeyValue) = -1 then Exit;
        FCheckState := FCheckState + [trncsChecked] - [trncsUnChecked]
      end
      else
    else
      if not (trncsUnChecked in FCheckState) then
      begin
        if DBTree.CheckList.Remove(DataEntry.KeyValue) = -1 then Exit;
        FCheckState := FCheckState - [trncsChecked] + [trncsUnChecked];
      end;
    //DBTree.ActiveChanged;
    {if (DBTree.IsMultiSelect) and (DBTree.Owner <> nil) and
      (DBTree.Owner.Owner <> nil) and (DBTree.Owner.Owner is TrDBLookupComboTree)
    then
      with TrDBLookupComboTree(DBTree.Owner.Owner) do begin
        Text := DisplayString;
        SelectAll;
      end;}

  //  NewState := GetChildState;

    if GetChildState = []
    then Exclude(FCheckState, trncsCheckChild)
    else Include(FCheckState, trncsCheckChild);
    ChangeParentState;
  finally
    DBTree.FInternalCheck := False;
  end;
end;

procedure TrDBTreeNode.ChangeParentState;
var
  Node: TrDBTreeNode;
  NewState: TrDBTreeNodeCheckStates;
begin
  Node := TrDBTreeNode(Parent);
  while Assigned(Node) do begin
    NewState := Node.GetChildState;
    if NewState = []
    then Exclude(Node.FCheckState, trncsCheckChild)
    else Include(Node.FCheckState, trncsCheckChild);
    Node := TrDBTreeNode(Node.Parent);
  end;
end;

function TrDBTreeNode.GetChildState: TrDBTreeNodeCheckStates;
var
  Node: TrDBTreeNode;
begin
  Result := [];
  Node := TrDBTreeNode(getFirstChild);
  while Assigned(Node) do begin
    if Node.FCheckState * [trncsChecked, trncsCheckChild] <> []
    then Include(Result, trncsCheckChild);
    Node := TrDBTreeNode(Node.getNextSibling);
  end;
end;

//procedure TrDBTreeNode.ChangeCheckNode(Value: Boolean; ChangeAtChild: Boolean = False);
procedure TrDBTreeNode.ChangeCheckNode(Value: Boolean{; ChangeAtChild: Boolean = False});
begin
  DoCheckChange(Value, trcmToggle);
end;

function TrDBTreeNode.GetCheckImageIndex: Integer;
begin
  Result := -1;
  if DBTree.IsMultiSelect
  then begin
    {if (Data <> nil) and (trncsUnChecked in FCheckState) and
      (DBTree.CheckList.IndexOf(DataEntry.KeyValue) > -1)
    then begin
      FCheckState := FCheckState + [trncsChecked] - [trncsUnChecked];
      ChangeParentState;
      DBTree.Invalidate;
    end;}
    if trncsChecked in FCheckState
    then Result := 1
    else if trncsCheckChild in FCheckState
      then Result := 2
      else if trncsUnChecked in FCheckState
        then Result := 0;
//    if DBTree.FMultiSelect.Option = tmsAllNotHaveChild
//tpoSelectNode, tpoSelectChild
    if ParentOptions * [tpoSelectNode, tpoSelectChild] = [tpoSelectChild]
    then case Result of
      0: if HasChildren then Result := 3;
      2: Result := 4;
    end;
  end;
end;

function TrDBTreeNode.GetChecked: Boolean;
begin
  Result := trncsChecked in FCheckState;
end;

procedure TrDBTreeNode.TestMultiSelect;
begin
  if DBTree.IsMultiSelect
  then if FCheckState = [] then FCheckState := [trncsUnChecked];
end;

function TrDBTreeNode.GetTreeDataEntry: TTreeDataEntry;
begin
  Result := PTreeDataEntry(Data)^;
end;

procedure TrDBTreeNode.SetTreeDataEntry(Value: TTreeDataEntry);
begin
  if Data <> @Value
  then Data := PTreeDataEntry(@Value);
end;

function TrDBTreeNode.GetMultiSelect: Boolean;
begin
  if Assigned(DBTree) and Assigned(DBTree.FMultiSelect)
  then Result := DBTree.FMultiSelect.FEnabled
  else Result := False;
end;

function TrDBTreeNode.GetParentOptions: TrParentOptions;
begin
  Result := [];
  if Assigned(DBTree) and Assigned(DBTree.FMultiSelect)
  then Result := DBTree.FMultiSelect.FParentOptions;
end;

{function TrDBTreeNode.GetOptions: TrMultiSelect;
begin
  Result := tmsNone;
  if Assigned(DBTree) and Assigned(DBTree.FMultiSelect)
  then Result := DBTree.FMultiSelect.FOption;
end;}

{ TTreeDataList }

constructor TTreeDataList.Create;
begin
  inherited Create;
  FList   := TList.Create;
  FStrings := TStringList.Create;
  FIndex := -1;
  FBuildMode := {trsmQuickSort}
//    trbmAsIs
    trbmDefault
  ;
  with FEmpty do begin
    VarClear(KeyValue);
    VarClear(ChildValue);
    VarClear(ParentValue);
    Node := nil;
  end;
end;

destructor TTreeDataList.Destroy;
var
  I: Integer;
  P: PTreeDataEntry;
begin
  for I := 0 to FList.Count-1 do begin
    P := PTreeDataEntry(FList.Items[I]);
    Dispose(P);
  end;
  FList.Free;
  FStrings.Free;
  inherited Destroy;
end;

function TTreeDataList.Compare(Index1, Index2: Integer): Boolean;
begin
  Result := FullCompareEntry(TreeEntry[Index1], TreeEntry[Index2])
end;

function TTreeDataList.CompareEntryData(Index1, Index2: Integer): Boolean;
begin
  Result := CompareEntry(TreeEntry[Index1], TreeEntry[Index2])
end;

function TTreeDataList.CompareEntry(Value1, Value2: TTreeDataEntry): Boolean;
begin
  Result := (CompareVar(Value1.KeyValue, Value2.KeyValue)=0) and
            (CompareVar(Value1.ChildValue, Value2.ChildValue)=0) and
            (CompareVar(Value1.ParentValue, Value2.ParentValue)=0);
end;

function TTreeDataList.FullCompareEntry(Value1, Value2: TTreeDataEntry): Boolean;
begin
  Result := CompareEntry(Value1, Value2) and
            (Value1.Node <> Value2.Node);
end;

procedure TTreeDataList.ChangePosition(OldPosition, NewPosition: Integer);
begin
  FList.Move(OldPosition, NewPosition);
end;

procedure TTreeDataList.SetIndex(Item: Integer);
begin
  if Item <> FIndex then
    if (Item >= 0) and (Item < FList.Count)
    then begin
      FIndex := Item;
      FCurrent := TreeEntry[Item];
    end
    else begin
      FIndex := -1;
      FCurrent := FEmpty;
    end
end;

function TTreeDataList.GetTreeEntry(Index: Integer): TTreeDataEntry;
begin
  Result := FEmpty;
  if (Index >= 0) and (Index < Count)
  then Result := TTreeDataEntry(FList.Items[Index]^);
end;

procedure TTreeDataList.SetTreeEntry(Index: Integer; AValue: TTreeDataEntry);
begin
  if not FullCompareEntry(AValue, TTreeDataEntry(FList.Items[Index]^))
  then with PTreeDataEntry(FList.Items[Index])^ do begin
    KeyValue  := AValue.KeyValue;
    ChildValue := AValue.ChildValue;
    ParentValue:= AValue.ParentValue;
    Node       := AValue.Node;
  end;
end;

function TTreeDataList.GetTreeNode(Index: Integer): TTreeNode;
begin
  Result := TreeEntry[Index].Node;
end;

procedure TTreeDataList.SetTreeNode(Index: Integer; Node: TTReeNode);
begin
  if PTreeDataEntry(FList.Items[Index])^.Node <> Node
  then PTreeDataEntry(FList.Items[Index])^.Node := Node;
end;

function TTreeDataList.GetItems(Index: Integer): PTreeDataEntry;
begin
  Result := nil;
  if (Index >= 0) and (Index < Count)
  then Result := PTreeDataEntry(FList.Items[Index]);
end;

procedure TTreeDataList.SetItems(Index: Integer; Item: PTreeDataEntry);
begin
  FList.Items[Index] := Item;
end;

function TTreeDataList.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TTreeDataList.Add(const KeyValue, ChildValue, ParentValue: Variant;
                  DisplayString: String): PTreeDataEntry;
begin
  New(Result);
  FillChar(Result^, SizeOf(TTreeDataEntry),#0);
  Result^.KeyValue   := KeyValue;
  Result^.ChildValue  := ChildValue;
  Result^.ParentValue := ParentValue;
  FStrings.AddObject(DisplayString, TObject(Result));
  FList.Add(Result);
end;

function TTreeDataList.AddEntry(const Entry: TTreeDataEntry;
                  DisplayString: String): PTreeDataEntry;
begin
  New(Result);
  FillChar(Result^, SizeOf(TTreeDataEntry),#0);
  with Result^ do begin
    KeyValue   := Entry.KeyValue;
    ChildValue  := Entry.ChildValue;
    ParentValue := Entry.ParentValue;
    Images := Entry.Images;
    if Assigned(Entry.Node)
    then begin
      Node        := Entry.Node;
      Node.Data   := Result;
    end;
  end;
  FList.Add(Result);
  FStrings.AddObject(DisplayString, TObject(Result));
end;

procedure TTreeDataList.Clear;
var
  I: Integer;
  P: PTreeDataEntry;
begin
  for I := 0 to FList.Count-1 do begin
    P := FList.Items[I];
    Dispose(P);
  end;
  FList.Clear;
  FStrings.Clear;
end;

function TTreeDataList.GetEntryLevel(Index: Integer): Integer;
begin
  Result := 0;
  if (Index >= 0) and (Index < Count) then
    Result := GetStringLevel(FStrings.IndexOfObject(TObject(FList.Items[Index])));
end;

function TTreeDataList.GetStringLevel(Index: Integer): Integer;
begin
  Result := 0;
  if (Index >= 0) and (Index < Count) and (FStrings.Strings[Index] <> EmptyStr)
  then while FStrings.Strings[Index][Result+1] = #9 do Inc(Result);
end;

function TTreeDataList.GetDisplayString(Index: Integer): String;
begin
  Result := '';
  if (Index >= 0) and (Index < Count)
  then Result := ReplaceStr(FStrings.Strings[Index], #9, EmptyStr);
end;

function TTreeDataList.GetDisplayText(Index: Integer): String;
begin
  Result := '';
  if (Index >= 0) and (Index < Count)
  then Result := GetDisplayString(FStrings.IndexOfObject(FList.Items[Index]));
end;

procedure TTreeDataList.EndOfCreate;
{
  function TreeCompare(List: TStringList; Index1, Index2: Integer): Integer;
  var
    Idx1, Idx2: Integer;
  begin
    Idx1 := FList.IndexOf(List.Objects[Index1]);
    Idx2 := FList.IndexOf(List.Objects[Index2]);

  end;
}
  function MoveParent(ParentValue: Variant; StartCount, Level: Integer): Integer;
  var
    I, CI: Integer;
    O: TObject;
  begin
    if StartCount < FList.Count then begin
      Result := StartCount;
      for I := 0 to FList.Count-1 do begin
       if CompareVar(TreeEntry[I].ChildValue, ParentValue) = 0
       then begin
         O := TObject(FList.Items[I]);
         CI := FStrings.IndexOfObject(O);
         if CI <> Result
         then FStrings.Move(CI, Result);
         FStrings.Strings[Result] := MakeStr(#9, Level)+ReplaceStr(FStrings.Strings[Result], #9,'');
         Result := MoveParent(TreeEntry[I].ParentValue, Result+1, Level+1);
       end;
      end;
    end
    else Result := StartCount - 1;
  end;
(*
  function MySort(Index1, Index2: Integer): Integer;
  var
    I1, I2: Integer;

    function CompareIndex(J1, J2: Integer): Integer;
    begin
      Result := I1-I2;
      if Result <> 0
      then if Result > 0
        then Result := 1
          else Result := -1
    end;

  begin
    I1 := FList.IndexOf(FStrings.Objects[Index1]);
    I2 := FList.IndexOf(FStrings.Objects[Index2]);
    if VarIsNull(TreeEntry[I1].ChildValue) and VarIsNull(TreeEntry[I2].ChildValue)
    then Result := CompareIndex(I1, I2)
    else begin
        Result := CompareVar(TreeEntry[I1].ChildValue, TreeEntry[I2].ChildValue);
        if Result = 0
        then Result := CompareIndex(I1, I2)
        else if not VarIsNull(TreeEntry[I1].ChildValue)
          then begin
            Result := CompareVar(TreeEntry[I2].ParentValue, TreeEntry[I1].ChildValue);
            if Result = 0
            then {Result := CompareIndex(I2, I1)
            else} Result := -1;
          end else begin
            Result := CompareVar(TreeEntry[I1].ParentValue, TreeEntry[I2].ChildValue);
            if Result = 0
            then {Result := CompareIndex(I1, I2)
            else} Result := 1;
          end
    end;
  end;

  procedure QuickSort(L, R: Integer);
  var
    I, J, P: Integer;
  begin
    repeat
      I := L;
      J := R;
      P := (L + R) shr 1;
      repeat
        while MySort(I, P) < 0 do Inc(I);
        while MySort(J, P) > 0 do Dec(J);
        if I <= J then
        begin
          FStrings.Exchange(I, J);
          if P = I then
            P := J
          else if P = J then
            P := I;
          Inc(I);
          Dec(J);
        end;
      until I > J;
      if L < J then QuickSort(L, J);
      L := I;
    until I >= R;
  end;
*)

  procedure MakeLevels;
  var
    I, J, K,  Level: Integer;

    function IOfParent(Value: Variant;
                       StartIndex: Integer): Integer;
    var
      I: Integer;
    begin
      Result := -1;
      if StartIndex >= 0 then
        for I := StartIndex downto 0 do
          if CompareVar(TreeEntry[I].ParentValue, Value) = 0
          then begin
            Result := I;
            Break;
          end;
    end;

  begin
    for I := 0 to FStrings.Count-1 do begin
      J := FList.IndexOf(FStrings.Objects[I]);
      Level := 0;
      if HasParent(TreeEntry[J]) then begin
        K := J;
        repeat
          K := IOfParent(TreeEntry[K].ChildValue, K-1);
          if K <> -1 then Inc(Level);
        until K = -1;
      end;
      FStrings.Strings[I] := MakeStr(#9, Level{CalcEntryLevel(TreeEntry[J])})+ReplaceStr(FStrings.Strings[I], #9,'');
    end;
  end;

begin

//  FOnMySort:=MySort;
//  FStrings.CustomSort(MySort);
  case FBuildMode of
  trbmDefault  : MoveParent(null, 0, 0);
(*
  trsmQuickSort: begin
    QuickSort(0, FStrings.Count-1);
  end;
*)
  trbmAsIs     : MakeLevels;
  end;
end;

function TTreeDataList.CalcEntryLevel(Item: TTreeDataEntry): Integer;
var
  I: Integer;
begin
  Result := 0;
  I := IndexOfParent(Item.ChildValue);
  while I >= 0 do begin
    Inc(Result);
    I := IndexOfParent(TreeEntry[I].ChildValue);
  end;
end;

function TTreeDataList.GetItem(Item: TTreeDataEntry; Delta: Integer; NoChild: Boolean): PTreeDataEntry;
begin
  Result := nil;
end;


function TTreeDataList.GetEntryIndex(Index: Integer): TTreeImageIndex;
var
  I: Integer;
begin
  FillChar(Result, SizeOf(TTreeImageIndex), -1);
  if (Index >= 0) and (Index < Count)
  then begin
    I := FList.IndexOf(FStrings.Objects[Index]);
    if I <> -1
    then Result := TreeEntry[I].Images;
  end;
end;

procedure TTreeDataList.BuildTreeNode(Tree: TTreeView; SelValue: Variant);
var
  ANode, NextNode: TTreeNode;
  ALevel, i, j: Integer;
  CurrStr: string;
begin
  Tree.Items.BeginUpdate;
  try
    try
      Tree.Items.Clear;
      if FStrings.Count = 0 then Exit;
      ANode := nil;
      for i := 0 to FStrings.Count - 1 do
      begin
        CurrStr := DisplayString[I];
        ALevel := StringLevel[I];
        if ANode = nil then
          ANode := Tree.Items.AddChild(nil, CurrStr)
        else if ANode.Level = ALevel then
          ANode := Tree.Items.AddChild(ANode.Parent, CurrStr)
        else if ANode.Level = (ALevel - 1) then
          ANode := Tree.Items.AddChild(ANode, CurrStr)
        else if ANode.Level > ALevel then
        begin
          NextNode := ANode.Parent;
          while NextNode.Level > ALevel do
            NextNode := NextNode.Parent;
          ANode := Tree.Items.AddChild(NextNode.Parent, CurrStr);
        end;
{
        ANode.ImageIndex := 0;
        ANode.SelectedIndex := 1;
        ANode.StateIndex := 1;
}
        PTreeDataEntry(FStrings.Objects[i])^.Node := ANode;
        ANode.Data := FStrings.Objects[i];
        ANode.ImageIndex := PTreeDataEntry(FStrings.Objects[i])^.Images.ChildIndex;
        ANode.SelectedIndex := PTreeDataEntry(FStrings.Objects[i])^.Images.ChildSelectedIndex;

        with TrDBTreeNode(ANode), TrDBTreeNode(ANode).DBTree do
          if IsMultiSelect and (CheckList.IndexOf(DataEntry.KeyValue) > -1) then
          begin
            FCheckState := FCheckState + [trncsChecked] - [trncsUnChecked];
            TrDBTreeNode(ANode).ChangeParentState;
          end;

//        ANode.StateIndex := -1;
        ANode.StateIndex := 1;
        if Assigned(ANode.Parent) then begin
          j := IndexOfNode(ANode.Parent);
          ANode.Parent.ImageIndex := PTreeDataEntry(FList[j])^.Images.ParentIndex;
          ANode.Parent.SelectedIndex := PTreeDataEntry(FList[j])^.Images.ParentSelectedIndex;
//          ANode.Parent.StateIndex := -1;
          ANode.Parent.StateIndex := 1;
          //  TVIS_CUT  NotSelectParent
          if TrCustomDBTreeView(Tree).NotSelectParent
          then SetNodeState(ANode.Parent, TVIS_CUT);
        end;
      end;
      i := IndexOfKey(SelValue);
      if i = -1
      then Tree.Selected := nil
      else Tree.Selected := TreeEntry[i].Node;
    finally
      Tree.Items.EndUpdate;
    end;
  except
//    Owner.Owner.Invalidate;  // force repaint on exception
    raise;
  end;
end;

function TTreeDataList.IndexOf(Item: Pointer): Integer;
begin
  Result := FList.IndexOf(Item);
end;

function TTreeDataList.IndexOfNode(Node: TTreeNode): Integer;
var
  I: Integer;
begin
//new
  Result := -1;
  if Assigned(Node)
  then
    if Assigned(Node.Data)
    then Result := FList.IndexOf(Node.Data)
    else for I := 0 to FList.Count-1 do
    if Node = TTreeDataEntry(FList.Items[I]^).Node
    then begin
      Result := I;
      Break;
    end;
end;

function TTreeDataList.IndexOfEntry(Item: TTreeDataEntry): Integer;
var
  I: Integer;
begin
  Result := -1;
  for I := 0 to FList.Count-1 do
    if CompareEntry(Item, TTreeDataEntry(FList.Items[I]^))
    then begin
      Result := I;
      Break;
    end;
end;

function TTreeDataList.IndexOfKey(KeyValue: Variant): Integer;
var
  I: Integer;
begin
  Result := -1;
  for I := 0 to FList.Count-1 do
    if CompareVar(KeyValue, TTreeDataEntry(FList.Items[I]^).KeyValue) = 0
    then begin
      Result := I;
      Break;
    end;
end;

function TTreeDataList.IndexOfParent(ParentValue: Variant): Integer;
var
  I: Integer;
begin
  Result := -1;
  for I := 0 to FList.Count-1 do
    if CompareVar(ParentValue, TTreeDataEntry(FList.Items[I]^).ParentValue) = 0
    then begin
      Result := I;
      Break;
    end;
end;

function TTreeDataList.ChildrenPos(ChildValue: Variant; var Pos: Integer): Integer;
var
  I: Integer;
begin
  Result := -1;
  if (Pos >= 0) and (Pos < FList.Count) then
    for I := Pos to FList.Count-1 do
      if CompareVar(ChildValue, TTreeDataEntry(FList.Items[I]^).ChildValue) = 0
      then begin
        Result := I;
        Break
      end;
end;

function TTreeDataList.HasChildren(Item: TTreeDataEntry): Boolean;
var
  I: Integer;
begin
  Result := False;
  for I := 0 to FList.Count - 1 do
    if CompareVar(Item.ParentValue, TTreeDataEntry(FList.Items[I]^).ChildValue) = 0
    then begin
      Result := True;
      Break
    end;
end;

function TTreeDataList.HasParent(Item: TTreeDataEntry): Boolean;
begin
  Result := not VarIsEmpty(Item.ChildValue) and not VarIsNull(Item.ChildValue);
end;

function TTreeDataList.IsEmptyEntry(Index: Integer): Boolean;
begin
  Result := FullCompareEntry(FEmpty, TreeEntry[Index]);
end;

{ TrTreeViewDataLink }

procedure TrTreeViewDataLink.ActiveChanged;
begin
  if FControl <> nil then FControl.ActiveChanged;
end;

//**
procedure TrTreeViewDataLink.DataSetScrolled(Distance: Integer);
begin
  if FControl <> nil then FControl.DataSetScrolled(Distance);
end;

procedure TrTreeViewDataLink.LayoutChanged;
begin
  if FControl <> nil then FControl.LayoutChanged;
//  inherited LayoutChanged;
end;

procedure TrTreeViewDataLink.UpdateData;
begin
  if FControl <> nil then FControl.UpdatingData;
end;
//**

procedure TrTreeViewDataLink.DataSetChanged;
begin
  if FControl <> nil then FControl.DataChanged;
end;

procedure TrTreeViewDataLink.EditingChanged;
begin
  if FControl <> nil then FControl.EditingChanged;
end;

procedure TrTreeViewDataLink.RecordChanged(Field: TField);
begin
  if FControl <> nil then FControl.RecordChanged(Field);
end;

{ TrTreeCheckList }

function TrTreeCheckList.Add(Item: Variant): Integer;
begin
  if AllowChange(Item, vlnAdd) then begin
    Result := inherited Add(Item);
    if Assigned(FTreeView) then FTreeView.AfterCheckChange(Item, vlnAdd);
    if Assigned(FCheckListChange) then FCheckListChange(FTreeView, Item, vlnAdd);
  end
  else Result := -1;
end;

function TrTreeCheckList.AllowChange(Item: Variant;
  Action: TVarListAction): Boolean;
begin
  Result := True;
  if Assigned(FCheckListChanging) then FCheckListChanging(FTreeView, Item, Action, Result);
end;

procedure TrTreeCheckList.Clear;
var
  I: Integer;
begin
  inherited Clear;
  if Assigned(FCheckListChange) then FCheckListChange(FTreeView, Null, vlnClear);
  with FTreeView do
    for i := 0 to Items.Count - 1 do begin
      Items.BeginUpDate;
      try
        TrDBTreeNode(Items[I]).FCheckState := [trncsUnChecked];
      finally
        Items.EndUpdate;
        Invalidate;
      end;
    end;
end;

constructor TrTreeCheckList.Create(ATree: TrCustomDBTreeView);
begin
  inherited Create;
  FTreeView := ATree;
end;

procedure TrTreeCheckList.Delete(Index: Integer);
var
  Value: Variant;
begin
  Value := Items[Index];
  if AllowChange(Value, vlnDelete) then begin
    inherited Delete(Index);
    if Assigned(FTreeView) then FTreeView.AfterCheckChange(Value, vlnDelete);
    if Assigned(FCheckListChange) then FCheckListChange(FTreeView, Value, vlnDelete);
  end;
end;

procedure TrTreeCheckList.Insert(Index: Integer; Item: Variant);
begin
  if AllowChange(Item, vlnInsert) then begin
    inherited Insert(Index, Item);
    if Assigned(FTreeView) then FTreeView.AfterCheckChange(Item, vlnAdd);
    if Assigned(FCheckListChange) then FCheckListChange(FTreeView, Item, vlnInsert);
  end;
end;

function TrTreeCheckList.Remove(Item: Variant): Integer;
begin
  if AllowChange(Item, vlnDelete) then begin
    Result := inherited Remove(Item);
    if Assigned(FTreeView) then FTreeView.AfterCheckChange(Item, vlnDelete);
    if Assigned(FCheckListChange) then FCheckListChange(FTreeView, Item, vlnDelete);
  end
  else Result := -1;
end;

{TrCustomDBTreeView}

constructor TrCustomDBTreeView.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
//***9
//  FOnlyChild    := True;
  FNoCloseUp    := False;
  FDestroingWindow := False;
  FOnlyChild    := False;
  FNotSelectParent:= False;
  FUpdateEnbled := False;
  FLastSelected := nil;
  FDataChangedCalled := False;
  FDisableStyle    := tdsDefault;
  FLookupControl   := False;
  FHasMoved        := True;
  FIsResync        := False;
  FAutoSelect      := False;
  FAutoMove        := True;
  FListActive      := False;
  FEditNode        := nil;
  FillChar(FImageItedexes, SizeOf(FImageItedexes), -1);
  VarClear(FKeyValue);
  FValue            := EmptyStr;
  FPrevState        := dsInactive;
  FCurrentRecord    := -1;
  FRecordCount      := 0;
  FRecordIndex      := 0;
  FRefreshEnabled   := True;
  FHasInsert        := False;
  FHasCreated       := False;
  FDataSet          := nil;
  FOnGetText        := nil;
  FOnGetImageIndexes:= nil;
  FListFields       := TList.Create;
  FTreeData         := TTreeDataList.Create;
  ReadOnly          := True;
  FSelectedEnabled  := False;
  FKeyFieldName     := EmptyStr;
  FDisplayFields    := EmptyStr;
  FChildrenFieldName:= EmptyStr;
  FParentFieldName  := EmptyStr;
  FKeyField         := nil;
  FChildrenField    := nil;
  FParentField      := nil;
  FLinkData         := TrTreeViewDataLink.Create;
  FLinkData.FControl:= Self;
  FIsFirst := True;
  FSearchString     := EmptyStr;
  FDefaultCheckImage := TCustomImageList.Create(Self);
  DrawDefaultCheck({$IFNDEF POLARIS_D4}TCrackCustomImageList(FDefaultCheckImage){$ELSE}FDefaultCheckImage{$ENDIF}, GetItemHeight);

//  StateImages := FDefaultCheckImage;

  FInternalCheck := False;
  FMultiSelect := TrTreeMultiSelect.Create(Self);
  FCheckList := TrTreeCheckList.Create(Self);
  FCheckKey := scCheckKey;
  FCheckShift := [ssShift];
end;

destructor TrCustomDBTreeView.Destroy;
begin
  FEditNode := nil;
//  ClearNodes;
  FDataSet := nil;
  FLinkData.FControl := nil;
  FLinkData.Free;
  FListFields.Free;
  FTreeData.Free;
  if Assigned(FDefaultCheckImage)
  then begin
    FDefaultCheckImage.Free;
    FDefaultCheckImage := nil;
  end;
  FMultiSelect.Free;
  FMultiSelect := nil;

  inherited Destroy;
end;

//Multi
function TrCustomDBTreeView.IsMultiSelect: Boolean;
begin
  Result := FMultiSelect.FEnabled;
end;

procedure TrCustomDBTreeView.SetMultiSelect(
  const Value: TrTreeMultiSelect);
begin
  FMultiSelect.Assign(Value);
end;

function TrCustomDBTreeView.GetItemHeight: Integer;
begin
  Result := 16;
  if HandleAllocated then
{$IFDEF POLARIS_D4}
    Result := TreeView_GetItemHeight(Handle);
{$ELSE}
    Result := 16;
{$ENDIF}
end;

procedure TrCustomDBTreeView.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
//  Params.Style := Params.Style or TVS_CHECKBOXES;
end;

procedure TrCustomDBTreeView.CreateWnd;
var
  B: Boolean;
begin
  B := HandleAllocated;
  inherited;
  if not B then begin
    if FTreeData.Count <> FLinkData.BufferCount
    then CreateEntryData;
    CreatedNodes;
    FIsFirst := True;
  end;
  if HandleAllocated and IsMultiSelect
  then TreeView_SetImageList(Handle, FDefaultCheckImage.Handle, TVSIL_STATE);
end;

procedure TrCustomDBTreeView.WMDestroy(var Message: TWMDestroy);
begin
  try
    FDestroingWindow := True;
    Selected := nil;
    inherited;
  finally
    FDestroingWindow := False;
  end;
end;

function TrCustomDBTreeView.ChangeCheckNode(Node: TTreeNode): Boolean;
begin
// 
  Result := False;
  if Assigned(Node) and IsMultiSelect then begin
  //Multi
  //  Check
{
    if TrDBTreeNode(Node).CheckState *
    TrDBTreeNode(Node).Checked := not TrDBTreeNode(Node).Checked;
}

    //    VarList
{
    case MultiSelect.Option of
    tmsAll:
      TrDBTreeNode(Node).Checked := not (trncsChecked in TrDBTreeNode(Node).CheckState);
    tmsAllNotHaveChild:
      TrDBTreeNode(Node).Checked := not (   (trncsChecked in TrDBTreeNode(Node).CheckState)
                                         or (trncsCheckChild in TrDBTreeNode(Node).CheckState));
    end;
}
//tpoSelectNode, tpoSelectChild
    if MultiSelect.FEnabled
    then begin
      if MultiSelect.FParentOptions * [tpoSelectNode, tpoSelectChild] = [tpoSelectChild]
      then TrDBTreeNode(Node).Checked := not ((trncsChecked in TrDBTreeNode(Node).CheckState)
                                         or (trncsCheckChild in TrDBTreeNode(Node).CheckState))
      else if MultiSelect.FParentOptions * [tpoSelectNode] <> []
        then TrDBTreeNode(Node).Checked := not (trncsChecked in TrDBTreeNode(Node).CheckState);
    end;
    Invalidate;
    Result := True;
  end
end;

procedure TrCustomDBTreeView.WndProc(var Message: TMessage);

  function TestCheck(var Message: TWMMouse): Boolean;
  var
    HTest: THitTests;
    Shift: TShiftState;
    Node : TTreeNode;
  begin
    Result := False;
    HTest := GetHitTestInfoAt(Message.XPos, Message.YPos);
    Shift := KeysToShiftState(Message.Keys);
    if (ssLeft in Shift) and (htOnStateIcon in HTest)
    then begin
      Exclude(Shift, ssLeft);
      Node := GetNodeAt(Message.XPos, Message.YPos);
      if Assigned(Node) and (Node <> Selected)
      then Node.Selected := True;
      if (Shift = [])
      then if ChangeCheckNode(Node)
        then begin
          Result := True;
          Message.Result := 1;
        end;
    end;
  end;

begin
  if not (csDesigning in ComponentState) and RightClickSelect and
    (Message.Msg = WM_RBUTTONDOWN) then Message.Msg := WM_LBUTTONDOWN;
  if not (csDesigning in ComponentState) and ((Message.Msg = WM_LBUTTONDOWN) or
   (Message.Msg = WM_LBUTTONDBLCLK))
   and not Dragging and not IsEditing
   then
  begin
    if (not TestCheck(TWMMouse(Message)) and not IsControlMouseMsg(TWMMouse(Message)))
    or not (Focused or FLookupControl)
    then inherited WndProc(Message);
  end
  else
    if not ((Message.Msg = WM_CONTEXTMENU) and (Owner <> nil) and (Owner is TrPopupTree))
    then inherited WndProc(Message);
  if FIsFirst and HandleAllocated and (Message.Msg = WM_Paint) and
    (not Assigned(StateImages) and
      Assigned(FDefaultCheckImage) and FDefaultCheckImage.HandleAllocated)
  then begin
    FIsFirst := False;
    Invalidate;
  end;
end;

{
procedure TrCustomDBTreeView.WndProc(var Message: TMessage);
begin
  inherited WndProc(Message);

  if FIsFirst and HandleAllocated and (Message.Msg = WM_Paint) and
  (TWMPaint(Message).DC = Canvas.Handle)
  then begin
    FIsFirst := False;
    Invalidate;
  end;

end;
}
function TrCustomDBTreeView.Locate(const S: String; var Index: Integer;
                                   var Node: TTreeNode): Boolean;
var
  L, I, J: Integer;
begin
  Result := False;
  Index := -1;
  J := -1;
  Node  := nil;
  L := Length(S);
  if L = 0 then Exit;
  for I := 0 to FTreeData.FStrings.Count-1 do
  if AnsiCompareText(Copy(FTreeData.DisplayString[I],1,L), S) =0
  then begin
    Result := True;
    J := I;
    Break;
  end;
  if Result then begin
    Index := FTreeData.IndexOf(FTreeData.FStrings.Objects[J]);
    Node  := FTreeData.TreeNode[Index];
  end;
end;

procedure TrCustomDBTreeView.DoEnter;
begin
  FHasMoved := False;
  Change(Selected);
  inherited DoEnter;
end;

procedure TrCustomDBTreeView.DoExit;
begin
  FHasMoved := True;
  inherited DoExit;
end;

function TrCustomDBTreeView.GetChildIndex: Integer;
begin
  Result := FImageItedexes[False, False]
end;

function TrCustomDBTreeView.GetChildSelected: Integer;
begin
  Result := FImageItedexes[False, True]
end;

function TrCustomDBTreeView.GetParentIndex: Integer;
begin
  Result := FImageItedexes[True, False]
end;

function TrCustomDBTreeView.GetParentSelected: Integer;
begin
  Result := FImageItedexes[True, True]
end;

function TrCustomDBTreeView.GetDisplayValue: String;
begin
  Result := EmptyStr;
  if Selected <> nil
  then Result := Selected.Text;
end;

procedure TrCustomDBTreeView.SetChildIndex(AValue: Integer);
begin
  if FImageItedexes[False, False] <> AValue
  then begin
    if AValue < -1 then AValue := -1;
    FImageItedexes[False, False] := AValue;
    ChangeImages;
  end;
end;

procedure TrCustomDBTreeView.SetChildSelected(AValue: Integer);
begin
  if FImageItedexes[False, True] <> AValue
  then begin
    if AValue < -1 then AValue := -1;
    FImageItedexes[False, True] := AValue;
    ChangeImages;
  end;
end;

procedure TrCustomDBTreeView.SetParentIndex(AValue: Integer);
begin
  if FImageItedexes[True, False] <> AValue
  then begin
    if AValue < -1 then AValue := -1;
    FImageItedexes[True, False] := AValue;
    ChangeImages;
  end;
end;

procedure TrCustomDBTreeView.SetParentSelected(AValue: Integer);
begin
  if FImageItedexes[True, True] <> AValue
  then begin
    if AValue < -1 then AValue := -1;
    FImageItedexes[True, True] := AValue;
    ChangeImages;
  end;
end;

procedure TrCustomDBTreeView.SetSearchString(AValue: String);
var
  I: Integer;
  ANode: TTreeNode;
begin
  I := -1;
  ANode := nil;
  if (ANSICompareText(AValue, FSearchString) <> 0) and
    (Locate(AValue, I, ANode) or (AValue = EmptyStr)) then FSearchString := AValue;
  if ANode <> nil then Selected := ANode;
end;

procedure TrCustomDBTreeView.ChangeImages;
var
  LR, J, I: Integer;
begin
//          DataSet
  if Items.Count = 0 then Exit;
  LR := FLinkData.ActiveRecord;
  try
    for I := 0 to Items.Count-1 do begin
      J := FTreeData.IndexOf(Items[I].Data);
      FLinkData.ActiveRecord := J;
      GetImageIndexes(Items[I]);
      if Assigned(FOnGetImageIndexes)
      then begin
        FOnGetImageIndexes(Self, FDataSet, FTreeData, PTreeDataEntry(Items[I].Data));
        if Items[I].HasChildren
        then begin
          Items[I].ImageIndex := PTreeDataEntry(Items[I].Data)^.Images.ParentIndex;
          Items[I].SelectedIndex := PTreeDataEntry(Items[I].Data)^.Images.ParentIndex;
//          Items[I].StateIndex := -1;
          Items[I].StateIndex := 1;
        end else begin
          Items[I].ImageIndex := PTreeDataEntry(Items[I].Data)^.Images.ChildIndex;
          Items[I].SelectedIndex := PTreeDataEntry(Items[I].Data)^.Images.ChildSelectedIndex;
//          Items[I].StateIndex := -1;
          Items[I].StateIndex := 1;
        end;
      end;
    end;
  finally
    FLinkData.ActiveRecord := LR;
  end;
end;

function TrCustomDBTreeView.CreateNode: TTreeNode;
begin
{
  Result := inherited CreateNode;
  GetImageIndexes(Result);
}
  Result := TrDBTreeNode.Create(Items);
  GetImageIndexes(Result);
//  if Assigned(FOnGetImageIndexes)
//  then FOnGetImageIndexes(Self, FDataSet, Result);
end;

function TrCustomDBTreeView.CanCollapse(Node: TTreeNode): Boolean;
{var
  N: TTreeNode;}
begin
  if Assigned(Selected) and Assigned(Node) and
    (Node <> Selected) then begin
    Result := True;
    {N := Selected;
    repeat
      N := N.Parent;
      if N = Node then begin
        Result := False;
        Break;
      end;
    until N = nil;}
    if Assigned(OnCollapsing) then OnCollapsing(Self, Node, Result);
  end
  else Result := inherited CanCollapse(Node);
end;

procedure TrCustomDBTreeView.GetImageIndexes(Node: TTreeNode);
begin
  if Node <> nil then begin
    Node.ImageIndex := FImageItedexes[Node.HasChildren, False];
    Node.SelectedIndex := FImageItedexes[Node.HasChildren, True];
  end;
end;

{$IFDEF POLARIS_D4}
function TrCustomDBTreeView.ExecuteAction(Action: TBasicAction): Boolean;
begin
  Result := inherited ExecuteAction(Action) or
    ((FLinkData <> nil) and FLinkData.ExecuteAction(Action));
end;

function TrCustomDBTreeView.UpdateAction(Action: TBasicAction): Boolean;
begin
  Result := inherited UpdateAction(Action) or
    ((FLinkData <> nil) and FLinkData.UpdateAction(Action));
end;

procedure TrCustomDBTreeView.SetDisableStyle(Value: TrTreeDisableStyle);
begin
  if FDisableStyle <> Value then begin
    FDisableStyle := Value;
    CheckDisableColor(True);
  end;
end;

procedure TrCustomDBTreeView.CheckDisableColor(Repaint: Boolean);
var
  cl: TColor;
begin
  if not FLookupControl and (FDisableStyle = tdsExplorer) then begin
    if Enabled then cl := Color
    else cl := clInactiveBorder;
  end
  else cl := Color;
  Brush.Color := cl;
  if HandleAllocated
  then TreeView_SetBkColor(Handle, ColorToRGB(cl));
  if Repaint then Invalidate;
end;

procedure TrCustomDBTreeView.CMEnabledChanged(var Message: TMessage);
begin
  CheckDisableColor(False);
  inherited;
end;

procedure TrCustomDBTreeView.InternalDrawItem(Sender: TCustomTreeView;
      Node: TTreeNode; State: TCustomDrawState; var DefaultDraw: Boolean);
begin
  if Assigned(FSaveProc)
  then FSaveProc(Sender, Node, State, DefaultDraw);
end;

procedure TrCustomDBTreeView.CustomDrawParam(Node: TTreeNode; State: TCustomDrawState;
      Stage: TCustomDrawStage);
begin
  if (Stage = cdPrePaint) and not Enabled then begin
    if not (cdsSelected in State)
    then Canvas.Brush.Color := Color
    else Canvas.Brush.Color := clInactiveBorder;
    Canvas.Font.Color := clGrayText;
  end;
end;

function TrCustomDBTreeView.CustomDrawItem(Node: TTreeNode; State: TCustomDrawState;
      Stage: TCustomDrawStage{$IFDEF POLARIS_D5}; var PaintImages: Boolean{$ENDIF}): Boolean;
begin
  if (Stage = cdPrePaint) and Assigned(OnCustomDrawItem)
  then CustomDrawParam(Node, State, Stage);
  Result := inherited CustomDrawItem(Node, State, Stage{$IFDEF POLARIS_D5}, PaintImages{$ENDIF});
end;
{
procedure TrCustomDBTreeView.TVMGETDISPINFO(var Message: TMessage);
var
  R: PRect;
begin
  inherited;
end;
}
procedure TrCustomDBTreeView.CNNotify(var Message: TWMNotify);
var
  Node: TTreeNode;
//Multi
{  MousePos: TPoint;
  HitTest: TTVHitTestInfo ;
  R: TRect;}

begin
  if TestChildEnabled
  then PreNotify(Message);

//Multi
  with Message do
  case NMHdr^.code of
  TVN_GETDISPINFO: if IsMultiSelect
    then with PTVDispInfo(NMHdr)^ do begin
      Node := Items.GetNode(item.hItem);
      if Node <> nil
      then with PTVDispInfo(NMHdr)^.item do begin
        mask := mask or TVIF_STATE or TVIF_HANDLE;
        stateMask := stateMask or TVIS_STATEIMAGEMASK;
        state := (state and not TVIS_STATEIMAGEMASK)
          or UINT(IndexToStateImageMask(TrDBTreeNode(Node).GetCheckImageIndex + 1));
      end;
    end;
  {NM_CLICK: begin
      GetCursorPos(MousePos);
      MousePos := ScreenToClient(MousePos);
      HitTest.pt := MousePos;
      TreeView_HitTest(Handle, HitTest);
      Node    := Items.GetNode(HitTest.hItem);
      if Assigned(Node)
      then begin
        TreeView_GetItemRect(Handle, HitTest.hItem, R, True);
        R.Right := R.Left;
        if Assigned(StateImages)
        then Dec(R.Left, StateImages.Width)
        else Dec(R.Left, FDefaultCheckImage.Width);
        if PtInRect(R,MousePos)
        then ChangeCheckNode(Node);
      end;
    end;}
  end;

  if (FLookupControl or (FDisableStyle = tdsGrayText))
  and ((Message.NMHdr^.code=NM_CUSTOMDRAW))
  then begin
    FSaveProc := OnCustomDrawItem;
    try
      OnCustomDrawItem := InternalDrawItem;
      inherited;
    finally
      OnCustomDrawItem := FSaveProc;
      FSaveProc := nil;
    end;
  end else inherited;
end;
{$ENDIF}

function TrCustomDBTreeView.GetNodeFromItem(const Item: TTVItem): TTreeNode;
begin
  Result := nil;
  if Items <> nil then
    with Item do
      if (state and TVIF_PARAM) <> 0 then Result := Pointer(lParam)
      else Result := Items.GetNode(hItem);
end;

function TrCustomDBTreeView.TestChildEnabled: Boolean;
begin
  Result := Focused and FNotSelectParent and not FHasCreated;
end;

function TrCustomDBTreeView.GetChildNode(Node: TTreeNode; Delta: Integer): TTreeNode;
begin
  Result := Node;
  if Assigned(Result) then begin
    if Delta < 0
    then begin
      Node := Result.getPrevSibling;
      if Assigned(Node) then begin
        Result := Node;
        while Result.HasChildren do
          Result := Result.GetLastChild
      end
      else begin
        while Result.HasChildren do
          Result := Result.getFirstChild;
      end
    end
    else while Result.HasChildren do
      Result := Result.getFirstChild;
  end;
end;

procedure TrCustomDBTreeView.PreNotify(var Message: TWMNotify);
var
  ADelta: Integer;
  B: Boolean;
  OldN, NewN: TTreeNode;
begin
  with Message do
    case NMHdr^.code of
    NM_SETFOCUS: if Assigned(Selected) then begin
        NewN := GetChildNode(Selected, 0);
        TreeView_SelectItem(Handle, NewN.ItemId);
      end;
    TVN_SELCHANGINGA, TVN_SELCHANGINGW:
      if Items.Count = FTreeData.Count then begin
        OldN := GetNodeFromItem(PNMTreeView(NMHdr)^.itemOld);
        NewN := GetNodeFromItem(PNMTreeView(NMHdr)^.itemNew);
        if Assigned(NewN)
        then B := NewN.HasChildren
        else Exit;
        ADelta := 0;
        if Assigned(OldN) and Assigned(NewN) and (PNMTreeView(NMHdr)^.action=TVC_BYKEYBOARD)
        then ADelta := NewN.AbsoluteIndex-OldN.AbsoluteIndex;
        NewN := GetChildNode(NewN, ADelta);
        if B then begin
          Result := 1;
          TreeView_SelectItem(Handle, NewN.ItemId);
          FNoCloseUp := True;
          Exit;
        end;
      end;
    TVN_GETDISPINFOA, TVN_GETDISPINFOW: begin
      NewN := Items.GetNode(PTVDispInfo(NMHdr)^.item.hItem);
      if Assigned(NewN) and NewN.HasChildren
      then begin
        PTVDispInfo(NMHdr)^.item.mask := PTVDispInfo(NMHdr)^.item.mask or TVIF_STATE;
        PTVDispInfo(NMHdr)^.item.stateMask := PTVDispInfo(NMHdr)^.item.stateMask or TVIS_CUT;
        PTVDispInfo(NMHdr)^.item.state := PTVDispInfo(NMHdr)^.item.state or TVIS_CUT;
      end;

    end;
    end; //case
end;

procedure TrCustomDBTreeView.ScrollTo(Pos: Smallint);
var
  M: TMessage;
begin
  if HandleAllocated then begin
    FillChar(M, SizeOf(TMessage), #0);
    TWMScroll(M).Msg := WM_HSCROLL;
    TWMScroll(M).ScrollCode := SB_THUMBPOSITION;
    TWMScroll(M).Pos := Pos;
    SendMessage(Handle, WM_HSCROLL, M.WParam, M.LParam);
    TWMScroll(M).ScrollCode := SB_ENDSCROLL;
    SendMessage(Handle, WM_HSCROLL, M.WParam, M.LParam);
  end;
end;

function TrCustomDBTreeView.ScrollPos: SmallInt;
begin
  if HandleAllocated
  then Result := GetScrollPos(Handle, SB_HORZ)
  else Result := 0;
end;

function TrCustomDBTreeView.CanEdit(Node: TTreeNode): Boolean;
begin
  inherited CanEdit(Node);
  Result := False;
end;

procedure TrCustomDBTreeView.Change(Node: TTreeNode);
var
  Child: TTreeNode;
begin
  if not (csDestroying in ComponentState) and not FDestroingWindow
    and (Node <> nil)
  then begin
    if FOnlyChild and
      ((not FLookupControl and Focused) or FLookupControl) and Assigned(Node)
    then begin
      Child := Node;
      while Child.HasChildren do
        Child := Child.getFirstChild;
      Selected := Child;
      if Node <> Child then exit;
    end;
    if not FHasCreated
    then begin
      if FAutoMove then
      SyncData(Node);
      inherited Change(Node);
    end
  end
  else inherited Change(Node);
end;

procedure TrCustomDBTreeView.InheritedReadOnly(AValue: Boolean);
begin
  inherited ReadOnly := AValue;
end;

procedure TrCustomDBTreeView.KeyValueChanged;
begin
  if Assigned(FOnKeyChanged) then FOnKeyChanged(Self, Selected);
end;

procedure TrCustomDBTreeView.ActiveChanged;
begin
  FListActive := False;
  if FDataSet <> FLinkData.DataSet
  then FDataSet := FLinkData.DataSet;

  if CheckActive then begin
    CheckNotCircular;
    InitFields(False);
//    UpdateRecordCount;
    CreateEntryData;
//!new
    if HandleAllocated
    then begin
      CreatedNodes;
      FRecordCount := FLinkData.RecordCount;
    end;
  end
  else begin
    FLinkData.BufferCount := 0;
    ClearNodes;
  end;
end;

procedure TrCustomDBTreeView.ChangeChildOf(Node: TTreeNode; ChildValue: Variant);
var
  I:  Integer;
  NN,
  TN: TTreeNode;
  B:  Boolean;
begin
  if (FTreeData <> nil) and (FTreeData.Count > 0) and
  (Node <> nil) and (Node.Data <> nil) and
  (CompareVar(PTreeDataEntry(Node.Data)^.ChildValue, ChildValue) <> 0)
  then begin
    I := FTreeData.IndexOfParent(ChildValue);
    B := False;
    if I >= 0
    then begin
      TN := FTreeData.TreeNode[I];
      B := not TN.HasChildren;
      if B then NN := Items.AddChild(TN, EmptyStr)
      else NN := TN.GetFirstChild;
    end
    else NN := Items.GetFirstNode;
    if NN <> nil then Node.MoveTo(NN, naAddFirst);
    if B then NN.Free;
    PTreeDataEntry(Node.Data)^.ChildValue := ChildValue;
  end;
end;

procedure TrCustomDBTreeView.ChangeParentOf(Node: TTreeNode; ParentValue: Variant);
var
  I, P: Integer;
  TN:   TTreeNode;
  B:    Boolean;
begin
  if (FTreeData <> nil) and (FTreeData.Count > 0) and (Node <> nil) and (Node.Data <> nil)
  and (CompareVar(PTreeDataEntry(Node.Data)^.ParentValue,ParentValue) <> 0)
  then begin
    PTreeDataEntry(Node.Data)^.ParentValue := ParentValue;
    P := 0;
    while P >= 0 do begin
      I := FTreeData.ChildrenPos(ParentValue, P);
      if I < 0 then Break;
      if (FTreeData.IndexOf(Node.Data)<>I)
      and (Node <> FTreeData.TreeNode[I].Parent)
      then begin
        B := not Node.HasChildren;
        if B then TN := Items.AddChild(Node,EmptyStr)
        else TN := Node.GetLastChild;
        FTreeData.TreeNode[I].MoveTo(TN, naAdd);
        if B then TN.Free;
      end;
      P := I+1;
    end;
  end;
end;

procedure TrCustomDBTreeView.ChangeDataOf(Node: TTreeNode; KeyValue: Variant);
begin
  if (FTreeData <> nil) and (FTreeData.Count > 0) and (Node <> nil) and (Node.Data <> nil)
  then begin
    PTreeDataEntry(Node.Data)^.KeyValue := KeyValue;
  end;
end;

procedure TrCustomDBTreeView.ReSyncRecords;
begin
  FIsResync := True;
  if CheckActive then
  try
    FRecordCount := FLinkData.RecordCount;
    UpdateRecordCount;
    if not (FDataSet.State in [dsEdit,dsInsert])
    then CreatedNodes
  finally
    FIsResync := False;
  end;
end;

procedure TrCustomDBTreeView.LayoutChanged;
begin
  NewLayout;
end;

function TrCustomDBTreeView.UpdateRecordCount: Integer;
begin
  Result := FLinkData.RecordCount;
  if FLinkData.BufferCount <= FLinkData.RecordCount then
    while FLinkData.RecordCount = FLinkData.BufferCount do begin
      FLinkData.BufferCount := FLinkData.BufferCount+1000;
      if FLinkData.BufferCount > FLinkData.RecordCount
      then Break;
    end;
  FLinkData.BufferCount := FLinkData.RecordCount;
end;

procedure TrCustomDBTreeView.DataSetScrolled(Distance: Integer);
{
var
  sp: SmallInt;
}
begin
  FCurrentRecord := FLinkData.ActiveRecord;
  if {FAutoMove and} HandleAllocated{ and not FLookupControl}
  then begin
    Selected := FTreeData.TreeNode[FCurrentRecord];
{
    if not Selected.IsVisible
    then begin
      sp := ScrollPos;
      Selected.MakeVisible;
      ScrollTo(sp);
    end;
}
  end;
end;

procedure TrCustomDBTreeView.DataChanged;
var
  TCN,
  CN: Boolean;
  ROC,
  RNC: Integer;
begin
  if CheckActive
  then begin
  if FDataChangedCalled or FHasCreated then Exit;
  try
    FDataChangedCalled := true;
//   Refresh
    CN := False;
    ROC := FRecordCount;
    TCN := FLinkData.RecordCount <> ROC;
    if not TCN then begin
      FLinkData.BufferCount := FLinkData.BufferCount+1;
      TCN := (ROC = FLinkData.RecordCount);
    end;
    if TCN
    then UpdateRecordCount;
    RNC := FLinkData.RecordCount;
//    FRecordCount := FLinkData.RecordCount;
    if RNC <> ROC then FRecordCount := -1;
    if
     {not (FPrevState in [dsInsert, dsEdit]) or  and}
{        (
        (not FDataSet.ControlsDisabled and FRefreshEnabled and (FCurrentRecord=FLinkData.ActiveRecord)) or
         (FRecordCount <> UpdateRecordCount))}
    (FPrevState in [dsInsert, dsEdit]) or
    (not (FDataSet.State in [dsInsert, dsEdit]) and
//        ((FRecordCount <> UpdateRecordCount) or (FRecordCount=-1))
        ((RNC <> UpdateRecordCount) or (FRecordCount=-1))
     )
    then begin
      FRecordCount := RNC;
      CreateEntryData;
      CN := True;
      FPrevState := FDataSet.State;
    end;

    if HandleAllocated
    then begin
      if CN
      then CreatedNodes;
      if FAutoMove then
      SyncDataSet;
    end;
    if not FLookupControl
    and (FCurrentRecord <> FLinkData.ActiveRecord)
    and (FTreeData.Count <> 0)
    and (FTreeData.TreeNode[FLinkData.ActiveRecord] <> Selected)
    then DataSetScrolled(0);

  finally
    FDataChangedCalled := False;
  end
  end
  else begin
    FCurrentRecord := -1;
    FRecordCount   := 0;
    FRecordIndex   := 0;
    FValue         := EmptyStr;
    FKeyValue      := null;
    ClearNodes;
  end;
end;

procedure TrCustomDBTreeView.ChangeNodeText(Node: TTreeNode; const NewText: String);
var
  I: Integer;
begin
  if Node <> nil then begin
//    I := ;
    I := TreeData.FStrings.IndexOfObject(TreeData.FList[Node.AbsoluteIndex]);
    if ANSICompareText(FTreeData.FStrings.Strings[I], NewText) <> 0
    then begin
      FTreeData.FStrings.Strings[I] := NewText;
      if FTreeData.TreeNode[FTreeData.IndexOf(FTreeData.FStrings.Objects[I])] <> Node
      then FTreeData.TreeNode[FTreeData.IndexOf(FTreeData.FStrings.Objects[I])] := Node;
      Node.Text := FTreeData.DisplayText[I];
    end;
  end;
end;

procedure TrCustomDBTreeView.EditingChanged;
var
  E: TTreeDataEntry;
  sp: Smallint;
  FNextSelected: TTreeNode;
begin
  if CheckActive then begin
    FHasCreated := True;
    Case FLinkData.DataSet.State of
    dsEdit  : begin
                FCurrentRecord := FLinkData.ActiveRecord;
                FEditNode := FTreeData.TreeNode[FLinkData.ActiveRecord];
                FPrevState := FLinkData.DataSet.State;
              end;
    dsInsert: begin
                with FTreeData.EmptyEntry do
//                FCurrentRecord := FTreeData.Count-1;
                FCurrentRecord := FLinkData.ActiveRecord;
                E := FTreeData.EmptyEntry;
                FEditNode := Items.Add(Selected, EmptyStr);
                if Selected <> nil then begin
                  FNextSelected := Selected.GetNextSibling;
                  if FNextSelected <> nil then FEditNode.MoveTo(FNextSelected, naInsert);
                end;
{
                E.KeyValue := FTreeData.EmptyEntry.KeyValue;
                E.ChildValue := FTreeData.EmptyEntry.ChildValue;
                E.ParentValue := FTreeData.EmptyEntry.ParentValue;
}
                E.Node := FEditNode;
                E.Images.ParentIndex := ParentIndex;
                E.Images.ParentSelectedIndex := ParentSelected;
                E.Images.ChildIndex := ChildIndex;
                E.Images.ChildSelectedIndex := ChildSelected;
                FTreeData.AddEntry(E, EmptyStr);
                FLinkData.BufferCount := FLinkData.BufferCount+1{2};
                sp := ScrollPos;
                FEditNode.MakeVisible;
                ScrollTo(sp);
//                Invalidate;
                FEditNode.Selected := True;
                FPrevState := FLinkData.DataSet.State;
              end;
    else begin
      FEditNode := nil;
      FHasCreated := False;
    end
    end;
{
    if (FPrevState in [dsInsert, dsEdit]) and
       (FLinkData.DataSet.State = dsBrowse)
    then FCurrentRecord := FLinkData.ActiveRecord;
}
//    FPrevState := FLinkData.DataSet.State;
  end;
end;

procedure TrCustomDBTreeView.UpdatingData;
begin
  FUpdateEnbled := True;
end;

procedure TrCustomDBTreeView.RecordChanged(Field: TField);
begin
//  if not FUpdateEnbled then Exit;
  if (Field = nil) or (Field=FKeyField)
  or (Field=FChildrenField) or (Field=FParentField)
  or ((Field<>nil) and
      (Pos(AnsiUpperCase(Field.FieldName),
           AnsiUpperCase(FDisplayFields))>0))
  then begin

    if FEditNode <> nil then
      if Field <> nil then
        if Field = FChildrenField
        then ChangeChildOf(FEditNode, Field.Value)
        else if Field = FParentField
        then ChangeParentOf(FEditNode, Field.Value)
          else if Field = FKeyField
          then ChangeDataOf(FEditNode, Field.Value)
            else ChangeNodeText( FEditNode, DisplayString)
      else begin
        ChangeChildOf (FEditNode, FChildrenField.Value);
        ChangeParentOf(FEditNode, FParentField.Value);
        CHangeDataOf  (FEditNode, FKeyField.Value);
        ChangeNodeText(FEditNode, DisplayString);
      end;
     FUpdateEnbled := False;
  end;
end;

function TrCustomDBTreeView.CheckSource: Boolean;
begin
  Result := FLinkData <> nil
end;

function TrCustomDBTreeView.CheckDataSet: Boolean;
begin
  Result := CheckSource and (FDataSet<>nil)
end;

function TrCustomDBTreeView.CheckActive: Boolean;
begin
  Result := CheckDataSet and FDataSet.Active
end;

function TrCustomDBTreeView.CheckBrowse: Boolean;
begin
  Result := CheckDataSet and (FDataSet.State in [dsBrowse]);
end;

procedure TrCustomDBTreeView.CheckNotCircular;
begin

end;

procedure TrCustomDBTreeView.SyncDataSet;
var
  sp: SmallInt;
begin
  if HandleAllocated and CheckActive then begin
    if {FAutoMove and}
    (FCurrentRecord <> FLinkData.ActiveRecord)
    then begin
      Selected := FTreeData.TreeNode[FLinkData.ActiveRecord];
      if Assigned(Selected)// and not Selected.IsVisible
      then begin
        sp := ScrollPos;
        Selected.MakeVisible;
        ScrollTo(sp);
//        Invalidate;
      end;
    end;
//    if FAutoMove then
    SyncData(Selected);
    FCurrentRecord := FLinkData.ActiveRecord;
  end;
end;

procedure TrCustomDBTreeView.SyncData(Node: TTreeNode);
var
  I, J: Integer;
  V: Variant;
  S: String;
begin
  if CheckActive
  then begin
    if (Node <> nil) then begin
//      I := FTreeData.IndexOf(FTreeData.FStrings.Objects[Node.AbsoluteIndex]);
      I := FTreeData.IndexOfNode(Node);//(FTreeData.FStrings.Objects[Node.AbsoluteIndex]);
      if not (I < 0)
      then begin
        J := I - FLinkData.ActiveRecord;
//        FLinkData.ActiveRecord := I;
        if (J <> 0) then FLinkData.DataSet.MoveBy(J);
//        Exit;
      end;
      if (FLinkData.ActiveRecord <> I) and (I >= 0)
      then FLinkData.ActiveRecord := I;
      FCurrentRecord := FLinkData.ActiveRecord;
      FTreeData.Index := FLinkData.ActiveRecord;
      if I >= 0
        then begin
          V := FKeyField.Value;
          S := FKeyField.AsString;
          if S <> FValue
          then SetValueKey(V, S);
        end
        else SetValueKey(null,'');
    end
    else SetValueKey(null,'');
  end;
{
  if CheckActive and (Node <> nil)
  then begin
    I := -1;
    if (Node <> nil)
    then I := FTreeData.IndexOf(Node.Data);
    if not (I < 0)
    then begin
      I := I - FLinkData.ActiveRecord;
      FLinkData.DataSet.MoveBy(I);
      I := FLinkData.ActiveRecord;
    end;
    FTreeData.Index := I;
    if I >= 0
      then begin
        V := FKeyField.Value;
        S := FKeyField.AsString;
        SetValueKey(V, S);
      end
      else SetValueKey(null,'');
  end;
}
end;

procedure TrCustomDBTreeView.ClearNodes;
begin
  if not (csDestroying in ComponentState)
  and HandleAllocated and (Items.Count > 0)
  then Items.BeginUpdate;
  if HandleAllocated then LockWindowUpdate(Handle);
  try
    Items.Clear;
    FTreeData.Clear;
  finally
    if HandleAllocated then LockWindowUpdate(0);
    if not (csDestroying in ComponentState)
    then Items.EndUpdate;
  end;
end;

procedure TrCustomDBTreeView.CreateEntryData;
var
  I, LastRecord: Integer;
  T: TTreeDataEntry;
  P: PTreeDataEntry;
begin
{
  if not ((FKeyField = nil) or (FChildrenField = nil) or
     (FParentField = nil) or (FListFields.Count = 0))
}
  if FListActive
  then try
    FHasCreated     := True;
    FRefreshEnabled := False;
//    FDataSet.DisableControls;
    LastRecord := FLinkData.ActiveRecord;
    try
      //    (Expanded)     
      if HandleAllocated {$IFDEF POLARIS_D4}and not AutoExpand{$ENDIF} then begin
        if Assigned(FLastSelected) then begin
          FLastSelected.Free;
          FLastSelected := nil;
        end;
        FLastSelected := TVariantList.Create;
        for I := Items.Count-1 downto 0 do
          if Items[I].Expanded
          then FLastSelected.Add(PTreeDataEntry(Items[I].Data)^.KeyValue);
      end;
//      FTreeData.Free;
//      FTreeData := TTreeDataList.Create;
      FTreeData.Clear;
      UpdateRecordCount;
      InitFields(False);
      for I := 0 to FLinkData.BufferCount-1{FRecordCount-1} do begin
        FLinkData.ActiveRecord := I;
        with T do begin
          KeyValue    := FKeyField.Value;
          ChildValue  := FChildrenField.Value;
          ParentValue := FParentField.Value;
          Node        := nil;
          Images.ParentIndex := ParentIndex;
          Images.ParentSelectedIndex := ParentSelected;
          Images.ChildIndex := ChildIndex;
          Images.ChildSelectedIndex := ChildSelected;
        end;
        P := FTreeData.AddEntry(T, DisplayString);
        if Assigned(FOnGetImageIndexes)
        then FOnGetImageIndexes(Self, FDataSet, FTreeData, P);
      end;
      FTreeData.EndOfCreate;
    finally
      FLinkData.ActiveRecord := LastRecord;
//      FDataSet.EnableControls;
    end;
  finally
    FHasCreated     := False;
    FRefreshEnabled := True;
  end;
end;

procedure TrCustomDBTreeView.CreatedNodes;
var
  L, I, sp{$IFDEF POLARIS_D4}, lp{$ENDIF}: Integer;
  TN: TTreeNode;
  LST: TSortType;
begin
  if FListActive and HandleAllocated
  then begin
    FHasCreated     := True;
    FRefreshEnabled := False;
    LST := SortType;
    LockWindowUpdate(Handle);
{$IFDEF POLARIS_D4}
    lp := ChangeDelay;
{$ENDIF}
    sp := ScrollPos;
    try
{$IFDEF POLARIS_D4}
      ChangeDelay := 0;
{$ENDIF}
      L := FLinkData.ActiveRecord;
      FTreeData.BuildTreeNode(TTreeView(Self), FKeyValue);
      if Assigned(FLastSelected) then begin
        while FLastSelected.Count <> 0 do begin
          I := FTreeData.IndexOfKey(FLastSelected.Items[0]);
          TN := nil;
          if I <> -1
          then TN := FTreeData.TreeNode[I];
          if Assigned(TN)
          then TN.Expand(False);
          FLastSelected.Delete(0);
        end;
        FLastSelected.Free;
        FLastSelected := nil;
      end;
      FLinkData.ActiveRecord := L;
      I := FTreeData.IndexOfKey(FKeyValue);
      if (I <> -1) and not Assigned(Selected)
      then Selected := FTreeData.TreeNode[I];

      if not FLookupControl then SyncDataSet;
    finally
      FHasCreated     := False;
      FRefreshEnabled := True;
      SortType        := LST;
      if Assigned(Selected) and not Selected.IsVisible
      then begin
        Selected.MakeVisible;
        ScrollTo(sp);
      end;
{$IFDEF POLARIS_D4}
      ChangeDelay     := lp;
{$ENDIF}
      LockWindowUpdate(0);
      Invalidate;
    end;
  end;
end;

procedure TrCustomDBTreeView.RefreshNodes;
begin
  if CheckActive then begin
    CreateEntryData;
    CreatedNodes;
  end;
end;

{
procedure TrCustomDBTreeView.ChangeDisplayNames;
var
  I, J, K: Integer;
begin
  FHasCreated := True;
  try
    K := FDataLink.ActiveRecord;
    for I := 0 to Items.Count-1 do begin
      J := FTreeData.IndexOf(Items[I].Data);
      if J >= 0 then begin
        FDataLink.ActiveRecord := J;
        Items[I].Text := DisplayString;
      end;
    end;
    FDataLink.ActiveRecord := K;
  finally
    FHasCreated := False;
  end;
end;
}

procedure TrCustomDBTreeView.InitFields(ShowError: Boolean);
var
  Pos: Integer;
  Dfn,
  Fld: String;
begin
  FListActive    := False;
  FKeyField      := nil;
  FChildrenField := nil;
  FParentField   := nil;
  FListFields.Clear;
  if not CheckActive
  or (FChildrenFieldName = EmptyStr)
  or (FParentFieldName = EmptyStr) then Exit;
  with FDataSet do begin
    FChildrenField := FindField(FChildrenFieldName);
    if (FChildrenField = nil) and ShowError then
      raise EInvalidOperation.Create(Format(SFieldNotFound, [Self.Name, FChildrenFieldName]))
    else if FChildrenField <> nil
      then begin
        FChildrenField.FreeNotification(Self);
        FParentField := FindField(FParentFieldName);
        if (FParentField = nil) and ShowError then
          raise EInvalidOperation.Create(Format(SFieldNotFound, [Self.Name, FParentFieldName]))
        else if FParentField <> nil then begin
          FParentField.FreeNotification(Self);
          if FKeyFieldName <> EmptyStr
          then begin
            FKeyField := FindField(FKeyFieldName);
            if (FKeyField = nil) and ShowError then
              raise EInvalidOperation.Create(Format(SFieldNotFound, [Self.Name, FKeyFieldName]))
          end;
          if FKeyField = nil then FKeyField := FParentField
          else FKeyField.FreeNotification(Self);

          Pos := 1;
          if FDisplayFields = EmptyStr then Dfn := FKeyFieldName
          else Dfn := FDisplayFields;
          while Pos < Length(Dfn) do begin
            Fld := ExtractFieldName(Dfn, Pos);
            if (FindField(Fld)=nil)
            and ShowError then
              raise EInvalidOperation.Create(Format(SFieldNotFound, [Self.Name, Fld]))
          end;
          GetFieldList(FListFields, Dfn);
//          FListActive := True;
          FListActive := Assigned(FKeyField) and
           Assigned(FChildrenField) and
           Assigned(FParentField) and
           (FListFields.Count = WordCount(Dfn,[';']))
        end
      end
  end;
end;

procedure TrCustomDBTreeView.NewLayout;
begin
  InitFields(True);
  if FListActive
  and (FLinkData.ActiveRecord <> TreeData.Index)
  and (TreeData.Index <> -1)
  then FLinkData.ActiveRecord := TreeData.Index;

//  LayoutChange;
{
  if FListActive
  then RefreshNodes;
}
//  ChangeDisplayNames;
//  DataChange;
end;

function TrCustomDBTreeView.GetLinkDataSource: TDataSource;
begin
  Result := FLinkData.DataSource;
end;

procedure TrCustomDBTreeView.SetLinkDataSource(AValue: TDataSource);
begin
  if FLinkData.DataSource <> AValue then begin
    ClearNodes;
    FLinkData.DataSource := AValue;
    if AValue <> nil then AValue.FreeNotification(Self);
  end;
end;

procedure TrCustomDBTreeView.SetKeyFieldName(AValue: String);
begin
  if ANSICompareText(AValue, FKeyFieldName) <> 0
  then begin
    FKeyFieldName := AValue;
    NewLayout;
  end;
end;

procedure TrCustomDBTreeView.SetChildrenFieldName(AValue: String);
begin
  if ANSICompareText(AValue, FParentFieldName) = 0
  then raise Exception.Create(trdbtreeEqualsField);
  if ANSICompareText(AValue, FChildrenFieldName) <> 0
  then begin
    FChildrenFieldName := AValue;
    NewLayout;
  end;
end;

procedure TrCustomDBTreeView.SetParentFieldName(AValue: String);
begin
  if ANSICompareText(AValue, FChildrenFieldName) = 0
  then raise Exception.Create(trdbtreeEqualsField);
  if ANSICompareText(AValue, FParentFieldName) <> 0
  then begin
    FParentFieldName := AValue;
    NewLayout;
  end;
end;

procedure TrCustomDBTreeView.SetDisplayFields(AValue: String);
begin
  if ANSICompareText(AValue, FDisplayFields) <> 0
  then begin
    FDisplayFields := AValue;
    NewLayout;
  end;
end;

function TrCustomDBTreeView.GetListField(Index: Integer): TField;
begin
  Result := nil;
  if (ListFieldCount > 0) and (Index <= ListFieldCount)
  then Result := TField(FListFields.Items[Index])
end;

function TrCustomDBTreeView.GetListFieldCount: Integer;
begin
  Result := FListFields.Count;
end;

function TrCustomDBTreeView.DisplayString: String;
var
  I: Integer;
  S: String;
  C: Char;
begin
  Result := EmptyStr;
  if FListFields.Count > 0
  then for I := 0 to FListFields.Count-1 do begin
    S := TField(FListFields.Items[I]).DisplayText;
    if Result = EmptyStr
    then Result := S
    else Result := Result+' '+S;
  end
  else if FKeyField <> nil
    then Result := FKeyField.DisplayText;
  if Assigned(FOnGetText)
  then FOnGetText(Self, FDataSet, Result);
  for C := #0 to #31 do begin
    S := C;
    Result := ReplaceStr(Result, S, '');
  end;
end;

procedure TrCustomDBTreeView.Notification(AComponent: TComponent; Operation: TOperation);
var
  I: Integer;
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove)
  then begin
    if (FLinkData <> nil) and (AComponent = LinkDataSource)
    then begin
      ClearNodes;
      LinkDataSource := nil;
      // BeginLayout NeedLayout := true
      if FKeyField = AComponent then FKeyField := nil;
      if FParentField = AComponent then FParentField := nil;
      if FChildrenField = AComponent then FChildrenField := nil;
      I := FListFields.IndexOf(AComponent);
      if I <> -1 then FListFields.Delete(I);
      //
    end;

  end;
end;

procedure TrCustomDBTreeView.SetValueKey(const AValue: Variant; const AString: String);
var
  PValue,
  S: String;
begin
  PValue := FValue;
  if CompareVar(FKeyValue, AValue) <> 0
  then begin
    FKeyValue := AValue;
    S := Trim(AString);
    if VarIsEmpty(AValue)
    then S := 'EMPTY'
    else if VarIsNull(AValue)
      then S := 'NULL';
    FValue    := AString;
    SelectKeyValue(FKeyValue);
  end;
// change()
  if PValue <> FValue
  then KeyValueChanged;
end;

procedure TrCustomDBTreeView.SelectKeyValue(const AValue: Variant);
begin
end;

procedure TrCustomDBTreeView.SetValue(AValue: String);
var
  V: Variant;
begin
  if (AValue <> FValue) or (AValue=EmptyStr)
  then begin
    AValue  := Trim(AValue);
    FValue  := AValue;
    if (AnsiCompareText(AValue,'NULL')=0) or (FValue=EmptyStr)
    then V := Null
    else if AnsiCompareText(AValue,'EMPTY')=0
      then VarClear(V)
      else V := AValue;
    if ANSICompareText(AValue, 'TRUE') = 0
    then KeyValue := True
    else if ANSICompareText(AValue,'FALSE') = 0
      then KeyValue := False
      else KeyValue := V;
  end;
end;

procedure TrCustomDBTreeView.SetNotSelectParent(Value: Boolean);
begin
  if Value <> FNotSelectParent then begin
    FNotSelectParent := Value;
    if FNotSelectParent and not (csDesigning in ComponentState)
    then Selected := GetChildNode(Selected, 0);
  end;
end;

function TrCustomDBTreeView.GetBuildMode: TrTreeBuildMode;
begin
  Result := FTreeData.BuildMode;
end;

procedure TrCustomDBTreeView.SetBuildMode(Value: TrTreeBuildMode);
begin
  FTreeData.BuildMode := Value;
end;

procedure TrCustomDBTreeView.SetKeyValue(AValue: Variant);
var
  I{$IFDEF POLARIS_D4}, L{$ENDIF}: Integer;
  N: TTreeNode;
begin
{
  if CompareVar(AValue, FKeyValue) <> 0
  then begin
}
    if FLinkData.Active
    then begin
      if not FHasCreated and FAutoSelect then SetValueKey(AValue, VarToStr(AValue));
      I := FTreeData.IndexOfKey(AValue);
      if I < 0
      then Selected := nil
      else begin
        N := FTreeData.TreeNode[I];
        if Assigned(N)
        then begin
{$IFDEF POLARIS_D4}
          L := ChangeDelay;
{$ENDIF}
          try
{$IFDEF POLARIS_D4}
            ChangeDelay := 0;
{$ENDIF}
            if Selected <> N
            then Selected := N
            else SyncData(Selected)
          finally
{$IFDEF POLARIS_D4}
            ChangeDelay := L;
{$ENDIF}
          end;
        end
        else begin
//          FLinkData.ActiveRecord := I;
          SetValueKey(AValue, VarToStr(AValue));
//          Change(nil);
        end
      end
    end
    else SetValueKey(AValue, VarToStr(AValue));
{
  end;
}
end;

{ TrDBTreeView }

{
function TrDBTreeView.GetDataSource: TDataSource;
begin
  Result := LinkDataSource;
end;

procedure TrDBTreeView.SetDataSource(Value: TDataSource);
begin
  LinkDataSource := Value;
end;
}
procedure TrCustomDBTreeView.SetCheckListChange(
  const Value: TTreeCheckListChange);
begin
  FCheckListChange := Value;
  if Assigned(FCheckList) then FCheckList.FCheckListChange := FCheckListChange;
end;

procedure TrCustomDBTreeView.SetCheckListChanging(
  const Value: TTreeCheckListChangingEvent);
begin
  FCheckListChanging := Value;
  if Assigned(FCheckList) then FCheckList.FCheckListChanging := FCheckListChanging;
end;

procedure TrCustomDBTreeView.WMChar(var Message: TWMChar);
{var
  sc: TShortCut;}
begin
  {sc := ShortCut(Message.CharCode, KeyDataToShiftState(Message.KeyData));
  if IsMultiSelect and (sc = FCheckKey) then ChangeCheckNode(Selected)
  else inherited;}
end;

procedure TrCustomDBTreeView.KeyDown(var Key: Word; Shift: TShiftState);
var
  AddAfter, AChecked: Boolean;
  ANode: TTreeNode;

  function GetNextNode(Key: Word): TTreeNode;
  begin
    if Key = VK_DOWN then Result := Selected.getNext{Sibling}
    else Result := Selected.getPrev{Sibling};
  end;

begin
  if IsMultiSelect and (FCheckKey = ShortCut(Key, Shift)) then ChangeCheckNode(Selected)
  else begin
    AddAfter := False;
    AChecked := Assigned(Selected) and (FCheckShift = Shift) and (Key in [VK_DOWN, VK_UP]);
    if AChecked then begin
      if not FSelecting then begin
        Direction := Key;
        FSelectionAnchor := Selected;
        ChangeCheckNode(FSelectionAnchor);
        FSelecting := True;
        AddAfter := True;
      end
      else begin
        AddAfter := Key = Direction;
        if not AddAfter then begin
          ChangeCheckNode(Selected);
          ANode := GetNextNode(Key);
          if Assigned(ANode) and (ANode = FSelectionAnchor) then begin
            AddAfter := True;
            FSelecting := False;
          end;
        end;
      end;
    end
    else FSelecting := False;
    inherited;
    if AChecked and AddAfter then begin
      ANode := GetNextNode(Key);
      if Assigned(ANode) then ChangeCheckNode(ANode);
    end;
  end;
end;

procedure TrCustomDBTreeView.DoCheck;
begin
  if Selected <> nil then ChangeCheckNode(Selected);
end;

procedure TrCustomDBTreeView.CheckAll;
var
  i: Integer;
begin
  for i := 0 to Items.Count - 1 do begin
    Items.BeginUpdate;
    try
      with TrDBTreeNode(Items[i]) do
        if not (trncsChecked in FCheckState) then SetCheckState(True);
    finally
      Items.EndUpdate;
      Invalidate;
    end;
  end;
end;

procedure TrCustomDBTreeView.MouseDown(Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  inherited;
  (*if (Button = mbRight) {and RightClickSelect} then begin
    //Selected := Selected;
    if CanChange(Selected) then Change(Selected);
  end;*)
end;

procedure TrCustomDBTreeView.AfterCheckChange(Item: Variant;
  Action: TVarListAction);
var
  ANode: TrDBTreeNode;
  I: Integer;
begin
  if not FInternalCheck and FLinkData.Active then
  begin
    I := FTreeData.IndexOfKey(Item);
    if I >= 0 then begin
      ANode := TrDBTreeNode(FTreeData.TreeNode[I]);
      if Assigned(ANode) then begin
        if Action = vlnAdd then
          ANode.FCheckState := ANode.FCheckState + [trncsChecked] - [trncsUnChecked]
        else
          ANode.FCheckState := ANode.FCheckState - [trncsChecked] + [trncsUnChecked];
        ANode.ChangeParentState;
        Invalidate;
      end;
    end;
  end;
end;

function TrCustomDBTreeView.GetNodeByID(Item: Variant): TTreeNode;
var
  i: Integer;
begin
  I := FTreeData.IndexOfKey(Item);
  if I >= 0 then
    Result := TrDBTreeNode(FTreeData.TreeNode[I])
  else Result := nil;
end;

{ TrDataSourceLink }

procedure TrDataSourceLink.ActiveChanged;
begin
  if FControl <> nil then FControl.DataLinkActiveChanged;
end;

procedure TrDataSourceLink.LayoutChanged;
begin
  if FControl <> nil then FControl.CheckDataLinkActiveChanged;
end;

procedure TrDataSourceLink.RecordChanged(Field: TField);
begin
  if FControl <> nil then FControl.DataLinkRecordChanged(Field);
end;

procedure TrDataSourceLink.DataSetChanged;
begin
  if FControl <> nil then FControl.DataLinkDataSetChanged;
end;

procedure TrDataSourceLink.FocusControl(Field: TFieldRef);
begin
  if (Field^ <> nil) and (FControl <> nil) and
    (Field^ = FControl.FDataField) and FControl.CanFocus then
  begin
    Field^ := nil;
    FControl.SetFocus;
  end;
end;

{???}
procedure TrDataSourceLink.EditingChanged;
begin
  if FControl <> nil then FControl.DataLinkEditChanged;
end;

{ TrDBLookupTree }

constructor TrDBLookupTree.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPrevValue     := null;
  FReadOnly      := False;
  FLookupControl := True;
  FDataField     := nil;
  FMasterField   := nil;
  FHasEdit       := False;
  FDataLink      := TrDataSourceLink.Create;
  FDataLink.FControl := Self;
end;

destructor TrDBLookupTree.Destroy;
begin
  FDataLink.FControl := nil;
  FDataLink.Free;
  inherited Destroy;
end;

procedure TrDBLookupTree.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then begin
    if (FDataLink <> nil) and (AComponent = DataSource) then DataSource := nil;
  end;
end;

procedure TrDBLookupTree.CheckNotCircular;
begin
  if FDataLink.Active and ((DataSource = LookupSource)
     or (FDataLink.DataSet = FLinkData.DataSet))
  then _DBError(SCircularDataLink);
end;

procedure TrDBLookupTree.DataLinkActiveChanged;
begin
  FDataField   := nil;
  FMasterField := nil;
  if FDataLink.Active and (FDataFieldName <> EmptyStr) then begin
    CheckNotCircular;
    FDataField := FDataLink.DataSet.FieldByName(FDataFieldName);
    FMasterField := FDataField;
  end;
  DataLinkRecordChanged(nil);
end;

procedure TrDBLookupTree.DataLinkDataSetChanged;
var
  I: Integer;
  TN: TTreeNode;
begin
  if FDataLink.Active and
     Assigned(FDataField) and
     Assigned(FMasterField)
  then begin
    FPrevValue := FDataField.Value;
    I := FTreeData.IndexOfKey(FDataField.Value);
    TN := nil;
    if I <> -1
    then TN := FTreeData.TreeNode[I];
    if Selected <> TN then begin
      Selected := TN;
      //KeyValue := FDataField.Value;
    end;

{
    I := FTreeData.IndexOfKey(FDataField.Value);
    TN := nil;
    if I <> -1
    then TN := FTreeData.TreeNode[I];
    Selected := TN;
}
//    SyncData(TN);
  end;
end;

procedure TrDBLookupTree.Change(Node: TTreeNode);
begin
  inherited Change(Node);
end;

procedure TrDBLookupTree.DataLinkRecordChanged(Field: TField);
begin
  if (Field = nil) or (Field = FMasterField) then begin
    if (FMasterField <> nil) then begin
      if FHasEdit and (CompareVar(FMasterField.Value, FPrevValue)=0) then Exit;
      Value := FMasterField.AsString;
      if FHasEdit and (CompareVar(FMasterField.Value, FPrevValue)<>0)
      then FPrevValue := FMasterField.Value;
    end// else Value := '';
  end;
end;

procedure TrDBLookupTree.CheckDataLinkActiveChanged;
var
  TestField: TField;
begin
  if FDataLink.Active and (FDataFieldName <> EmptyStr) then begin
    TestField := FDataLink.DataSet.FieldByName(FDataFieldName);
    if Pointer(FDataField) <> Pointer(TestField) then begin
      FDataField := nil;
      FMasterField := nil;
      CheckNotCircular;
      FDataField := TestField;
      FMasterField := FDataField;
      DataLinkRecordChanged(nil);
    end;
  end;
end;

{???}
procedure TrDBLookupTree.DataLinkEditChanged;
begin
  FHasEdit := FDataLink.DataSet.State in [dsInsert, dsEdit];
  if FHasEdit and (FMasterField <> nil)
  then FPrevValue := FMasterField.Value
  else FPrevValue := Null;
//  else VarClear(FPrevValue);
end;

{***}
procedure TrDBLookupTree.SetReadOnly(AValue: Boolean);
begin
  FReadOnly := AValue;
end;

procedure TrDBLookupTree.SetDataFieldName(AValue: String);
begin
  if FDataFieldName <> AValue then begin
    FDataFieldName := AValue;
    DataLinkActiveChanged;
  end;
end;

function TrDBLookupTree.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TrDBLookupTree.SetDataSource(AValue: TDataSource);
begin
  if FDataLink.DataSource <> AValue then begin
    FDataLink.DataSource := AValue;
    if AValue <> nil then AValue.FreeNotification(Self);
  end;
end;

function TrDBLookupTree.CanModify: Boolean;
begin
  Result := FListActive and not FMasterField.ReadOnly and ((FDataLink.DataSource = nil) or
    (FMasterField <> nil) and FMasterField.CanModify);
end;

procedure TrDBLookupTree.SetKeyValue(AValue: Variant);
begin
  if FMasterField <> nil
  then begin
    FKeyValue := AValue;
    if CompareVar(FMasterField.Value, AValue) <> 0 then
    inherited SetKeyValue(AValue);
  end
  else inherited SetKeyValue(AValue);
end;

procedure TrDBLookupTree.SelectKeyValue(const AValue: Variant);
begin
  if FMasterField <> nil then begin
    if (VarType(AValue) = varUnknown) or (CompareVar(FMasterField.Value, AValue)=0)
{
    then begin
       SetKeyValue(AValue);
       Exit;
    end;
}
    then Exit;
    if CanModify then begin
      if FDataField = FMasterField
      then FDataField.DataSet.Edit;
      FHasEdit := True;
      FMasterField.Value := AValue;
      FHasEdit := False;
    end
//    else Exit;
  end;
//  else SetValueKey(AValue);
end;

{ TrPopupTree }

type
  TLocDBLookupTree = class(TrCustomDBTreeView)
  private
    FRowCount: Integer;
    procedure CreatedNodes; override;
//    function GetDisplayText: String;
    procedure SetRowCount(AValue: Integer);
    procedure CMEnableChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure WMMouseActivate(var Message: TMessage); message WM_MOUSEACTIVATE;
    procedure CNNotify(var Message: TWMNotify); message CN_Notify;

  protected
//    procedure DataChanged; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure CreateParams(var Params: TCreateParams); override;
    function CalcHeight: Integer;

    procedure ActiveChanged; override; //dynamic;

    procedure DestroyWnd; override;

    function TestChildEnabled: Boolean; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;

{$IFDEF POLARIS_D4}
    procedure CustomDrawParam(Node: TTreeNode; State: TCustomDrawState;
      Stage: TCustomDrawStage); override;
{$ENDIF}

//    procedure SetValueKey(const AValue: Variant; const AString: String); override; //dynamic;

    procedure AfterCheckChange(Item: Variant; Action: TVarListAction); override;
  public
    constructor Create(AOwner: TComponent); override;
//    property DisplayText: String read GetDisplayText;
    property RowCount:Integer read FRowCount write SetRowCount;
    property SelectEnabled: Boolean read FSelectedEnabled;
  end;

constructor TLocDBLookupTree.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
//  ControlStyle := [csCaptureMouse, csClickEvents, csDoubleClicks];
  ControlStyle := [csClickEvents, csDoubleClicks];
  FLookupControl := True;
//  ChangeDelay := 200;
  FOnlyChild := False;

  FAutoMove  := False;

  HideSelection := False;
  Ctl3D        := False;
  Enabled      := False;
  BorderStyle  := bsNone;
  TabStop      := False;
  FRowCount    := 7;
  Height       := CalcHeight;
end;

procedure TLocDBLookupTree.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited;
  if Key in [VK_DOWN, VK_UP, VK_HOME, VK_END] then
    TrPopupTree(Parent).SearchString := EmptyStr;  
end;

procedure TLocDBLookupTree.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
    Style := Style and not (WS_BORDER or WS_TABSTOP or WS_DISABLED or WS_POPUP);
  Params.ExStyle := Params.ExStyle or WS_EX_TOOLWINDOW;
end;

procedure TLocDBLookupTree.DestroyWnd;
begin
  inherited;
end;

function TLocDBLookupTree.TestChildEnabled: Boolean;
begin
  Result := {Focused and }FNotSelectParent and not FHasCreated
    and not TrDBLookupComboTree(TrPopupTree(Parent).Parent).FOnDataChange
  ;
end;

procedure TLocDBLookupTree.WMMouseActivate(var Message: TMessage);
begin
  Message.Result := MA_NOACTIVATE;
end;

procedure TLocDBLookupTree.CNNotify(var Message: TWMNotify);
var
  Node: TTreeNode;
begin
  inherited;
  if HandleAllocated then
  with Message, NMHdr^ do
    if (Code = TVN_GETDISPINFO) and
       ((PTVDispInfo(Message.NMHdr)^.item.state and TVIS_SELECTED) <> 0)
    then begin
      Node := Items.GetNode(PTVDispInfo(Message.NMHdr)^.item.hItem);
      if (Node <> nil) and ((PTVDispInfo(Message.NMHdr)^.item.mask and TVIF_TEXT) <> 0)
      then TrPopupTree(Parent).Perform( RCT_LOOKUPCHANGE, Integer(Node.ItemId), 0)
    end;
end;

procedure TLocDBLookupTree.CreatedNodes;
//var
//  OldValue: Variant;
begin
  TrPopupTree(Parent).FIsCreateTree := False;
{  if not HandleAllocated
  then begin
    RecreateWnd;
  end
  else}
{
  TrPopupTree(Parent).HandleNeeded;
  TrPopupTree(Parent).FTreeList.HandleNeeded;
}
  inherited CreatedNodes;
  TrPopupTree(Parent).FIsCreateTree := True;
  TrPopupTree(Parent).EndCreated;
end;

procedure TLocDBLookupTree.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  try
    FSelectedEnabled := htOnItem in GetHitTestInfoAt(X,Y);
    inherited MouseDown(Button, Shift, X, Y);
    if htOnItem in GetHitTestInfoAt(X,Y)
    then begin
      if not (FNoCloseUp and FNotSelectParent)
      then TrDBLookupComboTree(Parent.Parent).PopupCloseUp(Parent.Parent, True);
  end;
  finally
    FNoCloseUp := False;
  end;
end;

procedure TLocDBLookupTree.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  FSelectedEnabled := False;
end;

procedure TLocDBLookupTree.SetRowCount(AValue: Integer);
begin
  if FRowCount <> AValue
  then begin
    FRowCount := AValue;
    if FRowCount < 1 then FRowCount := 1;
    if FRowCount > 50 then FRowCount := 50;
    Height := CalcHeight;
  end;
end;

function TLocDBLookupTree.CalcHeight: Integer;
var
  DC: HDC;
  SaveFont: HFont;
  SysMetrics, Metrics: TTextMetric;
  CH: Integer;
begin
  DC := GetDC(0);
  GetTextMetrics(DC, SysMetrics);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  CH := SysMetrics.tmHeight;
  Result := 0;   // ??????
  if (Images <> nil) and (Result < Images.Height) then CH := Images.Height;
  Result := CH*FRowCount+GetSystemMetrics(SM_CXHSCROLL)+2;
end;

procedure TLocDBLookupTree.ActiveChanged;
begin
  inherited;
  TrPopupTree(Parent).Perform( RCT_LOOKUPCHANGE, 0, 0);
end;

procedure TLocDBLookupTree.CMEnableChanged(var Message: TMessage);
begin
  if HandleAllocated and not (csDesigning in ComponentState) then
    EnableWindow(Handle, True);
end;

procedure TLocDBLookupTree.AfterCheckChange(Item: Variant;
  Action: TVarListAction);
begin
  inherited;
  if not FInternalCheck and FLinkData.Active then
    TrPopupTree(Parent).Perform(RCT_LOOKUPCHANGE, 0, 0);
end;

{$IFDEF POLARIS_D4}

procedure TLocDBLookupTree.CustomDrawParam(Node: TTreeNode; State: TCustomDrawState;
      Stage: TCustomDrawStage);
var
  cl: TColor;

begin
  cl := Canvas.Font.Color;
  Canvas.Font := Font;
  if (Stage = cdPrePaint) and (cdsSelected in State) then begin
    Canvas.Brush.Color := clHighlight;
    cl := clHighlightText;
    SetNodeState(Node, 0);
  end;
  Canvas.Font.Color := cl;
end;
{$ENDIF}

type
  TLocLabel = class(TStaticText)
  private
    procedure WMMouseActivate(var Message: TMessage); message WM_MOUSEACTIVATE;
    procedure CMEnableChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
  public
    constructor Create(AOwner: TComponent); override;
  end;

constructor TLocLabel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
//  ControlStyle := [csCaptureMouse, csClickEvents, csDoubleClicks];
  ControlStyle := [csClickEvents, csDoubleClicks];
  Ctl3D        := False;
  Enabled      := False;
end;


procedure TLocLabel.WMMouseActivate(var Message: TMessage);
begin
  Message.Result := MA_NOACTIVATE;
end;

procedure TLocLabel.CMEnableChanged(var Message: TMessage);
begin
  if HandleAllocated and not (csDesigning in ComponentState) then
    EnableWindow(Handle, True);
end;

procedure TLocLabel.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
    Style := Style and not (WS_BORDER or WS_TABSTOP or WS_DISABLED or WS_POPUP);
end;

{ TrPopupTree }

constructor TrPopupTree.Create(AOwner: TComponent);
var
  Control: TPanel;
begin
  inherited Create(AOwner);
  FIsCreateTree := False;
  ControlStyle := [csOpaque];
  Parent := TWinControl(AOwner);
  FIValue := null;
  FPrevValue := null;
  SNoneLabel := EmptyStr;
  FDropDownRows := 8;
  Width  := 300;
  Color  := clBtnFace;
  FTreeList := nil;
  FontSetDefault(Font);
  OnMouseUp := TreeMouseUp;
  if AOwner is TControl then ShowHint := TControl(AOwner).ShowHint
  else ShowHint := True;
  if (csDesigning in ComponentState) then Exit;
  Control := TPanel.Create(Self);
  with Control as TPanel do begin
    Parent := Self;
    Align := alTop;
    ParentColor := True;
    Width := Self.Width;
//    AutoSize := True;
    BevelInner := bvNone;
    BevelOuter := bvNone;
    Ctl3D  := False;
    ParentColor := True;
    ControlStyle := ControlStyle + [csReplicatable];
  end;
  FTreeList := TLocDBLookupTree.Create(Self);
  with FTreeList as TLocDBLookupTree do begin
{$IFDEF POLARIS_D4}
    AutoExpand := True;
{$ENDIF}
    Parent := Self;
    Align  := alClient;
    RowCount := FDropDownRows-1;
{$IFDEF POLARIS_D4}
    Control.Height := TreeView_GetItemHeight(FTreeList.Handle);
{$ELSE}
    Control.Height := 16;
{$ENDIF}
    Self.Height := Control.Height + Height;
    Ctl3D  := False;
    AutoSelectItem := False;
    OnChange := TreeChange;
    OnMouseUp := TreeMouseUp;
    ControlStyle := ControlStyle + [csReplicatable];
  end;
  FNoneLabel := TLocLabel.Create(Self);
  with FNoneLabel do begin
    Parent := Control;
    Color := FTreeList.Color;
    Align := alClient;
    Caption := SNoneLabel;
    OnClick := LabelClick;
    ControlStyle := ControlStyle + [csReplicatable];
    Visible := SNoneLabel <> EmptyStr;
  end;
{$IFDEF POLARIS_D4}
  Height := TreeView_GetItemHeight(FTreeList.Handle)*FDropDownRows+2
{$ELSE}
  Height := 16;
{$ENDIF}
end;

procedure TrPopupTree.CreateParams(var Params: TCreateParams);
begin
  inherited;
{
  Params.Style := Params.Style or WS_POPUPWINDOW or WS_SIZEBOX
      or SBS_SIZEGRIP;
}
end;

procedure TrPopupTree.CreateWnd;
begin
  inherited;
//  FTreeList.CreateWnd;
  FTreeList.HandleNeeded;
end;

procedure TrPopupTree.DestroyWnd;
begin
  if FTreeList.HandleAllocated
  then try
    FTreeList.FDestroingWindow := True;
    FTreeList.Selected := nil;
  finally
    FTreeList.FDestroingWindow := False;
  end;
  inherited;
end;

procedure TrPopupTree.SetNoneLabel(AValue: String);
begin
  if SNoneLabel <> Trim(AValue)
  then begin
    SNoneLabel := Trim(AValue);
    FNoneLabel.Caption := SNoneLabel;
    FNoneLabel.Visible := SNoneLabel <> EmptyStr;
  end;
end;

procedure TrPopupTree.EndCreated;
begin
  TrDBLookupComboTree(Parent).DataChange(Parent);
end;

procedure TrPopupTree.WMKeyDown(var Message: TWMKeyDown);
var
  M: TMessage;
  Key: Word;
  Shift: TShiftState;
  B: Boolean;
begin
  M := TMessage(Message);
  Shift := KeyDataToShiftState(Message.KeyData);
  Key   := Message.CharCode;
  B := TestKeyShiftState(Shift, [ssShift, ssAlt, ssCtrl]);
  if (Key in [VK_HOME, VK_END{, VK_LEFT, VK_RIGHT}])
  or (not Visible and B and (Key in [VK_UP, VK_DOWN, VK_PRIOR, VK_NEXT])) then Exit;
  if not (TLocDBLookupTree(FTreeList).IsMultiSelect and
    (Shift = TLocDBLookupTree(FTreeList).FCheckShift) and (Key in [VK_UP, VK_DOWN]))
  then inherited;
  if not (Visible or (Key in [VK_UP, VK_DOWN])) then Message.Result := 1;
  if not B
  then case Key of
    VK_PRIOR,
    VK_UP: with TLocDBLookupTree(FTreeList) do
      if (Selected = nil)
      or ((Selected <> nil) and (Selected = Items[0]))
      then begin
        if SNoneLabel <> EmptyStr then Self.SetValue(null);
        Message.Result := 1;
    end;
  end;
  if (Message.Result = 0)
  then
    FTreeList.Perform(M.Msg, M.WParam, M.LParam);
end;

procedure TrPopupTree.RCTLookupChange(var Message: TMessage);
begin
  if Assigned(Owner)
  then TWinControl(Owner).Perform(Message.Msg, Message.WParam, Message.LParam);
end;

procedure TrPopupTree.WMMouseActivate(var Message: TMessage);
begin
  Message.Result := MA_NOACTIVATE;
end;

function TrPopupTree.GetPopupText: String;
begin
  Result := DisplayValue;
end;

procedure TrPopupTree.Show(Origin: TPoint);
var
  Node: TTreeNode;
  sp: Integer;
begin
  inherited Show(Origin);
  with TLocDBLookupTree(FTreeList) do begin
    if Items.Count = 0 then Node := nil
    else if Assigned(Selected) then Node := Selected
      else Node := Items[0];
    if Assigned(Node)
    then begin
      sp := FTreeList.ScrollPos;
      TopItem := Node;
      FTreeList.ScrollTo(sp);
    end;
    TLocDBLookupTree(FTreeList).Update;
    TLocDBLookupTree(FTreeList).Invalidate;
  end;
end;

procedure TrPopupTree.SetDropDownRows(AValue: Integer);
begin
  if AValue <> FDropDownRows then begin
    FDropDownRows := AValue;
    if FDropDownRows < 2 then FDropDownRows := 2;
    if FDropDownRows > 50 then FDropDownRows := 50;
    if not (csDesigning in ComponentState) and (FTreeList <> nil)
    then with FTreeList as TLocDBLookupTree do begin
      RowCount := FDropDownRows-1;
{$IFDEF POLARIS_D4}
      Self.Height := TreeView_GetItemHeight(FTreeList.Handle)*FDropDownRows+2;
{$ELSE}
      Self.Height := 16;
{$ENDIF}
    end;
  end;
end;

procedure TrPopupTree.TreeMouseUp(Sender: TObject; Button: TMouseButton;
                          Shift: TShiftState; X, Y: Integer);
begin
  if (Button = mbLeft) and (Shift = []) then begin
    if htOnItem in TLocDBLookupTree(FTreeList).GetHitTestInfoAt(X,Y) then
      CloseUp(True);
  end;
end;

procedure TrPopupTree.LabelClick(Sender: TObject);
begin
  try
    SetValue(null);
    if not (FTreeList.FNoCloseUp and FTreeList.FNotSelectParent)
    then TrDBLookupComboTree(Parent).PopupCloseUp(Parent, True);
  finally
    FTreeList.FNoCloseUp := False;
  end;
end;

procedure TrPopupTree.TreeChange(Sender: TObject; Node: TTreeNode);
begin
  if SNoneLabel <> EmptyStr then
  if Node = nil
  then begin
{$IFDEF POLARIS_D4}
    FNoneLabel.Color := clHighlight;
    FNoneLabel.Font.Color := clHighlightText;
{$ELSE}
    FNoneLabel.Color := clBtnFace;
{$ENDIF}
  end
  else begin
{$IFDEF POLARIS_D4}
    FNoneLabel.Color := clWindow;
    FNoneLabel.Font.Color := clWindowText;
{$ELSE}
    FNoneLabel.Color := FTreeList.Color;
{$ENDIF}
  end;
end;

procedure TrPopupTree.KeyDown(var Key: Word; Shift: TShiftState);
var
  Delta: ShortInt;
//  B: Boolean;
  I, J, K: Integer;
  P: PTreeDataEntry;
begin
  inherited KeyDown(Key, Shift);
  if FTreeList <> nil
  then begin
    if Visible
    then //FTreeList.KeyDown( Key, Shift)
    else begin
      Delta := 0;
      if not TestKeyShiftState(Shift, [ssShift, ssAlt, ssCtrl])
      then Delta := byte(Key=VK_DOWN) - byte(Key=VK_UP);
      if Delta <> 0 then begin
        SearchString := EmptyStr;
        I := TLocDBLookupTree(FTreeList).TreeData.IndexOfKey(FIValue);
        if I <> -1
        then begin
          P := PTreeDataEntry(TLocDBLookupTree(FTreeList).TreeData.FList[I]);
          I := TLocDBLookupTree(FTreeList).TreeData.FStrings.IndexOfObject(TObject(P));
        end;
        J := I + Delta;
        if (Trim(FNoneLabel.Caption) = EmptyStr) and (J < 0) then J := 0;

        if J < 0
        then SetValue(null)
        else begin
          if J >= TLocDBLookupTree(FTreeList).TreeData.Count
          then J := TLocDBLookupTree(FTreeList).TreeData.Count-1;

          if TLocDBLookupTree(FTreeList).NotSelectParent then begin
            K := J;
            P := PTreeDataEntry(TLocDBLookupTree(FTreeList).TreeData.FStrings.Objects[K]);
            while TLocDBLookupTree(FTreeList).TreeData.HasChildren(P^) do begin
            //   
               if Delta > 0
               then Inc(K)
               else Dec(K);
               if K < 0 then begin
                 K := J;
                 Delta := 1;
               end;
              P := PTreeDataEntry(TLocDBLookupTree(FTreeList).TreeData.FStrings.Objects[K]);
          end;
          J := K;
          end;
          if (J <> I) then begin
            P := PTreeDataEntry(TLocDBLookupTree(FTreeList).TreeData.FStrings.Objects[J]);
            TLocDBLookupTree(FTreeList).TreeData.Index := TLocDBLookupTree(FTreeList).TreeData.FList.IndexOf(P);
            SetValue(P^.KeyValue);
          end;
        end;
      end;
      Key := 0;
    end;
  end;
end;

procedure TrPopupTree.KeyPress(var Key: Char);
begin
  if Key in ['+','-']
  then begin
    if (FTreeList <> nil) and (FTreeList.Selected <> nil) and
       (FTreeList.Selected.HasChildren) and Visible
    then
      case Key of
      '-': if FTreeList.Selected.Expanded then FTreeList.Selected.Collapse(True);
      '+': if not FTreeList.Selected.Expanded then FTreeList.Selected.Expand(True);

      end;
//    Key := #0;
  end;
  inherited KeyPress(Key);
  ProcessSearchKey(Key);
end;

procedure TrPopupTree.ProcessSearchKey(Key: Char);
var
//  TickCount: Longint;
  S: string;
begin
  S := '';
  case Key of
    #9, #27: begin
      SearchString := '';
      SetValue(FPrevValue);
    end;
    Char(VK_BACK), #32..#255:
//      if CanModify then
      begin
{
        if not FPopup then begin
          TickCount := GetTickCount;
          if TickCount - SearchTickCount > 2000 then FSearchText := '';
          SearchTickCount := TickCount;
        end;
}
        if (Key = Char(VK_BACK)) then
          S := Copy(SearchString, 1, Length(SearchString) - 1)
        else if Length(SearchString) < 32 then
          S := SearchString + Key;
        if S = EmptyStr then
          with TLocDBLookupTree(FTreeList) do begin
            Selected := Items.GetFirstNode;
            FSearchString := S;
          end
        else SearchString := S;
      end;
  end;
end;

function TrPopupTree.GetValue: Variant;
begin
  if (csDesigning in ComponentState)
  then Result := null
  else begin
    Result := FIValue;
{
    if (FTreeList <> nil)// and (TLocDBLookupTree(FTreeList).Selected <> nil)
    then Result := TLocDBLookupTree(FTreeList).KeyValue
    else Result := null
}
  end;
end;

procedure TrPopupTree.SetValue(const Value: Variant);
var
  B: Boolean;
begin
  if not FTreeList.HandleAllocated
  then try
    FTreeList.FDestroingWindow := True;
    FTreeList.Selected := nil;
    while not FTreeList.HandleAllocated do begin
      CreateWnd;
      Application.ProcessMessages;
    end;
  finally
    FTreeList.FDestroingWindow := False;
  end;
  if VarIsEmpty(Value) or VarIsNull(Value)
  then begin
//    TLocDBLookupTree(FTreeList).Value := EmptyStr;
    FIValue := null;
//    TLocDBLookupTree(FTreeList).Selected := nil;
{
    if FTreeList.HandleAllocated then begin
      TLocDBLookupTree(FTreeList).KeyValue := Value;
      TLocDBLookupTree(FTreeList).SetValueKey(FIValue, '');
      TreeChange(Self, nil);
    end;
}
    TLocDBLookupTree(FTreeList).KeyValue := Value;
    TLocDBLookupTree(FTreeList).SetValueKey(FIValue, '');
    TreeChange(Self, nil);
//    TLocDBLookupTree(FTreeList).KeyValueChanged;
//    TLocDBLookupTree(FTreeList).KeyValue := Value;
  end
  else begin
    FIValue := Value;
//    if FTreeList.HandleAllocated then begin
      B := TLocDBLookupTree(FTreeList).FAutoMove;
      TLocDBLookupTree(FTreeList).FAutoMove := True;
      TLocDBLookupTree(FTreeList).SetKeyValue(Value);
      TLocDBLookupTree(FTreeList).FAutoMove := B;
//    end;
//    TLocDBLookupTree(FTreeList).KeyValue := Value;
  end;
end;

function TrPopupTree.GetDisplayValue: String;

var
  I{, J}: Integer;

begin
{
  I := TLocDBLookupTree(FTreeList).FLinkData.ActiveRecord;
  J := TLocDBLookupTree(FTreeList).TreeData.FStrings.IndexOfObject(TLocDBLookupTree(FTreeList).TreeData.List[I]);
  Result := TLocDBLookupTree(FTreeList).TreeData.DisplayString[I];
}
  Result := TLocDBLookupTree(FTreeList).DisplayValue;
  I := TLocDBLookupTree(FTreeList).TreeData.IndexOfKey(TLocDBLookupTree(FTreeList).KeyValue);
  if (I <> -1)
  then Result := TLocDBLookupTree(FTreeList).TreeData.DisplayText[I]
  else if (Result = EmptyStr) and
    (VarIsNull(TLocDBLookupTree(FTreeList).KeyValue)
  or VarIsEmpty(TLocDBLookupTree(FTreeList).KeyValue))
  then Result := SNoneLabel;
{
  if (Result = EmptyStr) and
    (VarIsNull(TLocDBLookupTree(FTreeList).KeyValue)
  or VarIsEmpty(TLocDBLookupTree(FTreeList).KeyValue))
  then Result := SNoneLabel;
}
end;

function TrPopupTree.GetSearchString: String;
begin
  Result := TLocDBLookupTree(FTreeList).FSearchString;
end;

procedure TrPopupTree.SetSearchString(AValue: String);
begin
  TLocDBLookupTree(FTreeList).SetSearchString(AValue);
end;

{ TrDBLookupComboTree }

type
  ComboCrack = class(TCustomComboEdit);

constructor TrDBLookupComboTree.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FEscapeClear   := True;
  FValue         := EmptyStr;
  FKeyValue      := null;
  FDisplayString := SNoneLabel;
  FOnDataChange  := False;

  FDTBuildMode   := trbmDefault;
  FDTNotSelectParent := False;


  FCharKeyDown   := False;
  PopupAlign     := epaLeft;
  Parent         := TWinControl(AOwner);
  AlwaysEnable   := False;
  FChanging      := False;
  FListVisible   := False;
//  FCharPress := False;
  SNoneLabel     := EmptyStr;
  FDropDownWidth := 0;

  FDataLink                 := TFieldDataLink.Create;
  FDataLink.Control         := Self;
  FDataLink.OnActiveChange  := ActiveChange;
  FDataLink.OnDataChange    := DataChange;
  FDataLink.OnEditingChange := EditingChange;
  FDataLink.OnUpdateData    := UpdateData;

  EditMask  := EmptyStr;
  Alignment := taLeftJustify;

  FOnGetImageIndex    := nil;
  FOnGetSelectedIndex := nil;

  try
    if not (csDesigning in ComponentState)
    then begin
      FPopup := TrPopupTree.Create(Self);
      TrPopupTree(FPopup).FTreeList.OnKeyChanged := KeyChanged;
      FMultiSelect := TrTreeMultiSelect.Create(TrPopupTree(FPopup).FTreeList);
      TrPopupTree(FPopup).FTreeList.FMultiSelect := FMultiSelect;
//      TrPopupTree(FPopup).SearchString := EmptyStr;
    end
    else
      FMultiSelect := TrTreeMultiSelect.Create(nil);
    GlyphKind := gkDropDown; { force update }
    ComboCrack(Self).FButton.FStandart := False;
  finally
    ControlState := ControlState - [csCreating];
  end;
  if (csDesigning in ComponentState)
  then begin
    FDTDataSource    := nil;
    FDTLookupField   := EmptyStr;
    FDTLookupDisplay := EmptyStr;
    FDTChildrenField := EmptyStr;
    FDTParentField   := EmptyStr;
    FDTChildIndex    := -1;
    FDTChildSelected := -1;
    FDTParentIndex   := -1;
    FDTParentSelected:= -1;
    FDTImages        := nil;
    FDTDropDownCount := 8;
//    FDTBuildMode     := trbmDefault;
//    FDTNotSelectParent := False;
  end
  else begin
    ChildIndex     := -1;
    ChildSelected  := -1;
    ParentIndex    := -1;
    ParentSelected := -1;
    DropDownCount  := 8;
    Images         := nil;
  end;
  FOnDropDown   := nil;
  FOnCloseUp    := nil;
  FCheckKey := scCheckKey;
  FCheckShift := [ssShift];
end;

procedure TrDBLookupComboTree.CreateWnd;
begin
  inherited;
//  TrPopupTree(FPopup).CreateWnd;
end;

destructor TrDBLookupComboTree.Destroy;
begin
  FDTDataSource := nil;
  FDTImages     := nil;
  FDataLink.Free;
  FDataLink := nil;
  FPopup.Free;
  FPopup := nil;
  inherited Destroy;
end;

procedure TrDBLookupComboTree.RefreshNodes;
begin
  if Assigned(FPopup) and Assigned(TrPopupTree(FPopup).FTreeList)
  then TrPopupTree(FPopup).FTreeList.RefreshNodes;
end;

{$IFDEF POLARIS_D4}
function TrDBLookupComboTree.ExecuteAction(Action: TBasicAction): Boolean;
begin
  Result := inherited ExecuteAction(Action) or
    ((FDataLink <> nil) and FDataLink.ExecuteAction(Action));
end;

function TrDBLookupComboTree.UpdateAction(Action: TBasicAction): Boolean;
begin
  Result := inherited UpdateAction(Action) or
    ((FDataLink <> nil) and FDataLink.UpdateAction(Action));
end;
{$ENDIF}

procedure TrDBLookupComboTree.ActiveChange(Sender: TObject);
begin
  if FDataLink.Active then begin
    CheckNotCircular;
  end;
  DataChange(Sender);
end;

procedure TrDBLookupComboTree.DataChange(Sender: TObject);
var
  S: String;
begin
  FOnDataChange := True;
  try
    S := Text;
    if (FDataLink.Field <> nil) and not (csDesigning in ComponentState) then
    begin

      if VarIsEmpty(FDataLink.Field.Value) or VarIsNull(FDataLink.Field.Value)
      then begin
        FValue    := EmptyStr;
        FKeyValue := null;
        TrPopupTree(FPopup).SetValue(null);
        {if Assigned(FDataLink.Field)
          and FDataLink.Field.Required
        then FDisplayString := EmptyStr
        else} FDisplayString := SNoneLabel;
      end
      else begin
        FValue := FDataLink.Field.Value;
        FKeyValue := FDataLink.Field.Value;
        TrPopupTree(FPopup).SetValue(FDataLink.Field.Value);
        FDisplayString := TrPopupTree(FPopup).DisplayValue;
      end
    end
    else begin
      if csDesigning in ComponentState then EditText := Name
      else if (FDataLink.Field <> nil) then begin
  //      FValue := EmptyStr;
        FValue := FDataLink.Field.AsString;
  //      TrPopupTree(FPopup).SetValue(null);
        TrPopupTree(FPopup).SetValue(FKeyValue);
  //***
        {if Assigned(FDataLink.Field)
          and FDataLink.Field.Required
        then FDisplayString := EmptyStr
        else} FDisplayString := SNoneLabel;
      end
    end;
    Text := DisplayString;
    SetReadOnly(FDataLink.ReadOnly);
  finally
    FOnDataChange := False;
  end;
end;

procedure TrDBLookupComboTree.EditingChange(Sender: TObject);
begin
//  inherited ReadOnly := not FDataLink.Editing;
end;

procedure TrDBLookupComboTree.UpdateData(Sender: TObject);
begin
  ValidateEdit;
end;

procedure TrDBLookupComboTree.KeyChanged(Sender: TObject; Node: TTreeNode);
var
  B: Boolean;
  I: Integer;
begin
  if not (csDesigning in ComponentState)
  then
    if FPopupVisible
    then KeyValue := TrPopupTree(FPopup).FTreeList.FKeyValue;
  if not FPopupVisible {and
    (CompareVar(TrPopupTree(FPopup).FTreeList.FKeyValue, KeyValue) <> 0)}
  then begin

    B := False;
    if Assigned(FDataLink) and Assigned(FDataLink.Field)
    and (CompareVar(FDataLink.Field.Value,FKeyValue) <> 0)
(*
    and (CompareVar({TrPopupTree(FPopup).GetValue}
    TrPopupTree(FPopup).FTreeList.FKeyValue, FDataLink.Field.Value) = 0)
    and (CompareVar({TrPopupTree(FPopup).GetValue}
    TrPopupTree(FPopup).FTreeList.FKeyValue, KeyValue) <> 0)
*)
    then begin
      B := True;
      FKeyValue := FDataLink.Field.Value;
      FValue := VarToStr(FKeyValue);
    end
    else if (CompareVar({TrPopupTree(FPopup).GetValue}
    TrPopupTree(FPopup).FTreeList.FKeyValue, KeyValue) <> 0)
    then begin
      B := True;
      FKeyValue := TrPopupTree(FPopup).FTreeList.FKeyValue;
      FValue := VarToStr(FKeyValue);
    end;

    I := TrPopupTree(FPopup).FTreeList.TreeData.IndexOfKey(FKeyValue);
    if I <> -1
    then FDisplayString := TrPopupTree(FPopup).FTreeList.TreeData.GetDisplayText(I)
    else FDisplayString := SNoneLabel;
    Text := DisplayString;
//    SelectAll;

    if B then begin
      AcceptValue(KeyValue);
      DoChange;
    end;
  end;
//  Invalidate;

//  if not CanModify then Text := FDisplayString;
{
  I := TrPopupTree(FPopup).FTreeList.FLinkData.ActiveRecord;
//  TrPopupTree(FPopup).FTreeList.FTreeData.TreeEntry[I].KeyValue;
  B := CompareVar(TrPopupTree(FPopup).FTreeList.FTreeData.TreeEntry[I].KeyValue
  //KeyValue
  , TrPopupTree(FPopup).FTreeList.FKeyValue) = 0;
  if not (csDesigning in ComponentState)
  then KeyValue := TrPopupTree(FPopup).FTreeList.FKeyValue;
  if not FPopupVisible then begin
    if B then AcceptValue(KeyValue);
  end;
  Invalidate;
}
end;

procedure TrDBLookupComboTree.CMGetDataLink(var Message: TMessage);
begin
  Message.Result := Integer(FDataLink);
end;

procedure TrDBLookupComboTree.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  Button.Enabled := Enabled;
  Invalidate;
end;

procedure TrDBLookupComboTree.RCTLookupChange(var Message: TMessage);
var
  I: Integer;
//  S: String;
begin
  I := TrPopupTree(FPopup).FTreeList.FTreeData.IndexOfKey(FKeyValue);
  if I >= 0 then begin
    if (TrPopupTree(FPopup).FTreeList.Items.GetNode(HTreeItem(Message.WParam))
      = TrPopupTree(FPopup).FTreeList.FTreeData.TreeNode[I])
    then FDisplayString := TrPopupTree(FPopup).FTreeList.FTreeData.TreeNode[I].Text
//    else S := FDisplayString;
  end else begin
//***
    {if Assigned(FDataLink.Field)
    and FDataLink.Field.Required
    then S := EmptyStr
    else} FDisplayString := SNoneLabel;
  end;
  Text := DisplayString;
//  if S <> FDisplayString then FDisplayString := S;
  if Message.WParam = 0
  then if Enabled
    then Button.Enabled := TrPopupTree(FPopup).FTreeList.FLinkData.Active
    else Button.Enabled := Enabled;
  Message.Result := 1;
end;

procedure TrDBLookupComboTree.WMKillFocus(var Message: TWMKillFocus);
var
  H: {$IFDEF POLARIS_D4} HWND {$ELSE} Integer {$ENDIF};
begin
  H := Message.FocusedWnd;
  if (H = FPopup.Handle) or (H = Handle)
  then begin
    if H <> FPopup.Handle
    then begin
      Message.FocusedWnd := Handle;
      Message.Result := 1;
    end
    else begin
      Windows.SetFocus(Handle);
      Message.Result := 1;
    end;
//    inherited;
  end
  else inherited;
end;

procedure TrDBLookupComboTree.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then begin
    if (FDataLink <> nil) and (AComponent = DataSource) then DataSource := nil;
    if Assigned(Images) and (AComponent = Images) then Images := nil;
  end
end;

procedure TrDBLookupComboTree.WMKeyDown(var Message: TWMKeyDown);
var
  M: TMessage;
begin
  inherited;
   M := TMessage(Message);
  {if not ((Message.CharCode in [VK_UP, VK_PRIOR, VK_NEXT]) or
     (not TestKeyShiftState(KeyDataToShiftState(Message.KeyData), [ssAlt])
     and (Message.CharCode = VK_DOWN)))
  then  inherited
  else}
  if (FPopup <> nil) {and (Message.Result = 0)} and not ReadOnly
  then
    FPopup.Perform(M.Msg, M.WParam, M.LParam);
end;

procedure TrDBLookupComboTree.WMChar(var Message: TWMChar);
{var
  sc: TShortCut;}
begin
  {sc := ShortCut(Message.CharCode, KeyDataToShiftState(Message.KeyData));
  if Assigned(FPopup) and Assigned(TrPopupTree(FPopup).FTreeList) and
    (sc = TrPopupTree(FPopup).FTreeList.FCheckKey)
  then
    TrPopupTree(FPopup).FTreeList.ChangeCheckNode(TrPopupTree(FPopup).FTreeList.Selected)
  else}
    try
      FCharKeyDown := (Chr(Message.CharCode) in [#32..#255]) and
         not TestKeyShiftState(KeyDataToShiftState(Message.KeyData), [ssAlt,ssCtrl]);
      inherited;
    finally
      FCharKeyDown := False;
    end;
end;

procedure TrDBLookupComboTree.SetValue(AValue: String);
var
  I: Integer;
begin
  if (csDesigning in ComponentState) then exit;
  if FValue <> AValue
  then begin
    I := TrPopupTree(FPopup).FTreeList.FTreeData.IndexOfKey(AValue);
    if I >= 0 then begin
      FValue := AValue;
      FKeyValue := TrPopupTree(FPopup).FTreeList.FTreeData.TreeEntry[I].KeyValue;
      FDisplayString := TrPopupTree(FPopup).FTreeList.FTreeData.GetDisplayText(I)
    end
    else begin
      FValue := EmptyStr;
      FKeyValue := null;
//***
      {if Assigned(FDataLink.Field)
        and FDataLink.Field.Required
      then FDisplayString := EmptyStr
      else} FDisplayString := SNoneLabel;
    end;
  end;
end;

procedure TrDBLookupComboTree.SetKeyValue(AValue: Variant);
var
  I: Integer;
//  S: String;
begin
  if (csDesigning in ComponentState) then exit;
  if CompareVar(FKeyValue, AValue) <> 0
  then begin
    I := TrPopupTree(FPopup).FTreeList.FTreeData.IndexOfKey(AValue);
    if I >= 0 then begin

      FValue := VarToStr(AValue);
      FKeyValue := TrPopupTree(FPopup).FTreeList.FTreeData.TreeEntry[I].KeyValue;
      if Assigned(FDataLink.Field)
      then AcceptValue(AValue);
{$IFDEF POLARIS_D4}
      TrPopupTree(FPopup).FTreeList.FLinkData.MoveBy(I-TrPopupTree(FPopup).FTreeList.FLinkData.ActiveRecord);
{$ELSE}
      TrPopupTree(FPopup).FTreeList.FLinkData.ActiveRecord := I;
{$ENDIF}
//      TrPopupTree(FPopup).FTreeList.FDataSet.CursorPosChanged;
      TrPopupTree(FPopup).FTreeList.FKeyValue := FKeyValue;
      TrPopupTree(FPopup).FTreeList.FValue := FValue;

      FDisplayString := TrPopupTree(FPopup).FTreeList.TreeData.GetDisplayText(I);
      if (TrPopupTree(FPopup).SearchString <> EmptyStr) and FPopupVisible
      then Text := TrPopupTree(FPopup).SearchString
      else begin
        Text := DisplayString;
//        SelectAll;
      end;
{
      I := Length(Text);
      S := Copy( FDisplayString, 1, I);
      I := -1;
}
//      if AnsiCompareText(S, Text) <> 0 then I := -1;
    end
    else begin

      if TrPopupTree(FPopup).FTreeList.Items.Count = 0
      then begin
        FKeyValue := AValue;
// new
        if VarIsNull(AValue) then FValue := EmptyStr else FValue := AValue;
        if FCharKeyDown
          then FDisplayString := EmptyStr
          else FDisplayString := SNoneLabel;
//
        Change;
      end
      else begin
        FValue := EmptyStr;
        FKeyValue := null;
        if FCharKeyDown
        then FDisplayString := EmptyStr
        else {FDisplayString := SNoneLabel;}
        begin
  //***
          {if Assigned(FDataLink.Field)
            and FDataLink.Field.Required
          then FDisplayString := EmptyStr
          else} FDisplayString := SNoneLabel;
        end;
      end;
      if (TrPopupTree(FPopup).SearchString <> EmptyStr) and FPopupVisible
      then begin
        Text := TrPopupTree(FPopup).SearchString;
//        SelectAll;
      end
      else begin
        Text := DisplayString;
//        SelectAll;
      end;
//      Text := FDisplayString;
//      SelectAll;
//      I := -1;
    end;
{
    if not (Assigned(FDataLink) and Assigned(FDataLink.Field))
      and not FCharKeyDown
    then begin
      Text := FDisplayString;
      if I >= 0 then SetSel(I, Length(FDisplayString))
      else SelectAll;
    end;
}
    if not FPopupVisible then inherited DoChange;
  end;
//new  
//  TrPopupTree(FPopup).FTreeList.SyncData(TrPopupTree(FPopup).FTreeList.Selected);
end;

procedure TrDBLookupComboTree.SetDropDownWidth(AValue: Integer);
begin
  FDropDownWidth := AValue;
end;

procedure TrDBLookupComboTree.SetNoneLabel(AValue: String);
var
  B: Boolean;
begin
  if SNoneLabel <> Trim(AValue)
  then begin
    B := Text = SNoneLabel;
    SNoneLabel := Trim(AValue);
    if not (csDesigning in ComponentState)
    then TrPopupTree(FPopup).NoneLabel := SNoneLabel;
    if B then begin
      Text := SNoneLabel;
      FDisplayString := SNoneLabel;
    end;
  end;
end;

function TrDBLookupComboTree.GetDropDownCount: Integer;
begin
  if (csDesigning in ComponentState)
  then Result := FDTDropDownCount
  else Result := TrPopupTree(FPopup).FDropDownRows;
end;

procedure TrDBLookupComboTree.SetDropDownCount(AValue: Integer);
begin
  if (csDesigning in ComponentState)
  then begin
    if AValue <> FDTDropDownCount then begin
      FDTDropDownCount := AValue;
      if FDTDropDownCount < 2 then FDTDropDownCount := 2;
      if FDTDropDownCount > 50 then FDTDropDownCount := 50;
    end;
  end
  else TrPopupTree(FPopup).SetDropDownRows(AValue);
end;

function TrDBLookupComboTree.GetDataField: String;
begin
  Result := FDataLink.FieldName;
end;

procedure TrDBLookupComboTree.SetDataField(AValue: String);
begin
  FDataLink.FieldName := AValue;
  DataChange(Self);
end;

function TrDBLookupComboTree.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TrDBLookupComboTree.SetDataSource(AValue: TDataSource);
begin
  if FDataLink.DataSource <> AValue then begin
{$IFDEF POLARIS_D4}
    if not (FDataLink.DataSourceFixed and (csLoading in ComponentState)) then
{$ENDIF}
      FDataLink.DataSource := AValue;
    if AValue <> nil then AValue.FreeNotification(Self);
  end
end;

function TrDBLookupComboTree.GetLinkDataSource: TDataSource;
begin
  if (csDesigning in ComponentState)
  then Result := FDTDataSource
  else Result := TLocDBLookupTree(TrPopupTree(FPopup).FTreeList).GetLinkDataSource;
end;

procedure TrDBLookupComboTree.SetLinkDataSource(AValue: TDataSource);
begin
  if (csDesigning in ComponentState)
  then begin
    if FDTDataSource <> AValue then begin
      FDTDataSource := AValue;
      if AValue <> nil then AValue.FreeNotification(Self);
    end;
  end
  else begin
    TLocDBLookupTree(TrPopupTree(FPopup).FTreeList).SetLinkDataSource(AValue);
    DataChange(Self);
  end;
end;

function TrDBLookupComboTree.GetKeyField: String;
begin
  if (csDesigning in ComponentState)
  then Result := FDTLookupField
  else Result := TLocDBLookupTree(TrPopupTree(FPopup).FTreeList).KeyField;
end;

procedure TrDBLookupComboTree.SetKeyField(AValue: String);
begin
  if (csDesigning in ComponentState)
  then FDTLookupField := AValue
  else begin
    TLocDBLookupTree(TrPopupTree(FPopup).FTreeList).KeyField := AValue;
    DataChange(Self);
  end;
end;

function TrDBLookupComboTree.GetDisplayFields: String;
begin
  if (csDesigning in ComponentState)
  then Result := FDTLookupDisplay
  else Result := TLocDBLookupTree(TrPopupTree(FPopup).FTreeList).DisplayFields;
end;

procedure TrDBLookupComboTree.SetDisplayFields(AValue: String);
begin
  if (csDesigning in ComponentState)
  then FDTLookupDisplay := AValue
  else begin
    TLocDBLookupTree(TrPopupTree(FPopup).FTreeList).DisplayFields := AValue;
    DataChange(Self);
  end
end;

function TrDBLookupComboTree.GetChildrenField: String;
begin
  if (csDesigning in ComponentState)
  then Result := FDTChildrenField
  else Result := TLocDBLookupTree(TrPopupTree(FPopup).FTreeList).ChildrenField;
end;

procedure TrDBLookupComboTree.SetChildrenField(AValue: String);
begin
  if (csDesigning in ComponentState)
  then FDTChildrenField := AValue
  else begin
    TLocDBLookupTree(TrPopupTree(FPopup).FTreeList).ChildrenField := AValue;
    DataChange(Self);
  end
end;

function TrDBLookupComboTree.GetParentField: String;
begin
  if (csDesigning in ComponentState)
  then Result := FDTParentField
  else Result := TLocDBLookupTree(TrPopupTree(FPopup).FTreeList).ParentField;
end;

procedure TrDBLookupComboTree.SetParentField(AValue: String);
begin
  if (csDesigning in ComponentState)
  then FDTParentField := AValue
  else begin
    TLocDBLookupTree(TrPopupTree(FPopup).FTreeList).ParentField := AValue;
    DataChange(Self);
  end
end;

function TrDBLookupComboTree.GetReadOnly: Boolean;
begin
  Result := FDataLink.ReadOnly;
end;

procedure TrDBLookupComboTree.SetReadOnly(AValue: Boolean);
begin
  FDataLink.ReadOnly := AValue;
  if not (csDesigning in ComponentState) then begin
{
    if Assigned(FPopup)
    then inherited ReadOnly := AValue or not TrPopupTree(FPopup).FTreeList.FLinkData.Active
    else
}
    inherited ReadOnly := FDataLink.ReadOnly;
    if Enabled
    then begin
      if Assigned(FPopup)
      then Button.Enabled := not FDataLink.ReadOnly and TrPopupTree(FPopup).FTreeList.FLinkData.Active
      else Button.Enabled := Enabled;
    end
    else Button.Enabled := Enabled;
  end else begin
    inherited ReadOnly := AValue;
    Button.Enabled := AValue;
  end;
end;

function TrDBLookupComboTree.GetChildIndex: Integer;
begin
  if (csDesigning in ComponentState)
  then Result := FDTChildIndex
  else Result := TrPopupTree(FPopup).FTreeList.ChildIndex;
end;

procedure TrDBLookupComboTree.SetChildIndex(AValue: Integer);
begin
  if (csDesigning in ComponentState)
  then FDTChildIndex := AValue
  else TrPopupTree(FPopup).FTreeList.ChildIndex := AValue;
end;

function TrDBLookupComboTree.GetChildSelected: Integer;
begin
  if (csDesigning in ComponentState)
  then Result := FDTChildSelected
  else Result := TrPopupTree(FPopup).FTreeList.ChildSelected;
end;

procedure TrDBLookupComboTree.SetChildSelected(AValue: Integer);
begin
  if (csDesigning in ComponentState)
  then FDTChildSelected := AValue
  else TrPopupTree(FPopup).FTreeList.ChildSelected := AValue;
end;

function TrDBLookupComboTree.GetParentIndex: Integer;
begin
  if (csDesigning in ComponentState)
  then Result := FDTParentIndex
  else Result := TrPopupTree(FPopup).FTreeList.ParentIndex;
end;

procedure TrDBLookupComboTree.SetParentIndex(AValue: Integer);
begin
  if (csDesigning in ComponentState)
  then FDTParentIndex := AValue
  else TrPopupTree(FPopup).FTreeList.ParentIndex := AValue;
end;

function TrDBLookupComboTree.GetParentSelected: Integer;
begin
  if (csDesigning in ComponentState)
  then Result := FDTParentSelected
  else Result := TrPopupTree(FPopup).FTreeList.ParentSelected;
end;

procedure TrDBLookupComboTree.SetParentSelected(AValue: Integer);
begin
  if (csDesigning in ComponentState)
  then FDTParentSelected := AValue
  else TrPopupTree(FPopup).FTreeList.ParentSelected := AValue;
end;

function TrDBLookupComboTree.CanModify: Boolean;
begin
  if Assigned(FDataLink) and Assigned(FDataLink.DataSet)
  then Result := FDataLink.CanModify
  else Result := not ReadOnly;
end;

function TrDBLookupComboTree.GetImageList: TImageList;
begin
  if (csDesigning in ComponentState)
  then Result := FDTImages
  else
  if Assigned(FPopup) and Assigned(TrPopupTree(FPopup).FTreeList)
  then
  {$IFDEF POLARIS_D4}
  Result := TImageList(TrPopupTree(FPopup).FTreeList.Images)
  {$ELSE}
  Result := TrPopupTree(FPopup).FTreeList.Images
  {$ENDIF}
  else Result := nil;
end;

procedure TrDBLookupComboTree.SetImageList(AValue: TImageList);
begin
  if (csDesigning in ComponentState)
  then FDTImages := AValue
  else TrPopupTree(FPopup).FTreeList.Images := AValue;
end;

function TrDBLookupComboTree.GetBuildMode: TrTreeBuildMode;
begin
  if (csDesigning in ComponentState)
  then Result := FDTBuildMode
  else Result := TrPopupTree(FPopup).FTreeList.BuildMode;
end;

procedure TrDBLookupComboTree.SetBuildMode(Value: TrTreeBuildMode);
begin
  if (csDesigning in ComponentState)
  then FDTBuildMode := Value
  else TrPopupTree(FPopup).FTreeList.BuildMode := Value;
end;

function TrDBLookupComboTree.GetNotSelectParent: Boolean;
begin
  if (csDesigning in ComponentState)
  then Result := FDTNotSelectParent
  else Result := TrPopupTree(FPopup).FTreeList.NotSelectParent;
end;

procedure TrDBLookupComboTree.SetNotSelectParent(Value: Boolean);
begin
  if (csDesigning in ComponentState)
  then FDTNotSelectParent := Value
  else TrPopupTree(FPopup).FTreeList.NotSelectParent := Value;
end;

{function TrDBLookupComboTree.EditCanModify: Boolean;
begin
  Result := not ReadOnly;
end;}

procedure TrDBLookupComboTree.KeyDown(var Key: Word; Shift: TShiftState);
//var
//  Delta: ShortInt;
//  TValue,
begin
  if (ClickKey = ShortCut(Key, Shift)) and (ButtonWidth > 0) then begin
    ButtonClick;
    Key := 0;
  end;
  {else begin
    TrPopupTree(FPopup).KeyDown(Key, Shift);
    Key := 0;
  end;}
  inherited KeyDown(Key, Shift);
  //Invalidate;
end;

procedure TrDBLookupComboTree.KeyPress(var Key: Char);
//var
//  PKey: Char;
begin
  if (csDesigning in ComponentState) then begin
    inherited KeyPress(Key);
    exit;
  end;

  if FPopupVisible then begin
    TrPopupTree(FPopup).KeyPress(Key);
    if Key in [#13, #27]
    then PopupCloseUp(Self, Key=#13)
    else if TrPopupTree(FPopup).SearchString <> EmptyStr
      then Text := TrPopupTree(FPopup).SearchString;
    Key := #0;
  end
  else begin
    if Key <> #27 then begin
      if Key in [#32..#255] then begin
        PopupDropDown(True);
        if FPopupVisible
        then TrPopupTree(FPopup).KeyPress(Key);
      end else if not (Key in [^C]) then Key := #0;
    end
    else if (Key = #27) and FEscapeClear
      and ((FValue <> EmptyStr) and (FValue<>SNoneLabel))
      and CanModify then begin
        ResetField;
        inherited SelectAll;
        if (FValue = EmptyStr) or (FValue=SNoneLabel) then Key := #0;   {ALEX}
      end;
  end;
  inherited KeyPress(Key);
  if (Key in [#13, #27]) then
    GetParentForm(Self).Perform(CM_DIALOGKEY, Byte(Key), 0);
//  Invalidate;
end;

procedure TrDBLookupComboTree.DoChange;
begin
  {if (CompareVar(FKeyValue, TrPopupTree(FPopup).FIValue) <> 0)
  or Assigned(FDataLink.Field)
  then}
  inherited DoChange;
end;

procedure TrDBLookupComboTree.Change;
begin
{
  if not PopupVisible then DoChange
  else PopupChange;
}
  if PopupVisible then begin
    PopupChange;
  end

end;

procedure TrDBLookupComboTree.PopupChange;
begin
  inherited PopupChange;
end;

procedure TrDBLookupComboTree.PopupDropDown(DisableEdit: Boolean);
var
  P: TPoint;
  Y: Integer;
begin
  if (csDesigning in ComponentState) then Exit;
  if (FPopup <> nil) and not (ReadOnly or FPopupVisible) then begin
    TrPopupTree(FPopup).FTreeList.FAutoMove := True;
    TrPopupTree(FPopup).FPrevValue := KeyValue;

      {if Assigned(FDataLink.Field)
        and FDataLink.Field.Required
      then TrPopupTree(FPopup).SNoneLabel := EmptyStr
      else} TrPopupTree(FPopup).SNoneLabel := SNoneLabel;

    if FDropDownWidth <> 0
    then TrPopupTree(FPopup).Width := FDropDownWidth
    else TrPopupTree(FPopup).Width := Width;
    P := Parent.ClientToScreen(Point(Left, Top));
    Y := P.Y + Height;
    if Y + FPopup.Height > Screen.Height then
      Y := P.Y - FPopup.Height;
    case PopupAlign of
      epaRight:
        begin
          Dec(P.X, FPopup.Width - Width);
          if P.X < 0 then Inc(P.X, FPopup.Width - Width);
        end;
      epaLeft:
        begin
          if P.X + FPopup.Width > Screen.Width then
            Dec(P.X, FPopup.Width - Width);
        end;
    end;
    if P.X < 0 then P.X := 0
    else if P.X + FPopup.Width > Screen.Width then
      P.X := Screen.Width - FPopup.Width;
    if (Text <> EmptyStr) and (Text <> SNoneLabel)
    then SetPopupValue(FKeyValue{Text})
    else SetPopupValue(Null);
    if CanFocus then SetFocus;
    ShowPopup(Point(P.X, Y));
    TrPopupTree(FPopup).FTreeList.FSearchString := '';
    FPopupVisible := True;
    if DisableEdit then begin
      inherited ReadOnly := True;
      HideCaret(Handle);
    end;
  end;
//new
//  TrPopupTree(FPopup).FTreeList.SyncData(TrPopupTree(FPopup).FTreeList.Selected);
end;

function TrDBLookupComboTree.AcceptPopup(var Value: Variant): Boolean;
begin
  if Assigned(FDataLink) and Assigned(FDataLink.Field)
  then Result := CompareVar(KeyValue, FDataLink.Field.Value) <> 0
  else Result := True;
end;

procedure TrDBLookupComboTree.SetPopupValue(const Value: Variant);
begin
  if not (csDesigning in ComponentState)
  then begin
    TrPopupTree(FPopup).SetValue(Value);
    if not (VarIsEmpty(Value) or VarIsNull(Value))
    then TrPopupTree(FPopup).FTreeList.SyncDataSet;
  end
end;

function TrDBLookupComboTree.GetPopupValue: Variant;
begin
  if not (csDesigning in ComponentState)
  then Result := TrPopupTree(FPopup).DisplayValue;
end;

procedure TrDBLookupComboTree.AcceptValue(const Value: Variant);
var
  AValue: Variant;
begin
  AValue := Value;
  if Assigned(FDataLink) and Assigned(FDataLink.Field)
  and not FPopupVisible and not FOnDataChange
  then begin
    if CompareVar(AValue, FDataLink.Field.Value) <> 0 then begin
      Modified := True;
      if CanModify
      then begin
        if FDataLink.Edit then
          FDataLink.Field.Value := AValue
        else SetKeyValue(FDataLink.Field.Value);
//        SelectAll;
      end
    end
  end;
  UpdatePopupVisible;
end;

procedure TrDBLookupComboTree.PopupCloseUp(Sender: TObject; Accept: Boolean);
var
  AValue: Variant;
  I: Integer;
begin
  if (FPopup <> nil) and FPopupVisible then begin
    if GetCapture <> 0 then SendMessage(GetCapture, WM_CANCELMODE, 0, 0);
    if not (csDesigning in ComponentState)
    then
    AValue := TrPopupTree(FPopup).FTreeList.FKeyValue;
    HidePopup;
    FPopupVisible := False;

    I := TrPopupTree(FPopup).FTreeList.FTreeData.IndexOfKey(AValue);
    if I >= 0 then
      FDisplayString := TrPopupTree(FPopup).FTreeList.FTreeData.GetDisplayText(I)
    else begin
      {if Assigned(FDataLink.Field) and FDataLink.Field.Required
      then FDisplayString := EmptyStr
      else} FDisplayString := SNoneLabel;
    end;

    try
      try
        if CanFocus then begin
          SetFocus;
          if GetFocus = Handle then ComboCrack(Self).SetShowCaret;
        end;
      except
        { ignore exceptions }
      end;
      SetDirectInput(DirectInput);
      Invalidate;
      if Accept and AcceptPopup(AValue) and CanModify then begin
        AcceptValue(AValue);
        DoChange;
      end;
      Text := DisplayString;
      if FFocused then inherited SelectAll;
      if Assigned(FOnCloseUp) then FOnCloseUp(Self);
    finally
      FPopupVisible := False;
      inherited ReadOnly := GetReadOnly;
      TrPopupTree(FPopup).FTreeList.FAutoMove := False;
    end;
  end;
end;

procedure TrDBLookupComboTree.ButtonClick;
begin
  if Assigned(OnButtonClick) then OnButtonClick(Self);
  DropDown;
end;

procedure TrDBLookupComboTree.DropDown;
begin
  if FPopup <> nil then begin
    if FPopupVisible
    then PopupCloseUp(Self, False)
    else PopupDropDown(True);
  end;
end;

procedure TrDBLookupComboTree.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  if (FPopup <> nil) and (Button = mbLeft) then begin
    if CanFocus then SetFocus;
    if not FFocused then Exit;
    if FPopupVisible then PopupCloseUp(Self, False);
  end;
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TrDBLookupComboTree.ResetField;
begin
  if (FDataLink.DataSource <> nil) and (FDataLink.Field <> nil)
  and CanModify then begin
    if FDataLink.Edit then FDataLink.Field.Value := null;
  end;
  KeyValue := null;
end;

procedure TrDBLookupComboTree.DoEnter;
begin
  inherited DoEnter;
  inherited SelectAll;
//  TrPopupTree(FPopup).HandleNeeded;
  TrPopupTree(FPopup).SetValue(FKeyValue);
end;

procedure TrDBLookupComboTree.DoExit;
{var
  I: Integer;
  S: String;}
begin
  (*I := TrPopupTree(FPopup).FTreeList.FTreeData.IndexOfKey(FKeyValue);
  if I >= 0 then begin
    S := TrPopupTree(FPopup).FTreeList.FTreeData.GetDisplayText(I)
  end else
    {if Assigned(FDataLink.Field) and FDataLink.Field.Required
      then S := EmptyStr
      else} S := SNoneLabel;

  if S <> Text then Text := S;*)
  inherited DoExit;
  inherited SelectAll;
end;

procedure TrDBLookupComboTree.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned( FPopup) then
    if FDropDownWidth <> 0
      then TrPopupTree(FPopup).Width := FDropDownWidth
      else TrPopupTree(FPopup).Width := Width;
end;

procedure TrDBLookupComboTree.CheckNotCircular;
begin
  if (LookupSource <> nil) and FDataLink.Active and ((DataSource = LookupSource)
     or (FDataLink.DataSet = LookupSource.DataSet))
  then _DBError(SCircularDataLink);
end;

procedure TrDBLookupComboTree.SetOnGetImageIndex(AValue: TTVExpandedEvent);
begin
  if (csDesigning in ComponentState)
  then FOnGetImageIndex := AValue
  else TrPopupTree(FPopup).FTreeList.OnGetImageIndex := AValue
end;

procedure TrDBLookupComboTree.SetOnGetSelectedIndex(AValue: TTVExpandedEvent);
begin
  if (csDesigning in ComponentState)
  then FOnGetSelectedIndex := AValue
  else TrPopupTree(FPopup).FTreeList.OnGetSelectedIndex := AValue
end;

procedure TrDBLookupComboTree.SetMultiSelect(
  const Value: TrTreeMultiSelect);
begin
  FMultiSelect.Assign(Value);
  if Assigned(FPopup) and Assigned(TrPopupTree(FPopup).FTreeList) then
    TrPopupTree(FPopup).FTreeList.MultiSelect := Value;
end;

function TrDBLookupComboTree.GetCheckList: TrTreeCheckList;
begin
  if Assigned(FPopup) and Assigned(TrPopupTree(FPopup).FTreeList) then
    Result := TrPopupTree(FPopup).FTreeList.CheckList
  else Result := nil;
end;

procedure TrDBLookupComboTree.SetCheckListChange(
  const Value: TTreeCheckListChange);
begin
  FCheckListChange := Value;
  if Assigned(FPopup) and Assigned(TrPopupTree(FPopup).FTreeList) then
    TrPopupTree(FPopup).FTreeList.OnCheckListChange := Value;
end;

procedure TrDBLookupComboTree.SetCheckListChanging(
  const Value: TTreeCheckListChangingEvent);
begin
  FCheckListChanging := Value;
  if Assigned(FPopup) and Assigned(TrPopupTree(FPopup).FTreeList) then
    TrPopupTree(FPopup).FTreeList.OnCheckListChanging := Value;
end;

function TrDBLookupComboTree.GetDisplayString: String;
var
  n: Integer;
begin
  if (FPopup <> nil) and TrPopupTree(FPopup).FTreeList.IsMultiSelect
  then begin
    n := TrPopupTree(FPopup).FTreeList.CheckList.Count;
    if n > 0 then
      if n = 1 then
        Result := Format(srChecked, [NumSomething(Max(n, 1), srRecord, n > 0)])
      else
        Result := Format(srCheckedMulti, [NumSomething(Max(n, 1), srRecord, n > 0)])
    else Result := FDisplayString;
  end
  else Result := FDisplayString;
end;

procedure TrDBLookupComboTree.SetCheckKey(const Value: TShortCut);
begin
  FCheckKey := Value;
  if Assigned(FPopup) and Assigned(TrPopupTree(FPopup).FTreeList) then
    TrPopupTree(FPopup).FTreeList.FCheckKey := Value;
end;

procedure TrDBLookupComboTree.SetCheckShift(const Value: TShiftState);
begin
  FCheckShift := Value;
  if Assigned(FPopup) and Assigned(TrPopupTree(FPopup).FTreeList) then
    TrPopupTree(FPopup).FTreeList.FCheckShift := Value;
end;

function TrDBLookupComboTree.GetIDText: String;
begin
  if Assigned(FPopup) and Assigned(TrPopupTree(FPopup).FTreeList) then
    with TrPopupTree(FPopup).FTreeList do
      if CheckList.Count > 1 then Result := CheckList.SaveToString(',')
      else Result := Self.Value
  else Result := EmptyStr;
end;

procedure TrDBLookupComboTree.SetIDText(const Value: String);
begin
  if Assigned(FPopup) and Assigned(TrPopupTree(FPopup).FTreeList) then
  begin
    TrPopupTree(FPopup).FTreeList.CheckList.Clear;
    if Value = EmptyStr then KeyValue := Null
    else if WordCount(Value, [',']) = 1 then KeyValue := Value
    else begin
//******    !!!
      TrPopupTree(FPopup).FTreeList.CheckList.LoadFromString(Value, [','], varInteger);
      KeyValue := TrPopupTree(FPopup).FTreeList.CheckList[0];
    end;
  end;
end;

end.
