//==============================================
//      File rdbutils.pas
//
//      Components library for Delphi and BCB++.
//      ,       .
//      Procedures, functions and types
//      for working with database.
//
//      Copyright 1998-2002 Polaris Software
//      http://polesoft.da.ru
//      mailto: PolarisLib@mail.ru
//==============================================
unit rDBUtils;

{$I POLARIS.INC}

interface

uses Windows, Classes, Forms, Menus, rDBConst, DB, DBConsts, DBCtrls, DBUtils,
     SysUtils, Registry, rUtils, rVCLUtils, ErrorMes,
     {$IFDEF POLARIS_D6}RxStrUtils, Variants{$ELSE}StrUtils{$ENDIF};

type
  TrDBOption =
    (tboFirst,tboLast,tboEmpty,
     tboCanModify,tboInserting,tboModifying,tboEditing,
     tboAfterEditAssigned,tboNULL);
  TrDBOptionSet = set of TrDBOption;

  TrDBAction = (tbaCustom,tbaFirst,tbaPrior,tbaNext,tbaLast,tbaRefresh,
                tbaInsert,tbaAppend,tbaDelete,tbaEdit,tbaView,tbaPost,tbaCancel,
                tbaConfirmDelete,tbaNULL);

  TOnActiveChanged =
    procedure (Sender: TObject; DataLink: TFieldDataLink; var SetEnabled: Boolean) of object;

const

  Negatives: array[TrDBAction] of TrDBOptionSet =
    ([],[tboFirst,tboEditing],[tboFirst,tboEditing],[tboLast,tboEditing],[tboLast,tboEditing],[tboEditing],
     [tboEditing],[tboEditing],[tboEmpty,tboEditing],[tboEmpty,tboEditing],[tboEmpty,tboEditing],[],[],
     [tboEmpty,tboEditing],[tboNULL]);

  Positives: array[TrDBAction] of TrDBOptionSet =
    ([],[],[],[],[],[],
     [tboCanModify],[tboCanModify],[tboCanModify],[tboCanModify],[tboAfterEditAssigned],[tboEditing],[tboEditing],
     [tboCanModify],[tboEditing]);

  DBActionResources: array[TrDBAction] of string[20] =
    ('','RDBN_FIRST','RDBN_PRIOR','RDBN_NEXT','RDBN_LAST','RDBN_REFRESH',
     'RDBN_INSERT','RDBN_INSERT','RDBN_DELETE','RDBN_EDIT','RDBN_VIEW','RDBN_POST','RDBN_CANCEL',
     'RDBN_DELETE','NULL_BTN');

  // 666 - set in initialization section
  DBActionShortCuts: array[TrDBAction] of TShortCut =
    (0, 666, VK_UP, VK_DOWN, 666, 666,
     VK_INSERT, VK_INSERT, VK_DELETE, VK_F4, 666, 666, VK_ESCAPE,
     VK_DELETE, VK_DELETE);

  DBActionHints: Array[TrDBAction] of Pointer = (
    @RDBN_Custom,
    @RDBN_First,
    @RDBN_Prior,
    @RDBN_Next,
    @RDBN_Last,
    @RDBN_Refresh,
    @RDBN_Insert,
    @RDBN_Insert,
    @RDBN_Delete,
    @RDBN_Edit,
    @RDBN_View,
    @RDBN_Post,
    @RDBN_Cancel,
    @RDBN_Delete,
    @RDBN_NULL);

  { TrActions }
  RDBA_Caption: Array[TrDBAction] of Pointer = (
    @RDBA_Custom,
    @RDBA_First,
    @RDBA_Prior,
    @RDBA_Next,
    @RDBA_Last,
    @RDBA_Refresh,
    @RDBA_Insert,
    @RDBA_Insert,
    @RDBA_Delete,
    @RDBA_Edit,
    @RDBA_View,
    @RDBA_Post,
    @RDBA_Cancel,
    @RDBA_Delete,
    @RDBA_NULL
  );

  DBImageIndexes: array[TrDBAction] of Integer = (-1,0,1,2,3,4,5,5,6,7,8,9,10,6,11);

//    ()
// Delete record(s) warning messagebox
function DeleteMessage(Multi: Boolean): Boolean; {$IFDEF POLARIS_D4}overload;
function DeleteMessage(Q: Integer): Boolean; overload;
{$ENDIF}

//    dataset  field
// Check options on dataset and field
function TestOptions(DataSet: TDataSet; Field: TField; PositiveOptions,NegativeOptions: TrDBOptionSet): Boolean;

//    dataset  field
// Execute action with dataset or field
function DoAction(DataSet: TDataSet; Field: TField; Action: TrDBAction): Boolean;

//   
// Recalculate calculating fields
procedure ReCalcField(DataSet: TDataSet);

//       
//      
// Format "Name" field: cutting blanks and convert first letter to upper case
procedure FormatNameField(Field: TStringField);

//   
// Format "Initials" field
procedure FormatFIOField(Field: TStringField);

//    DataSet
// Change dataset filter
procedure ChangeFilter(DataSet: TDataSet; NewFilter: string);

//     
// Set filter value from system registry
procedure SetDBFilter(AModule: TComponent; Reg: TRegistry; ARegKey: String);

{$IFDEF POLARIS_D4}
//   
// Assign parameter (TParam) value
procedure AssignParam(Param: TParam; Value: Variant);

//    
// Clear all TParam's values
procedure ClearParams(Params: TParams);

//     
// Assign parameter (TParam) value from field
procedure AssignParamFromField(Param: TParam; Field: TField);
{$ENDIF}

//   
// DB errors handler
procedure DBErrorEHProc(E: Exception; Handler: TErrorHandler);

implementation

type
  TCrackDataSet=class(TDataSet);

function DeleteMessage(Multi: Boolean): Boolean; {$IFDEF POLARIS_D4}overload;{$ENDIF}
var s: string;
begin
  if Multi then s := SDeleteMultipleRecordsQuestion
  else s := SDeleteRecordQuestion;
  Result := rMsgBox(PChar(s), MB_YESNO+MB_ICONQUESTION+MB_DEFBUTTON2) = IDYES;
end;
{$IFDEF POLARIS_D4}
function DeleteMessage(Q: Integer): Boolean; overload;
var s: string;
begin
  Result := False;
  if Q <= 0 then exit;
  s := Format(srDeleteMultipleRecordsQuestion,
              [Q, ExtractWord(TestNum(Q)+1, srRecord, [','])]);
  Result := rMsgBox(PChar(s), MB_YESNO+MB_ICONQUESTION+MB_DEFBUTTON2) = IDYES;
end;
{$ENDIF}

function TestOptions(DataSet: TDataSet; Field: TField; PositiveOptions,NegativeOptions: TrDBOptionSet): Boolean;
var
  I: TrDBOption;

  function TestOption(O: TrDBOption): Boolean;
  begin
    case O of
      tboFirst     : Result := DataSet.BOF;
      tboLast      : Result := DataSet.EOF;
      tboEmpty     : Result := IsDataSetEmpty(DataSet);
      tboCanModify : Result :=
        (not Assigned(Field) and DataSet.CanModify) or (Assigned(Field) and Field.CanModify);
      tboInserting : Result := DataSet.State=dsInsert;
      tboModifying : Result := DataSet.State=dsEdit;
      tboEditing   : Result := DataSet.State in [dsInsert,dsEdit];
      tboAfterEditAssigned : Result := Assigned(DataSet.AfterEdit);
      tboNULL      : Result := Assigned(Field) and Field.IsNull;
    else
      Result := False;
    end;
  end;

begin
  Result := Assigned(DataSet) and DataSet.Active;
  if (PositiveOptions=[]) and (NegativeOptions=[]) then exit;
  if Result and (PositiveOptions<>[]) then
    for I := Low(TrDBOption) to High(TrDBOption) do
      if I in PositiveOptions then begin
        Result := Result and TestOption(I);
        if not Result then Break;
      end;
  if Result and (NegativeOptions<>[]) then
    for I := Low(TrDBOption) to High(TrDBOption) do
      if I in NegativeOptions then begin
        Result := Result and not TestOption(I);
        if not Result then Break;
      end;
end;

function DoAction(DataSet: TDataSet; Field: TField; Action: TrDBAction): Boolean;
var
  i: Integer;
begin
  if not Assigned(DataSet) then begin
    Result := False;
    exit;
  end;
  Result := True;
  with DataSet do
    case Action of
      tbaFirst  : First;
      tbaPrior  : Prior;
      tbaNext   : Next;
      tbaLast   : Last;
      tbaRefresh: begin
                    if Assigned(Field) then Field.RefreshLookupList
                    else
                      for i:=0 to FieldCount-1 do
                        if (Fields[i].FieldKind = fkLookup) and Fields[i].LookupCache then
                          Fields[i].RefreshLookupList;
                    try Refresh except RefreshQuery(DataSet) end;
                  end;
      tbaInsert : Insert;
      tbaAppend : Append;
      tbaDelete : Delete;
      tbaEdit   : Edit;
      tbaView   : if Assigned(AfterEdit) then AfterEdit(DataSet)
                  else Result := False;
      tbaPost   : Post;
      tbaCancel : Cancel;
      tbaConfirmDelete: if DeleteMessage(False) then Delete else Result := False;
      tbaNULL   : if Field <> nil then Field.Clear else Result := False;
    else
      Result := False;
    end;
end;

procedure ReCalcField(DataSet: TDataSet);
begin
  if DataSet <> nil then
  TCrackDataSet(DataSet).GetCalcFields(DataSet.ActiveBuffer);
end;

procedure FormatNameField;
begin
  with Field do begin
    if not(DataSet.State in [dsEdit,dsInsert]) then exit;
    if Length(Trim(AsString))=0 then begin
      AsVariant := Null;
      exit;
    end;
    AsString := Trim(AsString);
    AsString := Stuff(AsString,1,1,AnsiUpperCase(String(AsString[1])))
  end;
end;

procedure FormatFIOField;
begin
  with Field do begin
    if not(DataSet.State in [dsEdit,dsInsert]) then exit;
    if Length(Trim(AsString))=0 then begin
      AsVariant := Null;
      exit;
    end;
    FormatNameField(Field);
    AsString := AnsiProperCase(Trim(AsString),[' ','.',',','-']);
  end;
end;

procedure ChangeFilter;
begin
  with DataSet do
    if Filter <> NewFilter then begin
      Filter := NewFilter;
      Filtered := (Trim(Filter)<>'') or Assigned(OnFilterRecord);
    end;
end;

procedure SetDBFilter(AModule: TComponent; Reg: TRegistry; ARegKey: String);
Var
  i: Integer;
  RealKey: String;
  r, old_r: HKEY;
begin
  For i:=0 to AModule.ComponentCount-1 do
    if (AModule.Components[i] is TDataSet) then begin
      SeparateRegKey(ARegKey + srFilterKey + (AModule.Components[i] as TDataSet).Name, r, RealKey);
      With Reg do begin
        old_r := RootKey;
        try
          RootKey := r;
          if OpenKey(RealKey, False) then
            try
              (AModule.Components[i] as TDataSet).Filter := ReadString(srFilterValue);
            finally
              CloseKey;
            end;
        finally
          RootKey := old_r;
        end;
      end;
    end;
end;

{$IFDEF POLARIS_D4}
procedure AssignParam(Param: TParam; Value: Variant);
begin
  if VarIsNull(Value) then Param.Clear else Param.Value := Value;
  Param.Bound := True;
end;

procedure ClearParams(Params: TParams);
var
  I: Integer;
begin
  for I := 0 to Params.Count - 1 do Params[I].Clear;
end;

procedure AssignParamFromField(Param: TParam; Field: TField);
begin
  if Field.IsNull then Param.Clear else Param.Value := Field.Value;
  Param.Bound := True;
end;
{$ENDIF}

procedure DBErrorEHProc(E: Exception; Handler: TErrorHandler);
begin
  with Handler do begin
    TitleMessage := srDBErrorTitle;
    Message := ConvertMessage(E.Message);
    if Message = '' then Message := E.Message;
  end;
end;

initialization
  AddError('EDatabaseError',DBErrorEHProc);

  DBActionShortCuts[tbaFirst]   := ShortCut(VK_HOME,[ssCtrl]);
  DBActionShortCuts[tbaLast]    := ShortCut(VK_END,[ssCtrl]);
{$IFDEF REFRESH_F5}
  DBActionShortCuts[tbaRefresh] := ShortCut(VK_F5,[]);
{$ELSE}
  DBActionShortCuts[tbaRefresh] := ShortCut(Word('R'),[ssCtrl]);
{$ENDIF}
  DBActionShortCuts[tbaView]    := ShortCut(VK_F3,[ssCtrl]);
  DBActionShortCuts[tbaPost]    := ShortCut(VK_RETURN,[ssCtrl]);

{$IFDEF POLARIS_D6}
  RHandles.Add(Pointer(GetModuleHandle('vcldb60.bpl')));
  RHandles.Add(Pointer(GetModuleHandle('rxdb6.bpl')));
{$ELSE}
{$IFDEF POLARIS_D5}
  RHandles.Add(Pointer(GetModuleHandle('vcldb50.bpl')));
  RHandles.Add(Pointer(GetModuleHandle('rxdb5.bpl')));
{$ELSE}
{$IFDEF POLARIS_D4}
  RHandles.Add(Pointer(GetModuleHandle('vcldb40.bpl')));
  RHandles.Add(Pointer(GetModuleHandle('rxdb4.bpl')));
{$ELSE}
  RHandles.Add(Pointer(GetModuleHandle('vcldb30.dpl')));
  RHandles.Add(Pointer(GetModuleHandle('rxdb.dpl')));
{$ENDIF}
{$ENDIF}
{$ENDIF}
end.

