//==============================================
//      File rmenubar.pas
//
//      Components library for Delphi, BCB++, Kylix.
//      Menu as in Delphi 4,5.
//
//      Copyright (c) 1995,98 Inprise Corporation
//      Copyright 1998-2002 Polaris Software
//      http://polesoft.da.ru
//      mailto: PolarisLib@mail.ru
//==============================================
unit rMenuBar;

{$I POLARIS.INC}

interface

uses
  SysUtils, Classes,
{$IFDEF LINUX}
  QControls, QComCtrls, QMenus;
{$ELSE}
  Windows, Messages, Graphics, Controls, Forms, Dialogs,
  ToolWin, ComCtrls, Menus;
{$ENDIF}
type
  TrMenuBar = class(TToolBar)
  private
{$IFNDEF POLARIS_D6}
    FMenu: TMainMenu;
    procedure SetMenu(const Value: TMainMenu);
{$ENDIF}
{$IFNDEF LINUX}
    function HandleAppKeyDown(var Message: TWMKey): Boolean;
    procedure CMDialogKey(var Message: TWMKey); message CM_DIALOGKEY;
{$ENDIF}
  protected
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override;
{$IFNDEF POLARIS_D6}
    procedure Notification(AComponent: TComponent;
      Operation: TOperation); override;
{$ENDIF}
    procedure ClearButtons;
  public
    constructor Create(AOwner: TComponent); override;
    procedure RefreshButtons;
  published
    property EdgeBorders default [];
{$IFNDEF POLARIS_D6}
    property Menu: TMainMenu read FMenu write SetMenu;
{$ENDIF}
  end;

implementation

{ TrMenuBar }

constructor TrMenuBar.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Flat := True;
  ShowCaptions := True;
  EdgeBorders := [];
  ControlStyle := [csCaptureMouse, csClickEvents,
    csDoubleClicks, csMenuEvents, csSetCaption];
  RecreateWnd;
end;

{$IFNDEF LINUX}
procedure TrMenuBar.CMDialogKey(var Message: TWMKey);
begin
  if not HandleAppKeyDown(Message) then Inherited;
end;

function TrMenuBar.HandleAppKeyDown(var Message: TWMKey): Boolean;
begin
  if Assigned(Menu) and Menu.IsShortCut(Message) then begin
    Message.Result := 1;
    Result := True;
  end else
    Result := False;
end;
{$ENDIF}

procedure TrMenuBar.ClearButtons;
var
  i: Integer;
begin
{$IFDEF LINUX}
  for i := ControlCount - 1 downto 0 do Controls[i].Free;
{$ELSE}
  for i := ButtonCount - 1 downto 0 do Buttons[i].Free;
{$ENDIF}
end;

procedure TrMenuBar.GetChildren(Proc: TGetChildProc; Root: TComponent);
begin
end;

{$IFNDEF POLARIS_D6}
procedure TrMenuBar.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FMenu)
  then begin
    ClearButtons;
    FMenu := nil;
  end;
end;

procedure TrMenuBar.SetMenu(const Value: TMainMenu);
var
  i: Integer;
begin
  if FMenu = Value then exit;
  if Assigned(FMenu) then ClearButtons;
  FMenu := Value;
  if not Assigned(FMenu) then exit;
  Value.FreeNotification(Self);
  for i := 0 to FMenu.Items.Count - 1 do
    with TToolButton.Create(Self) do
      try
        AutoSize := True;
        Grouped := True;
        Parent := Self;
      except
        Free;
        raise;
      end;
{$IFDEF LINUX}
//  for i := 0 to ControlCount - 1 do TToolButton(Controls[i]).MenuItem := FMenu.Items[i];
{$ELSE}
  for i := 0 to ButtonCount - 1 do Buttons[i].MenuItem := FMenu.Items[i];
{$ENDIF}
end;
{$ENDIF}

procedure TrMenuBar.RefreshButtons;
var
  i: Integer;
begin
{$IFNDEF LINUX}
  for i := 0 to ButtonCount - 1 do with Buttons[i] do Enabled := MenuItem.Enabled;
{$ENDIF}
end;

end.
