//==============================================
//      File rperiodcomboedit.pas
//
//      Components library for Delphi and BCB++.
//        .
//      Choice dates period.
//
//      Copyright 1998-2002 Polaris Software
//      http://polesoft.da.ru
//      mailto: PolarisLib@mail.ru
//==============================================
unit rPeriodComboEdit;

{$I POLARIS.INC}

interface

uses
  SysUtils, Classes, Controls, rUtils, rConst,
  rButtons, rComboEdit, rPeriod, rPeriodDlg;

type
  TrPeriodComboEdit = class(TrCustomComboEdit)
  private
    FPeriod: TrPeriod;
    FNeedPrefix: Boolean;
    function GetDate1: TDateTime;
    function GetDate2: TDateTime;
    procedure SetDate1(Value: TDateTime);
    procedure SetDate2(Value: TDateTime);
  protected
    procedure ButtonClick; override;
    function IsNotDefaultGR: Boolean;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Period: TrPeriod read FPeriod;
    property Date1: TDateTime read GetDate1 write SetDate1;
    property Date2: TDateTime read GetDate2 write SetDate2;
    procedure AssignDate(dDate: TDateTime; dTypePeriod: TTypePeriod; Offset: Integer);
    procedure AssignNum(NumPeriod,Year: Integer; dTypePeriod: TTypePeriod);
  published
    property NumGlyphs default 2;
    property GlyphResKind;
    property GlyphResource stored IsNotDefaultGR;
    property OnCloseUp;
    property NeedPrefix: Boolean read FNeedPrefix write FNeedPrefix default True;
  end;

implementation

{ TrPeriodComboEdit }

constructor TrPeriodComboEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPeriod := TrPeriod.Create;
  GlyphResource := srPeriodGlyph;
  NumGlyphs := 2;
  FNeedPrefix := True;
  Date1 := SysUtils.Date;
  Date2 := SysUtils.Date;
end;

destructor TrPeriodComboEdit.Destroy;
begin
  FPeriod.Free;
  inherited;
end;

procedure TrPeriodComboEdit.ButtonClick;
begin
  inherited;
  if EditPeriod(Period) then begin
    Text := Period.GetName(FNeedPrefix);
    if Assigned(OnCloseUp) then OnCloseUp(Self);
  end;
end;

function TrPeriodComboEdit.IsNotDefaultGR: Boolean;
begin
  Result := UpperCase(GlyphResource) <> srPeriodGlyph;
end;

function TrPeriodComboEdit.GetDate1: TDateTime;
begin
  Result := FPeriod.D1;
end;

function TrPeriodComboEdit.GetDate2: TDateTime;
begin
  Result := FPeriod.D2;
end;

procedure TrPeriodComboEdit.SetDate1(Value: TDateTime);
begin
  if FPeriod.D1 <> Value then begin
    FPeriod.D1 := Value;
    Text := FPeriod.GetName(FNeedPrefix);
  end;
end;

procedure TrPeriodComboEdit.SetDate2(Value: TDateTime);
begin
  if FPeriod.D2 <> Value then begin
    FPeriod.D2 := Value;
    Text := FPeriod.GetName(FNeedPrefix);
  end;
end;

procedure TrPeriodComboEdit.AssignDate(dDate: TDateTime; dTypePeriod: TTypePeriod; Offset: Integer);
begin
  FPeriod.AssignDate(dDate, dTypePeriod, Offset);
  Text := FPeriod.GetName(FNeedPrefix);
end;

procedure TrPeriodComboEdit.AssignNum(NumPeriod,Year: Integer; dTypePeriod: TTypePeriod);
begin
  FPeriod.AssignNum(NumPeriod, Year, dTypePeriod);
  Text := FPeriod.GetName(FNeedPrefix);
end;

end.
